#include <itf/option.h>
#include <itf/itf.h>
#include <sim/sim.h>
#include "prop-bp.h"


int CompileNet(net)
NETWORK *net;
{
  NET_PARAMS *np = net->p;
  int n_units = 0;
  int n_links = 0;
  int n_bunits = 0;
  int n_blinks = 0;
  int total_n_units = 0;
  int total_n_bunits = 0;
  int total_n_links  = net->n_links;
  int u,l;
  /*
   * Count units and links
   */
  ForEachForwardUnit(UNIT) {
    total_n_units++;
    if (!unit->lesioned) n_units++;
    ApplyToLinksOfUnit(LINK,UNIT,if (!link->lesioned) n_links++;);
  } EndForEach; /* Unit */
#ifndef DBM_NETWORK
  ForEachBackwardUnit(UNIT) {
    total_n_bunits++;
    if (!unit->lesioned) n_bunits++;
    ApplyToLinksOfUnit(LINK,UNIT,if (!link->lesioned) n_blinks++;);
  } EndForEach; /* Unit */
#endif  DBM_NETWORK
  np->n_units  = n_units;
  np->n_links  = n_links;
  np->n_bunits = n_bunits;
  np->n_blinks = n_blinks;
  /* 
   * Create vectors for unit/link pointers
   */
  if (!np->compiled) {
    if (np->n_units) {
      np->units = (UNIT_PTR*)calloc(total_n_units,sizeof(UNIT_PTR));
      if (np->units == NULL) {
	IOutOfMemory("Compile", "allocating np->units (requested %d * %d)",
		     total_n_units,sizeof(UNIT_PTR));
	goto out_of_memory;
      }
      np->links = (LINK_PTR*)calloc(total_n_links,sizeof(LINK_PTR));
      if (net->p->links == NULL) {
	IOutOfMemory("Compile", "allocating np->links (requested %d * %d)",
		     total_n_links,sizeof(LINK_PTR));
	goto out_of_memory;
      }
      np->s_units = (UNIT_PTR*)calloc(total_n_links,sizeof(UNIT_PTR));
      if (net->p->s_units == NULL) {
	IOutOfMemory("Compile", "allocating np->s_units (requested %d * %d)",
		     total_n_links,sizeof(UNIT_PTR));
	goto out_of_memory;
      }
      np->d_units = (UNIT_PTR*)calloc(total_n_links,sizeof(UNIT_PTR));
      if (net->p->d_units == NULL) {
	IOutOfMemory("Compile", "allocating np->d_units (requested %d * %d)",
		     total_n_links,sizeof(UNIT_PTR));
	goto out_of_memory;
      }
    }
    if (np->n_bunits) {
      np->bunits = (UNIT_PTR*)calloc(total_n_bunits,sizeof(UNIT_PTR));
      if (np->bunits == NULL) {
	IOutOfMemory("Compile", "allocating np->bunits (requested %d * %d)",
		     total_n_bunits,sizeof(UNIT_PTR));
	goto out_of_memory;
      }
      np->blinks = (LINK_PTR*)calloc(total_n_links,sizeof(LINK_PTR));
      if (net->p->blinks == NULL) {
	IOutOfMemory("Compile", "allocating np->blinks (requested %d * %d)",
		     total_n_links,sizeof(LINK_PTR));
	goto out_of_memory;
      }
      np->s_bunits = (UNIT_PTR*)calloc(total_n_links,sizeof(UNIT_PTR));
      if (net->p->s_bunits == NULL) {
	IOutOfMemory("Compile", "allocating np->s_bunits (requested %d * %d)",
		     total_n_links,sizeof(UNIT_PTR));
	goto out_of_memory;
      }
      np->d_bunits = (UNIT_PTR*)calloc(total_n_links,sizeof(UNIT_PTR));
      if (net->p->d_bunits == NULL) {
	IOutOfMemory("Compile", "allocating np->d_bunits (requested %d * %d)",
		     total_n_links,sizeof(UNIT_PTR));
	goto out_of_memory;
      }
    }
  }
  /*
   * Put pointers in arrays
   */
  l = 0; u = 0;
  ForEachForwardUnit(UNIT) {
    if (!unit->lesioned) np->units[u++] = unit;
    ApplyToLinksOfUnit
      (LINK,UNIT,
       if (!link->lesioned) {
	 np->s_units[l] = other;
	 np->d_units[l] = unit;
	 np->links[l++] = link;
       }
       );
  } EndForEach; /* Unit */
#ifndef DBM_NETWORK
  l = 0; u = 0;
  ForEachBackwardUnit(UNIT) {
    if (!unit->lesioned) np->bunits[u++] = unit;
    ApplyToLinksOfUnit
      (LINK,UNIT,
       if (!link->lesioned) {
	 np->s_bunits[l] = other;
	 np->d_bunits[l] = unit;
	 np->blinks[l++] = link;
       }
       );
  } EndForEach; /* Unit */
#endif DBM_NETWORK
  np->compiled = 1;
  return 1;
 out_of_memory:
  return 0;
}


int command_compile(tokc,tokv)
int tokc;
char *tokv[];
{
  IUsage("");
  if (GiveHelp(tokc)) {
    ISynopsis("compile current net into efficient data structures.");
    IHelp(IHelpArgs,
"Compile network into efficient data structures for forward/backward",
"pass.  Should be called after network is completely built.",
	  NULL);
    return;
  }
  if (!NetExists(net)) {
    IError("current net is null");
    return;
  }
  fprintf(dout,"Compiling network...");
  fflush(dout);
  if (!CompileNet(net)) {
    IError("error in compiling net");
    return;
  }
  fprintf(dout,"done\n");
}
