#include <itf/option.h>
#include <itf/itf.h>
#include <sim/sim.h>
#include <lib/example.h>
#include "prop-bp.h"
#include "example-bp.h"

#if defined(ibmrt)
#define RangePosition(a,b,c,d,e)	AreaPosition(a,b,c,d,e)
#define RangeDimensions(a,b,c,d,e) 	AreaDimensions(a,b,c,d,e) 
#endif
	
#define DB(x)

#define LayerFormat(n) (&(net->layer[n].format))

#define MAX_COMMAND_LEN 10000

extern FILE *IOpenFileOrAbort();	/* itf.c */

real GetLinkField(net,sl,su,dl,du,field_name)
NETWORK *net;
int sl,su,dl,du;
char *field_name;
{
  real weight;
  int take_address;
  char *p;
  ADDRESS_ENV env;
  DoLayerUnit(dl,UNIT,du) {
    ForEachLinkSet(LINK,UNIT) {
      if (field_name!=NULL) ICreateEnv(&env,nameof_link[ls->link_tag],NULL);
      if (OtherLayerN == sl)
	ApplyToEachLink
	  (
	   if (OtherUnitN == su) {
	     if (link->lesioned)        return 0.0;
	     else if (field_name==NULL) return LINK_WEIGHT(link);
	     else {
	       env.poi.ptr = (char*)link;
	       if (!IContParseObject(field_name,&env,&take_address)) 
		 IErrorAbort("trouble parsing field of link");
	       p = IGetSimpleValue(&env,0,NULL,0);
	       if (p==NULL || itf_errno)
		 IErrorAbort("trouble parsing field of link");
	       weight = atof(p);
	       if (!strcmp(field_name,HEBB_NAME)) weight *= net->p->hebb_scale;
	     }
	     return weight;
	   }
	   );
    } EndForEach; /* Link */
  } EndDo; /* Unit */
  return 0.0;
}


/*****************************************************************************
 * command_display
 */

int command_display(tokc,tokv)
int tokc;
char *tokv[];
{
  int n, i, t, max_width, bias_pos;
  int sl,su,dl,du;
#ifdef TIME_NETWORK
  int time_network = TRUE;
#else
  int time_network = FALSE;
#endif TIME_NETWORK
#ifdef DBM_NETWORK
  int incoming = TRUE, outgoing = TRUE;
#else
  int incoming = -1, outgoing = FALSE;
#endif DBM_NETWORK
  int units_printed, t_min = 0, t_max = 0;
  int t_inc = 0;		/* extra space inside state box for targets */
  int t_inc_total = 0;		/* extra space inside state box for targets */
  int show_targets, example = 0, example_n = -1;
  char *example_str = NULL;
  int debug = FALSE, states = FALSE, weights = TRUE, boxed = FALSE, boxmarks = FALSE, map = FALSE;
  int do_bias = TRUE, clear = TRUE, force = FALSE, bias = FALSE, ranges = FALSE, invert = FALSE;
  int spread = FALSE, targets = TRUE;
  char *label_str = NULL;
  int number_links = 0, number_size = 0, complex_ranges = FALSE, space = FALSE;
  int *layer_pos; /* records the position of each layer */
  int *layer_width, *layer_height, *layer_display_targets;
  real scale=5;
  int fixed_scale = FALSE;
  real s, ss, mean, var, max_abs;
  char unit_name[10];
  char *field_name = NULL;
  real weight;
  int *slayer, *dlayer;
  struct RANGE *sunit_range;
  struct RANGE *dunit_range;
  static int last_n_links=0;	/* if these are the same, assume we have */
  static int last_net_n=0;	/* sent the plan of the net */
  static int last_option_check;	/* sum of all the characters in the options */
  int option_check = FALSE, repetitive = FALSE;
  int rows, cols, depth, row, col;
  int weight_precision = 2;	/* number of decimal places of accuracy */
  extern char *IGetSimpleValue();
  static struct OPTION options[] =
    {
      {OPTPRES,			"ex",		0},
      {OPTPRES|OPTREALARG,	"scale",	0},
      {OPTPRES|OPTINTARG,	"precision",	0},
      {OPTPRES|OPTINTARG,	"tmin",		0},
      {OPTPRES|OPTINTARG,	"tmax",		0},
      {OPTPRES,			"debug",	0},
      {OPTPRES,			"reverse",	0},
      {OPTPRES,			"number",	0},
      {OPTSIGN,          	"incoming",	0},
      {OPTSIGN,          	"outgoing",	0},
      {OPTSIGN,			"clear",	0},
      {OPTSIGN,          	"time",		0},
      {OPTSIGN,          	"map",		0},
      {OPTSIGN,          	"states",	0},
      {OPTSIGN,          	"targets",	0},
      {OPTSIGN,          	"bias",		0},
      {OPTSIGN,          	"weights",	0},
      {OPTSIGN,          	"box",		0},
      {OPTSIGN,          	"marks",	0},
      {OPTSIGN,          	"ranges",	0},
      {OPTSIGN,          	"repetitive",	0},
      {OPTSIGN,          	"forcerep",	0},
      {OPTSIGN,          	"space",	0},
      {OPTSIGN,          	"spread",	0},
      {OPTPRES,          	"slayer",	0},
      {OPTPRES,          	"sunit",	0},
      {OPTPRES,          	"dlayer",	0},
      {OPTPRES,          	"dunit",	0},
      {OPTPRES|OPTSTRARG,	"field",	0},
      {OPTPRES|OPTSTRARG,	"label",	0},
      {OPTLAST,NULL,0}
    };
  IUsage(" [%pex <int>] [%pscale <real>] [%pprecision <int>]\
 [%preverse] [%sclear] [%pnumber] [%stime] [%smap] [%sstates]\
 [%sbias] [%sweights] [%sbox] [%smarks] [%sranges] [%srepetitive]\
 [%sforcerep] [%sspace] [%sspread] [%pfield <var>] [%sincoming]\
 [%soutgoing] [%ptmin <int>] [%ptmax <int>] [%plabel <string>]\
 [%stargets]");
  if (GiveHelp(tokc)) {
    ISynopsis("display graphically the weights/states of the network");
    IHelp
      (IHelpArgs,
"This command produces a description of the network which can be",
"interpreted by the `bboard' program.",
"To show the network once, use the simple command:",
"    display | bboard",
"The display of weights may flash and dissappear, if this happens, do:",
"    set TERM = xterm-nas",
"to tell curses not to use the alternate screen.",
"OPTIONS",
"%pscale <SCALE>: use <SCALE> as the scaling value for weights",
"%pprecision <P>: use <P> decimal places of precision when writing weights",
"%pex <n>       : show states for example n",
"%ptmin <n>     : first iteration to display states (default 0).",
"%ptmin <n>     : last iteration to display states (default np->t_max).",
"%preverse      : tell bboard to reverse the color of the blots",
"%pnumber       : print out numbers of the links, rather than weights",
"%sclear      : don't send a clear message (default %y)",
"%stime       : do (not) show states for each iteration for time-network",
"%smap        : do (not) produce a map of the network (default %y)",
"%sstates     : do (not) show the states of the units (default %n)",
"%stargets    : do (not) show the targets of the units (default %y)",
"%sbias       : do (not) show a weight map for the bias unit (default %y)",
"%sweights    : do (not) show weight maps (default %y)",
"%sbox        : do (not) put boxes around all the maps (default %n)",
"%smarks      : do (not) put alphabetical marks on boxes (default %y)",
"%sspace      : do (not) put space between unboxed displays (default %y)" ,
"%sspread     : do (not) put space bewteen layers within displays (default %n)" ,
"%sranges     : do (not) show a map with all the unit ranges (default %n)",
"%sincoming   : do (not) display incoming weights (default %y)",
"%soutgoing   : do (not) display outgoing weights (default %n)",
"%prepetitive   : the use is repetitive, so don't output a complete plan",
"                  each time the command is called.  Saves much time.",
"%pforcerep     : do as though repetitive, even if command thinks not.",
"                  (default + if +sunit is used)",
"%pfield <var>  : display field <var> (e.g. .momentum) instead of weights.",
"%plabel <str>  : display string.",
"%pdebug        : print out which units and links will be displayed,",
"                  and then return without doing anything else",
"SPECIFYING WHICH UNITS AND LINKS TO DISPLAY",
"The dlayer and dunit options control which units get a rectangle with",
"their weights, and the slayer and sunit options control which weights",
"are shown within the rectangles.  There can be any number of dlayer",
"and slayer options, but there can be only one sunit and one dunit",
"option for each layer.",
"%sdlayer <range>: do (not) show units in layers in <range>",
"%pdunit <layerrange>:<unitrange>",
"                : only show these units in these layers",
"%sslayer <range>: do (not) show links from units in layers in <range>",
"%psunit <layerrange>:<unitrange>",
"                : only show links from these units for these layers",
"The slayer and sunit options control which links get shown for each unit,",
"it does not matter, and it has nothing to do with whether the links are",
"inward or outward links.",
"FORMATTING LAYERS",
"Layer formats are specified (permanently) using the \"layer-format\"",
"command.",
"For help with using ranges, see the \"range\" command.",
"PRINTOUTS",
"To get a printout of a weight diagram, use the \"xdpr\" command, which",
"will dump the contents of a window to the printer.  The easiest way",
"to get labels on the printout is to use the \"xstring\" command to put",
"up a small window with a string in it, and use the mouse to position",
"the string where desired.",
"REPETITIVE USAGE",
"To be continued...",
NULL);
    return;
  }
  if (!NetExists(net)) {
    IError("current net is null");
    return;
  }
  layer_pos    = ITempCalloc(int,net->n_layers);
  layer_width  = ITempCalloc(int,net->n_layers);
  layer_height = ITempCalloc(int,net->n_layers);
  layer_display_targets = ITempCalloc(int,net->n_layers);
  AllocDisplayRanges(net,&dlayer,&slayer,&dunit_range,&sunit_range);
  for (i=0;i<net->n_layers;i++) layer_pos[i] = -1;

  tokv++;	/* ignore command name */
  while (IsOption(tokv[0])) {
    if        (OptionMatch(tokv[0],"ex",options,&example)) {
      if (IIsInteger(tokv[1])) example_n = atoi(*++tokv);
      else 		       example_str = *++tokv; 
      tokc--;
      states = TRUE;
      weights = FALSE;
    } else if (OptionMatch(tokv[0],"scale",	options,NULL)) {
      if (!IExpectNumber(tokv[1])) return;
      scale = atof(*++tokv); tokc--;
      fixed_scale = TRUE;
    } else if (OptionMatch(tokv[0],"precision",	options,NULL)) {
      if (!IExpectInteger(tokv[1])) return;
      weight_precision = atoi(*++tokv); tokc--;
    } else if (OptionMatch(tokv[0],"tmin",	options,NULL)) {
      if (!IExpectInteger(tokv[1])) return;
      t_min = atoi(*++tokv); tokc--;
    } else if (OptionMatch(tokv[0],"tmax",	options,NULL)) {
      if (!IExpectInteger(tokv[1])) return;
      t_max = atoi(*++tokv); tokc--;
    } else if (OptionMatch(tokv[0],"debug",	options,&debug)) {
    } else if (OptionMatch(tokv[0],"incoming",	options,&incoming)) {
    } else if (OptionMatch(tokv[0],"outgoing",	options,&outgoing)) {
    } else if (OptionMatch(tokv[0],"reverse",	options,&invert)) {
    } else if (OptionMatch(tokv[0],"clear",	options,&clear)) {
    } else if (OptionMatch(tokv[0],"number",	options,NULL)) {
      number_links = TRUE;
      fixed_scale = TRUE;
      scale = TRUE;
      number_size = 1+floor(1+log10((real)net->n_links));
      map = FALSE;
    } else if (OptionMatch(tokv[0],"time",	options,&time_network)) {
    } else if (OptionMatch(tokv[0],"map",	options,&map)) {
    } else if (OptionMatch(tokv[0],"states",	options,&states)) {
    } else if (OptionMatch(tokv[0],"targets",	options,&targets)) {
    } else if (OptionMatch(tokv[0],"bias",	options,&do_bias)) {
      dlayer[np->bias_layer_n] = do_bias ;
    } else if (OptionMatch(tokv[0],"weights",	options,&weights)) {
      map = weights;
    } else if (OptionMatch(tokv[0],"space",	options,&space)) {
    } else if (OptionMatch(tokv[0],"spread",	options,&spread)) {
    } else if (OptionMatch(tokv[0],"box",	options,&boxed)) {
      if (boxed) space = FALSE;
    } else if (OptionMatch(tokv[0],"marks",	options,&boxmarks)) {
      if (boxmarks) boxed = TRUE;
    } else if (OptionMatch(tokv[0],"ranges",	options,&ranges)) {
    } else if (OptionMatch(tokv[0],"repetitive",options,&repetitive)) {
    } else if (OptionMatch(tokv[0],"forcerep",	options,&force)) {
    } else if (OptionMatch(tokv[0],"slayer",	options,&complex_ranges) ||
	       OptionMatch(tokv[0],"dlayer",	options,&complex_ranges) ||
	       OptionMatch(tokv[0],"sunit", 	options,&complex_ranges) ||
	       OptionMatch(tokv[0],"dunit", 	options,&complex_ranges)) {
      if (!SetDisplayRanges(net,&tokv,dlayer,slayer,dunit_range,sunit_range))
	return;
    } else if (OptionMatch(tokv[0],"field",options,NULL)) {
      field_name = *++tokv; tokc--;
    } else if (OptionMatch(tokv[0],"label",options,NULL)) {
      label_str = *++tokv; tokc--;
     } else {
       IError("option not understood: \"%s\"",tokv[0]);
       return;
     }
     tokv++;
     tokc--;
   }
  if (repetitive) {
    if ((last_net_n!=net->my_n || last_n_links!=net->n_links
	 || last_option_check != option_check) && !force)
      repetitive = FALSE;
    last_net_n = net->my_n;
    last_n_links = net->n_links;
    last_option_check = option_check;
  }
  if (!complex_ranges) dlayer[np->bias_layer_n] = FALSE;
  if (outgoing) {
    if (incoming==-1) incoming = FALSE;
  } else if (np->input_layer_n < np->forward_layer_n)
    dlayer[np->input_layer_n] = FALSE;
  if (np->bias_layer_n<0 || !dlayer[np->bias_layer_n]) bias = FALSE;
  if (debug) {
    PrintDisplayRanges(net,dlayer,slayer,dunit_range,sunit_range);
    return;
  }
  if (!repetitive && clear) fprintf(dout,"clear\n");
  /******************************************************************************
   * calculate the weight scaling, and mark layers which are sources of links
   */
  n = s = ss = max_abs = 0;
  if (weights) {
    if (!fixed_scale) {
      for (dl=0; dl<net->n_layers; dl++) {
	if (!dlayer[dl]) continue;
	for (du=dunit_range[dl].lo; du<=dunit_range[dl].hi; du++) 
	  for (sl=0; sl<net->n_layers; sl++) {
	    if (!slayer[sl]) continue;
	    for (su=sunit_range[sl].lo; su<=sunit_range[sl].hi; su++) {
	      if (incoming &&
		  (weight=GetLinkField(net,sl,su,dl,du,field_name))!= 0.0) {
		n++;  s += fabs(weight);  ss += weight*weight;
		if (fabs(weight)>max_abs) max_abs = fabs(weight);
	      }
	      if (outgoing &&
		  (weight=GetLinkField(net,dl,du,sl,su,field_name))!= 0.0) {
		n++;  s += fabs(weight);  ss += weight*weight;
		if (fabs(weight)>max_abs) max_abs = fabs(weight);
	      }
	    }
	  }
      }
      mean = s/n;
      var = (ss  - s*s/n)/(n-1);
      scale = mean + 3*sqrt(var);
      if (scale>max_abs) scale=max_abs;
      if (fabs(scale)<1e-20) scale=1;
      fprintf(dout,"name %s\nstdscr\ns 0 0 Network \"%s\" u(|w|)=%.*g sd(|w|)=%.*g\ns 0 -25 weight scaling = %.*g\n",
	      net->name,net->name,weight_precision,mean,weight_precision,
	      sqrt(var),weight_precision,scale);
    } else
      fprintf(dout,"name %s\nstdscr\ns 0 0 Network \"%s\" \ns 0 -31 fixed weight scaling = %.*g\n",
	      net->name,net->name,weight_precision,scale);
  } else
    fprintf(dout,"name %s\nstdscr\ns 0 0 Network \"%s\"\n",net->name,net->name);
  if (!repetitive) {
    if (boxed) fprintf(dout,"allbox\n");
    if (!boxmarks) fprintf(dout,"noboxmarks\n");
    if (invert) fprintf(dout,"invert\n");
    if (number_size) fprintf(dout,"numsize %d\n",number_size);
  }
  /******************************************************************************
   * Plan the layout for a complete display
   */
  for (i=0;i<net->n_layers;i++) {
    layer_pos[i] = 0;
    layer_width[i] = 0;
    layer_height[i] = 0;
  }
  depth = 0;
  max_width = 0;
  ForEachLayerDownType(UNIT) {
    if (!slayer[layer_n]) continue;
    if (layer_n==np->bias_layer_n) continue;
    RangeDimensions(net->layer[layer_n].n_units,&(sunit_range[layer_n]),
		   LayerFormat(layer_n),&rows,&cols);
    layer_pos[layer_n] = depth;
    depth += rows;
    layer_width[layer_n] = cols;
    layer_height[layer_n] = rows;
    if (cols > max_width) max_width = cols;
  } EndForEach;
  /*
   * find a position for the bias weight
   */
  if (np->bias_layer_n>=0) {
    bias_pos = -1;
    ForEachLayerDownType(UNIT) {
      if (layer_n==np->bias_layer_n) continue;
      if (slayer[layer_n] && layer_width[layer_n]<(max_width-1)) {
	bias_pos = layer_pos[layer_n]
	  +(layer_pos[layer_n]>0 && layer_height[layer_n]>1);
	break;
      }
    } EndForEach;
    if (bias_pos==-1) {
      bias_pos = 0;
      max_width += 2; /* need to add 2 because of dumb way we center layers*/
    }
    layer_pos[np->bias_layer_n] = bias_pos;
  }
  bias = FALSE;
  if (do_bias && weights && np->bias_layer_n>=0) {
    DoLayer(np->bias_layer_n) {
      DoUnit(UNIT,0) {
	if (info->fan_in>0) bias=TRUE;
      } EndDo;
    } EndDo;
  }
/******************************************************************************
   * Present example
   */
  if (example) {
    if (!LoadExample(net,example_n,example_str)) return;
    PerformNet(net);
    fprintf(dout,"s 0 -60 Example %d", np->example_n);
    if (current_el->index[np->example_n]->comment)
      fprintf(dout,"(%s)",current_el->index[np->example_n]->comment);
    fprintf(dout,": sumsq-errors = %f  n-errors = %d\n",
	    np->ex_sumsq_errors,np->ex_n_errors);
  }
  /*
   * Start generating actual display(s)
   */
  if (label_str!=NULL) fprintf(dout,"newrow\ns 1 0 %s\nnewrow\n",label_str);
  if (map || states || bias || ranges) fprintf(dout,"newrow\nrowhint %d\n",map+states+bias+ranges);
/******************************************************************************
   * States
   */
  if (states) {
    fprintf(dout,"newrow\nrowhint 1\n");
    /*
     * Display each iteration....
     */
    for (i=0;i<net->n_layers;i++) {
      layer_display_targets[i] = FALSE;
#ifdef DBM_NETWORK
      if (   targets
	  && slayer[i] 
	  && (net->layer[i].p->is_input_layer || net->layer[i].p->is_output_layer )
	  && i!=np->output_layer_n) {
	layer_display_targets[i] = TRUE;
	t_inc_total += layer_height[i];
      }
#endif DBM_NETWORK
    }
    if (targets && slayer[np->output_layer_n]) {
      layer_display_targets[np->output_layer_n] = TRUE;
      t_inc_total += layer_height[np->output_layer_n];
    }
    if (spread) t_inc_total += net->n_layers-2;		/* only between and no bias */
    if (time_network && !t_max) t_max = np->max_t;
    for (t=t_min; t<=t_max; t++) {
      t_inc = 0;
      show_targets = (targets && (!time_network || IsTrainingIteration(t)));
      if (time_network)	fprintf(dout,"window %d %d Iteration %d\n",depth+(show_targets?t_inc_total:0),max_width,t);
      else		fprintf(dout,"window %d %d States\n",      depth+t_inc_total,max_width);
      ForEachLayerDownType(UNIT) {
	if (!slayer[layer_n] || IsBiasLayer) continue;
	/*
	 * Output targets
	 */
	if (show_targets && layer_display_targets[layer_n] && !IsInputLayer) {
	  fprintf(dout,"width %d\n",layer_width[layer_n]);
	  ForEachUnit(UNIT) {
	    if (RangePosition(unit_n,&(sunit_range[layer_n]),&layer->format,&row,&col))
	      fprintf(dout,"b %d %d %.*f\n", 
		      row+t_inc+layer_pos[layer_n], col, weight_precision, 
		      (time_network?UNIT_TARGET_T(unit,t):UNIT_TARGET(unit)));
	  } EndForEach;
	  t_inc += layer_height[layer_n];
	} /* targets */
	/*
	 * Unit states
	 */
	fprintf(dout,"width %d\n",layer_width[layer_n]);
	ForEachUnit(UNIT) {
	  if (RangePosition(unit_n,&(sunit_range[layer_n]),&layer->format,&row,&col))
	    fprintf(dout,"b %d %d %.*f\n",
		    row+t_inc+layer_pos[layer_n], col, weight_precision, 
		    (time_network?UNIT_OUTPUT_T(unit,t):UNIT_OUTPUT(unit)));
	} EndForEach; /* Unit */
	/*
	 * Input targets (DBM)
	 */
	if (show_targets && layer_display_targets[layer_n] && IsInputLayer) {
	  t_inc += layer_height[layer_n];
	  fprintf(dout,"width %d\n",layer_width[layer_n]);
	  ForEachUnit(UNIT) {
	    if (RangePosition(unit_n,&(sunit_range[layer_n]),&layer->format,&row,&col))
	      fprintf(dout,"b %d %d %.*f\n", 
		      row+t_inc+layer_pos[layer_n], col, weight_precision, 
		      (time_network?UNIT_TARGET_T(unit,t):UNIT_TARGET(unit)));
	  } EndForEach;
	} /* input targets */
	if (spread) t_inc++;
      } EndForEach; /* Layer */
      if (space) fprintf(dout,"newrow\nrowhint 1\n");
    } /* Iteration */
  } /* States */
/******************************************************************************
   * Map
   */
  if (map) {
    fprintf(dout,"window %d %d Map\n",depth,max_width);
    i = 0;
    ForEachLayerDownType(UNIT) {
      if (!slayer[layer_n]) continue;
      fprintf(dout,"width %d\n",layer_width[layer_n]);
      ForEachUnit(UNIT) {
	if (RangePosition(unit_n,&(sunit_range[layer_n]),&layer->format,&row,&col))
	  fprintf(dout,"b %d %d %.2f\n",layer_pos[layer_n]+row,col,
		  layer_n==np->bias_layer_n?-0.9:i%2?-0.5:0.5);
      } EndForEach;
      i++; /* just for counting odd/even to decide color of map */
    } EndForEach;
  }
  if (ranges) {
    fprintf(dout,"window %d %d $Layer--Units\n",depth,16);
    i = 0;
    ForEachLayerDownType(UNIT) {
      char str[10];
      int lo,hi;
      if (!slayer[layer_n]) continue;
      if (layer_n==np->bias_layer_n) continue;
      sprintf(str,"%d",layer_n);
      fprintf(dout,"s %d 4 $%s\n",layer_pos[layer_n],str);
      lo = sunit_range[layer_n].lo;
      hi = sunit_range[layer_n].hi;
      for (i=0;i<sunit_range[layer_n].depth;i++) {
	sprintf(str,"%d-%d",lo,hi);
	fprintf(dout,"s %d 8 $%s\n",layer_pos[layer_n]+i,str);
	lo += layer->format.cols;
	hi += layer->format.cols;
      }
    } EndForEach;
    if (slayer[np->bias_layer_n]) fprintf(dout,"s %d 0 $B\n",bias_pos);
  }
/******************************************************************************
   * Weights
   */
  if (weights) {
    if (number_links) {
      i = 0;
      ApplyToLinksOfNet(UNIT,LINK,UNIT,if (IsDestSet && !link->lesioned) LINK_WEIGHT(link)=i++;);
    }
    /*
     * Major unit(s)
     */
    for (dl=net->n_layers-1; dl>=0; dl--) {
      if (!dlayer[dl] || (!do_bias && dl==np->bias_layer_n)) continue;
      RangeDimensions(net->layer[dl].n_units,&(dunit_range[dl]),LayerFormat(dl),&rows,&cols);
      units_printed = 0;
      for (du=dunit_range[dl].lo; du<=dunit_range[dl].hi; du++) {
	if (!InRange(du,&(dunit_range[dl]),LayerFormat(dl))) continue;
	if ((dl!=np->bias_layer_n || outgoing) && (units_printed % cols)==0)
	  fprintf(dout,"newrow\nrowhint %d\n",cols);
	units_printed++;
	if (dl==np->bias_layer_n)
	  sprintf(unit_name,(net->layer[dl].n_units>1?"Bias %d":"Bias"),du);
	else
	  sprintf(unit_name,"Unit %d:%d",dl,du);
	fprintf(dout,"window %d %d %s\n",depth,max_width,unit_name);
	if (slayer[dl] &&
	    RangePosition(du,&(sunit_range[dl]),LayerFormat(dl),&row,&col))
	  fprintf(dout,"width %d\nx %d %d\n",layer_width[dl],layer_pos[dl]+row,col);
	/*
	 * Minor units
	 */
	for (sl=0; sl<net->n_layers; sl++) {
	  if (!slayer[sl]) continue;
	  fprintf(dout,"width %d\n",layer_width[sl]);
	  for (su=sunit_range[sl].lo; su<=sunit_range[sl].hi; su++) {
	    if (RangePosition(su,&(sunit_range[sl]),LayerFormat(sl),&row,&col)) {
	      if (incoming)
		weight = GetLinkField(net,sl,su,dl,du,field_name);
	      else
		weight = 0.0;
	      if (outgoing && weight==0.0)
		weight = GetLinkField(net,dl,du,sl,su,field_name);
	      if (dl!=sl || du != su || weight!=0.0)
		fprintf(dout,"b %d %d %.*f\n",layer_pos[sl]+row,col,
			weight_precision,weight/scale);
	    }
	  } /* su */
	} /* sl */
	if (space) fprintf(dout,"newrow\nrowhint 1\n");
      } /* du */
    } /* dl */
  } /* weights */
  fprintf(dout,"newrow\ns -1 0 b\nend\n");
  return;
}

