#include <itf/option.h>
#include <itf/itf.h>
#include <sim/sim.h>
#include <lib/example.h>
#include <lib/ex-load.h>
#include "prop-bp.h"
#include "example-bp.h"

#define DB(x) x

struct EXAMPLE_LIST *current_el = NULL;
struct EXAMPLE_REC *current_er = NULL;
struct EXAMPLE_LIST **ela = NULL;
struct EXAMPLE example_rec;
struct EXAMPLE *example = &example_rec;
int n_els = 0;
int example_show_end = 1;
string example_format;
string example_interp;
string example_type;
string ex_validation_range;
string ex_training_range;
string ex_testing_range;
string ex_ignore_range;
string ex_class_override;

static PFEC DoAtResetExamples;
static PFEC DoBeforeNextExample;
static PFEC DoAfterNextExample;

#define NO_FOLLOW	0
#define FOLLOW		1
#define AUTO_END	2

#define FollowAction(PRE, POST) action_follow[(int)PRE][(int)POST]

IUsage("[-c <comment>] [-t <time>] [label] [: val1 val2 ...]");

unsigned char action_follow[EXAMPLE_ACTION_TOP][EXAMPLE_ACTION_TOP];

static char *action_name[EXAMPLE_ACTION_TOP] = {
  "(bad code)",
  "null action",
  "input",
  "output",
  "layer",
  "predict",
  "copy",
  "clear",
  "fill",
  "pop",
  "push",
  "clearstack",
  "enqueue",
  "dequeue",
  "clearqueue",
  "ecall",
  "edo",
  "end",
};

EXAMPLE_REC *NewExampleRec();
EXAMPLE_LIST *NewExampleList();
EXAMPLE_CODE DoERList();

autoinit_example()
{
  int i, j;
  for (i=(int)EXAMPLE_ACTION_BOT;i<(int)EXAMPLE_ACTION_TOP;i++)
    for (j=(int)EXAMPLE_ACTION_BOT;j<(int)EXAMPLE_ACTION_TOP;j++)
      action_follow[i][j] = NO_FOLLOW;
  FollowAction(             INPUT_EX,             INPUT_EX) = AUTO_END;
  FollowAction(             INPUT_EX,             LAYER_EX) = AUTO_END;
  FollowAction(             INPUT_EX,            OUTPUT_EX) = FOLLOW;
  FollowAction(             INPUT_EX,               END_EX) = FOLLOW;

  FollowAction(             LAYER_EX,             INPUT_EX) = FOLLOW;
  FollowAction(             LAYER_EX,             LAYER_EX) = FOLLOW;
  FollowAction(             LAYER_EX,            OUTPUT_EX) = AUTO_END;
  FollowAction(             LAYER_EX,               END_EX) = FOLLOW;

  FollowAction(            OUTPUT_EX,             INPUT_EX) = AUTO_END;
  FollowAction(            OUTPUT_EX,             LAYER_EX) = FOLLOW;
  FollowAction(            OUTPUT_EX,            OUTPUT_EX) = AUTO_END;
  FollowAction(            OUTPUT_EX,           PREDICT_EX) = AUTO_END;
  FollowAction(            OUTPUT_EX,               END_EX) = FOLLOW;

  FollowAction(               END_EX,             INPUT_EX) = FOLLOW;
  FollowAction(               END_EX,             LAYER_EX) = FOLLOW;
  FollowAction(               END_EX,            OUTPUT_EX) = FOLLOW;
  FollowAction(               END_EX,           PREDICT_EX) = FOLLOW;

  FollowAction(           PREDICT_EX,           PREDICT_EX) = AUTO_END;
  FollowAction(           PREDICT_EX,            OUTPUT_EX) = AUTO_END;
  FollowAction(           PREDICT_EX,               END_EX) = FOLLOW;

  FollowAction(       NULL_ACTION,           PREDICT_EX) = FOLLOW;
  FollowAction(       NULL_ACTION,             INPUT_EX) = FOLLOW;
  FollowAction(       NULL_ACTION,            OUTPUT_EX) = FOLLOW;
  ISetValue("example-format", "%3g");
  ISetValue("example-type", "el");
  ISetValue("example-interp", "vector");
  DoAtResetExamples = (PFEC) IGetFunction("DoAtResetExamples");
  DoBeforeNextExample = (PFEC) IGetFunction("DoBeforeNextExample");
  DoAfterNextExample = (PFEC) IGetFunction("DoAfterNextExample");
}


/*************************************************************************
 * Functions to modify input
 */

real GaussNoise(sd)
real sd;
{
  static int iset=FALSE;
  static real gset;
  real v1,v2,r,fac;
  if (iset) {
    iset = FALSE;
    return (gset*sd);
  } else {
    do {
      v1 = Noise(-1.0,1.0);
      v2 = Noise(-1.0,1.0);
      r  = v1*v1 + v2*v2;
    } while (r >= 1.0);
    fac = sqrt(-2.0*log(r)/r);
    gset = v1*fac;
    iset = TRUE;
    return (v2*fac*sd);
  }
}

real
  UnitNoise(y,sd)
real y,sd;
{
  if      (sd==0.0)           return 0.0;
  else if (y==UnitLowerBound) return fabs(GaussNoise(sd));
  else if (y==UnitUpperBound) return -fabs(GaussNoise(sd));
  else                        return GaussNoise(sd);
}

void
  AddGaussianInputNoise(net,sd)
NETWORK *net;
real sd;
{
  ApplyToEachInputUnit(UNIT_OUTPUT(unit) += UnitNoise(UNIT_OUTPUT(unit),sd););
}

void SaveOutputInInput(net)
NETWORK *net;
{
  ApplyToEachInputUnit(unit->x = UNIT_OUTPUT(unit););	/* save noiseless values */
}

int RestoreInputToOutput(net)
NETWORK *net;
{
  ApplyToEachInputUnit(UNIT_OUTPUT(unit) = unit->x;);/* restore noiseless values */
  return 1;
}

real
  input_cos_angle(net,values)
NETWORK *net;
real *values;
{
  int u = 0;
  real l1 = 0.0;
  real l2 = 0.0;
  real dst = 0.0;
  ApplyToEachInputUnit
    (
     l1  += sqr(UNIT_OUTPUT(unit));
     l2  += sqr(values[u]);
     dst += UNIT_OUTPUT(unit)*values[u++];
     );
  return (dst/sqrt(l1*l2));
}

real
  output_cos_angle(net)
NETWORK *net;
{
  real l1 = 0.0;
  real l2 = 0.0;
  real dst = 0.0;
  ApplyToEachOutputUnit
    (
     l1  += sqr(UNIT_OUTPUT(unit));
     l2  += sqr(UNIT_TARGET(unit));
     dst += UNIT_OUTPUT(unit)*UNIT_TARGET(unit);
     );
  if (l1==0.0 || l2==0.0) return 0.0;
  return (dst/sqrt(l1*l2));
}

int OtherExampleCloser(net)
NETWORK *net;
{
  int this_ex_n   = net->p->example_n;
  real this_ex_angle = input_cos_angle(net,current_el->index[this_ex_n]->value);
  int ex;
  real tmp;
  for (ex=0; ex<current_el->n_examples; ex++) 
    if (   (ex != this_ex_n)
	&& (input_cos_angle(net,current_el->index[ex]->value) > this_ex_angle))
      return 1;
  return 0;
}


void
  AddInputNoise(net)
NETWORK *net;
{
  real sd = net->p->input_noise;
  if (!zerop(sd)) {
    if (np->ensure_ex_closest) {
      SaveOutputInInput(net);
      do {
	AddGaussianInputNoise(net,sd);
      } while (OtherExampleCloser(net) && RestoreInputToOutput(net));
    } else
      AddGaussianInputNoise(net,sd);
  }
}

void
  ComputeExpectedValues(net)
NETWORK *net;
{
  NET_PARAMS *np = net->p;
  int e,u;
  real input_sum = 0;
  real output_sum = 0;
  int old_example_n = np->example_n;
  int old_modify_input = np->modify_input;
  real old_input_noise = np->input_noise;
  if (current_el == NULL) {
    IError("current example set is NULL");
    return;
  }
  np->modify_input = 0;
  np->input_noise = 0.0;
  np->input_expected_value  = (real*)calloc(np->input_width+1,sizeof(real));
  np->output_expected_value = (real*)calloc(np->output_width+1,sizeof(real));
  for (u=0; u<np->input_width;  u++) np->input_expected_value[u]  = 0.0;
  for (u=0; u<np->output_width; u++) np->output_expected_value[u] = 0.0;
  ResetExamples(net);
  while (NextExample(net) == EXAMPLE_OK) {
    u = 1;  
    ApplyToEachInputUnit
      (
       input_sum += UNIT_OUTPUT(unit);
       np->input_expected_value[u++] += UNIT_OUTPUT(unit);
       );
    u = 1;
    ApplyToEachOutputUnit
      (
       output_sum += UNIT_TARGET(unit);
       np->output_expected_value[u++] += UNIT_TARGET(unit);
       );
  }
  /* slot 0 contains overall average expected value */
  np->input_expected_value[0] =  input_sum  / (np->input_width * current_el->n_examples);
  np->output_expected_value[0] = output_sum / (np->output_width * current_el->n_examples);
  for (u=1; u<=np->input_width;  u++) np->input_expected_value[u]  /= current_el->n_examples;
  for (u=1; u<=np->output_width; u++) np->output_expected_value[u] /= current_el->n_examples;
  DoExample(net,old_example_n);
  np->modify_input = old_modify_input;
  np->input_noise  = old_input_noise;
  np->computed_expected_values = 1;
}
  
void
  SampleInputFeatures(net)
NETWORK *net;
{
  NET_PARAMS *np = net->p;
  int u;
  real e,sample_p,p = np->input_feature_prob;

  if (!np->computed_expected_values) ComputeExpectedValues(net);
  e        = np->input_expected_value[0];
  sample_p = (1.0-p) * e/(1-e);
  u = 1;
  ApplyToEachInputUnit
    (
     if (onep(UNIT_OUTPUT(unit)) && with_prob(p)) u++;
     else if (with_prob(np->input_expected_value[u++]*sample_p))
       UNIT_OUTPUT(unit) = UnitLowerBound;
     );
}

void
  TurnOffOneInputFeature(net)
NETWORK *net;
{
  NET_PARAMS *np = net->p;
  int i;
  DoLayer(np->input_layer_n) {
    do {
      i = RandInt(np->input_width);
      DoUnit(UNIT,i) {
	if (onep(UNIT_OUTPUT(unit))) {
	  UNIT_OUTPUT(unit) = UnitLowerBound;
	  return;
	}
      } EndDo;
    } while (1);
  } EndDo;
}

void
  ModifyInput(net)
NETWORK *net;
{
  switch (net->p->modify_input) {
  case 0:
    break;
  case 1:
    SampleInputFeatures(net);
    break;
  case 2:
    TurnOffOneInputFeature(net);
    break;
  default:
    IError("unknown modify_input function specified: %d",net->p->modify_input);
    return;
  }
#ifndef DBM_NETWORK		/* done in SetInitialStates() in prop-dbm.c */
  AddInputNoise(net);
#endif  DBM_NETWORK
}


/****************************************************************************
 * 5 functions are defined (and are used by the example module):
 *
 * LoadInput(net,i,x,t)		- load input unit i with value x
 * LoadInputVector(net,v,n,x,t)	- load input layer with values in v or x
 * LoadTarget(net,i,x,t)	- load target unit i with value x
 * LoadTargetVector(net,v,n,x,t)- load target layer with values in v or x
 * LoadUnit(net,i,j,x,t)	- load unit j in layer i with value x
 *
 * The vector loading functions will be used by the example module
 * if the macro VECTOR_EX_LOAD is defined at the beginning of ul.h.
 *
 * All these functions also have a 't' parameter which is used when
 * the network is a TIME_NETWORK, and there is a different input and
 * target for each time frame for each unit.  The 't' parameter is
 * ignored unless the macro TIME_NETWORK is defined (by putting
 * "#define TIME_NETWORK" at the beginning of ul.h).
 *
 * Currently both unit functions (in prop-bp.h) return values in [0,1].
 * If this is changed (e.g. to [-1,1]) then the example loading functions
 *    below must be changed to accomdate this.
 */

LoadInput(net,i,x,t)
NETWORK *net;
int i;
real x;
int t;
{
  int it;
  DoLayer(net->p->input_layer_n) {
    DoUnit(UNIT,i) {
      UNIT_OUTPUT(unit) = ScaleOutput(x);
#ifdef TIME_NETWORK
      if (t==T_SENTINEL) t=0;
      if (t>=net->p->max_t) net->p->max_t = t+1;
      for (it=0; it<=net->p->max_t; it++)
	UNIT_INPUT_T(unit,it) = UNIT_OUTPUT(unit);
#endif
    } EndDo;
  } EndDo;
}

LoadInputVector(net,v,n,x,t)
NETWORK *net;
real v[];
real x;
int t;
{
  int it, u = 0;
  if (v!=NULL) {
    ApplyToEachInputUnit
      (
       if (u>=n) break;
       UNIT_OUTPUT(unit) = ScaleOutput(v[u++]);
       );
  } else {
    ApplyToEachInputUnit
      (
       if (unit_n>=n) break;
       UNIT_OUTPUT(unit) = ScaleOutput(x);
       );
  }
#ifdef DBM_NETWORK
  ApplyToEachInputUnit(UNIT_TARGET(unit) = UNIT_OUTPUT(unit););
#endif DBM_NETWORK
  ModifyInput(net);
#ifdef TIME_NETWORK
  if (t==T_SENTINEL) t=0;
  if (t>=net->p->max_t) net->p->max_t = t+1;
  ApplyToEachInputUnit
    (
     if (unit_n>=n) break;
     for (it=0; it<=net->p->max_t; it++)
       UNIT_INPUT_T(unit,it) =
         (IsInputIteration(it+1) ? UNIT_OUTPUT(unit) : layer->p->initial_y);
     );
#endif
}

LoadTarget(net,i,x,t)
NETWORK *net;
int i;
real x;
int t;
{
  DoLayer(net->p->output_layer_n) {
    DoUnit(UNIT,i) {
      UNIT_TARGET(unit) = ScaleOutput(x);
#ifdef TIME_NETWORK
      if (t==T_SENTINEL) t=0;
      if (t>=net->p->max_t) net->p->max_t = t+1;
      UNIT_TARGET_T(unit,t) = UNIT_TARGET(unit);
#endif
    } EndDo;
  } EndDo;
}

LoadTargetVector(net,v,n,x,t)
NETWORK *net;
real v[];
real x;
int t;
{
  int u = 0;
  if (v!=NULL) {
    ApplyToEachOutputUnit
      (
       if (u>=n) break;
       UNIT_TARGET(unit) = ScaleOutput(v[u++]);
        );
  } else {
    ApplyToEachOutputUnit
      (
       if (unit_n>=n) break;
       UNIT_TARGET(unit) = ScaleOutput(x);
       );
  }
#ifdef TIME_NETWORK
  if (t==T_SENTINEL) t=0;
  if (t>=net->p->max_t) net->p->max_t = t+1;
  ApplyToEachOutputUnit
    (
     if (unit_n>=n) break;
     UNIT_TARGET_T(unit,t) = UNIT_TARGET(unit);
     );
#endif
}

LoadUnit(net,i,j,x,t)
NETWORK *net;
int i,j;
real x;
int t;
{
  int it;
  if (i < 0 || i >= net->n_layers) return 0;
  DoLayer(i) {
    if (j < 0 || j >= layer->n_units) return 0;
    DoUnit(UNIT,j) {
      UNIT_OUTPUT(unit) = ScaleOutput(x);
#ifdef TIME_NETWORK
      if (t==T_SENTINEL) t=0;
      for (it=0; it<=net->p->max_t; it++)
	UNIT_OUTPUT_T(unit,it) = UNIT_OUTPUT(unit);
#endif
    } EndDo;
  } EndDo;
  return 1;
}


/******************************************************************************
 * We now return you to your regularly scheduled example-bp.c
 */

EXAMPLE_CODE ResetExamples(net)
NETWORK *net;
{
  PFEC func;
  char name[60];
  EXAMPLE_CODE retval;
  char *str;
  int net_n;
#ifdef EX_LOAD_ALL_NETS
  assert(net==NULL);
  for (net_n=0, net=neta[0];net_n<n_nets && net!=NULL;net_n++, net=neta[net_n])
    if (net->p!=NULL)
      net->p->example_n = -1;
  net = NULL;
#else
  if (!NetExists(net)) {
    IError("network is not built");
    return EXAMPLE_ERROR;
  }
  net->p->example_n = -1;
#endif
  str = IGetValue("example_type", NULL);
  if (str == NULL) return EXAMPLE_ERROR;
  strcpy(name, str);
  strcat(name, "_reset");
  func = (PFEC) IGetFunction(name);
  if (func==NULL) {
    IError("there is no function with name \"function-%s\"", name);
    return EXAMPLE_ERROR;
  }
  retval = func(net);
  if (retval==EXAMPLE_OK && DoAtResetExamples!=NULL)
    DoAtResetExamples(net);
  return retval;
}

EXAMPLE_CODE GetExample(net, action, n, load, example, train, validate, test)
NETWORK *net;
char *action;
int n, load;
struct EXAMPLE *example;
int train, validate, test;	/* example classes we want to load */
{
  PFER func;
  char name[60];
  char *str;
  int net_n;
  int testing, training, validation, ignore;	/* class of the example */
  EXAMPLE_CODE retval = EXAMPLE_OK;
  EXAMPLE_REC *er;
  str = IGetValue("example_type", NULL);
  if (str == NULL) return EXAMPLE_ERROR;
  strcpy(name, str);
  strcat(name, action);
  func = (PFER) IGetFunction(name);
  if (func==NULL) {
    IError("there is no function with name \"function_%s\"", name);
    return EXAMPLE_ERROR;
  }
  if (example!=NULL) {
    example->predict = 0;
    example->output = 0;
    example->input = 0;
    example->training = 0;
    example->validation = 0;
    example->testing = 0;
    example->props_in_text = "";
  }
  do {
    er = func(net, n);
    if (er==NULL)
      return EXAMPLE_END;
    if (true(ex_class_override)) {
      /* CompareWithRangeString returns -1, 0, 1, or 2, 0 means in range */
      training = CompareWithRangeString(er->my_n, ex_training_range, NULL);
      validation = CompareWithRangeString(er->my_n, ex_validation_range, NULL);
      testing = CompareWithRangeString(er->my_n, ex_testing_range, NULL);
      ignore = CompareWithRangeString(er->my_n, ex_ignore_range, NULL);
      if (   (ignore==0 || ignore==1)
	  && (training>0) && (validation>0) && (testing>0))
	return EXAMPLE_END; /* no more will be in a not-ignore range */
      if (training==0 || validation==0 || testing==0 || ignore==0) {
	/* n is in one of the ranges */
	training = !training;
	validation = !validation;
	testing = !testing;
	ignore = !ignore;
      } else {
	/* n is not in any range - either ignore or treat as training */
	if (training==2 || validation==2 || testing==2 || ignore==2) {
	  training = 1;
	  validation = 0;
	  testing = 0;
	  ignore = 0;
	} else {
	  training = 0;
	  validation = 0;
	  testing = 0;
	  ignore = 1;
	}
      }
    } else {
      training = er->training;
      validation = er->validation;
      testing = er->testing;
      ignore = 0;
      if (!training && !testing && !validation)
	training = 1;
    }
  } while (   !(   training && train
		|| validation && validate
		|| testing && test)
	   || ignore);
  if (example!=NULL) {
    example->training = training;
    example->validation = validation;
    example->testing = testing;
    example->position_loading = 0;
    example->output_er = NULL;
    example->input_er = NULL;
    example->n = er->my_n;
    if (training && validation && testing)
      example->props_in_text = "training+validation+testing";
    else if (training && validation && !testing)
      example->props_in_text = "training+validation";
    else if (training && !validation && testing)
      example->props_in_text = "training+testing";
    else if (training && !validation && !testing)
      example->props_in_text = "training";
    else if (!training && validation && testing)
      example->props_in_text = "validation+testing";
    else if (!training && validation && !testing)
      example->props_in_text = "validation";
    else if (!training && !validation && testing)
      example->props_in_text = "testing";
  }
#ifdef EX_LOAD_ALL_NETS
  assert(net==NULL);
  for (net_n=0, net=neta[0];net_n<n_nets && net!=NULL;net_n++, net=neta[net_n])
    if (net->p!=NULL)
      net->p->example_n = er->my_n;
  net = NULL;
#else
  if (!NetExists(net)) {
    IError("network is not built");
    return EXAMPLE_ERROR;
  }
  net->p->example_n = er->my_n;
#endif
  if (load) {
    if (DoBeforeNextExample!=NULL)
      retval = DoBeforeNextExample(net);
    if (retval==EXAMPLE_OK)
      retval = DoERList(net, er, example);
    if (retval==EXAMPLE_OK && DoAfterNextExample!=NULL)
      retval = DoAfterNextExample(net);
  }
  if (retval!=EXAMPLE_OK)
    IReportError(stderr);
  return retval;
}

EXAMPLE_CODE NextExample(net)
NETWORK *net;
{
  return GetExample(net, "_next", 0, 1, example, 1, 1, 1);
}

EXAMPLE_CODE DoExample(net, n)
NETWORK *net;
int n;
{
  return GetExample(net, "_do", n, 1, example, 1, 1, 1);
}

EXAMPLE_CODE SkipExample(net)
NETWORK *net;
{
  return GetExample(net, "_next", 0, 0, example, 1, 1, 1);
}

EXAMPLE_CODE LoadNextExample(net, example, train, validate, test)
NETWORK *net;
struct EXAMPLE *example;
int train, validate, test;
{
  return GetExample(net, "_next", 0, 1, example, train, validate, test);
}

EXAMPLE_CODE DoExampleN(net, example, n)
NETWORK *net;
struct EXAMPLE *example;
int n;
{
  return GetExample(net, "_do", n, 1, example, 1, 1, 1);
}

EXAMPLE_CODE SkipNextExample(net, example)
NETWORK *net;
struct EXAMPLE *example;
{
  return GetExample(net, "_next", 0, 0, example, 1, 1, 1);
}

EXAMPLE_CODE ResetEL(el)
EXAMPLE_LIST *el;
{
  if (el==NULL) {
    warn("ResetExamples", "example list is NULL");
    return EXAMPLE_ERROR;
  } else {
    if (!el->complete || el->tail->action!=END_EX) {
      if (   !TryToAdd(NewExampleRec(END_EX, 0, NULL))) {
	IError("example list is incomplete and END cannot be added");
	return EXAMPLE_ERROR;
      }
    }
    if (el->n_examples==0) {
      IError("example list has no example!");
      return EXAMPLE_ERROR;
    }
    el->current = NULL;
    el->next = el->head;
    el->example_n = -1;	/* ready to start at zero */
  }
  return EXAMPLE_OK;
}

EXAMPLE_CODE function_el_reset()
{
  return ResetEL(current_el);
}

EXAMPLE_CODE DoERList(net, er, example)
NETWORK *net;
EXAMPLE_REC *er;
EXAMPLE *example;
{
  if (er==NULL) return EXAMPLE_END;
  /* now advance the current pointer past */
  while (er!=NULL && er->action!=END_EX) {
    if (!DoExampleAction(net, er, example))
      return EXAMPLE_ERROR;
    er = er->next;
  }
  return EXAMPLE_OK;
}

EXAMPLE_REC *function_el_next()
{
  EXAMPLE_LIST *el = current_el;
  EXAMPLE_REC *er = NULL;

  do {
    if (el==NULL || !el->complete)
      return NULL;
    er = el->current = current_er = el->next;
    if (er==NULL)
      return er;
    el->example_n = er->my_n;
    
    /* now advance the current pointer past the end of this section */
    while (el->next!=NULL && el->next->action!=END_EX)
      el->next = el->next->next;
    if (el->next!=NULL)
      el->next = el->next->next;
  } while (er->ignore);
  return er;
}

static void MakeELIndex(el)
EXAMPLE_LIST *el;
{
  EXAMPLE_ACTION last_action;
  EXAMPLE_REC *er;
  int n;
  if (el==NULL) IErrorAbort("current example list is NULL");
  if (!el->complete && ResetEL(el)!=EXAMPLE_OK) IAbort();  /* bad */
  if (el->index_size==el->n_examples && el->index!=NULL) return;  /* ok */
  if (el->n_examples<=0) return;
  if (el->index) free(el->index);
  el->index = Calloc(EXAMPLE_REC*, el->n_examples);
  if (el->index==NULL) {
    IOutOfMemory("MakeELIndex", "allocating index (%d)", el->n_examples);
    IAbort();
  }
  el->index_size = el->n_examples;
  last_action = END_EX;
  n = 0;
  for (er=el->head; er!=NULL; er=er->next) {
    if (last_action==END_EX)
      el->index[n++] = er;
    if ((n-1)!=er->my_n)
      warn("MakeELIndex", "example %d improperly numbered", n);
    last_action = er->action;
  }
}

EXAMPLE_REC *function_el_do(net, n)
NETWORK *net;
int n;
{
  EXAMPLE_REC *er = NULL;
  if (current_el->index_size!=current_el->n_examples||current_el->index==NULL)
    MakeELIndex(current_el);

  if (n<0 || n>=current_el->n_examples) {
    IError("example # %d is not in the current example list", n);
    return NULL;
  }

  er = current_el->index[n];
  current_el->example_n = er->my_n;
  return er;
}

DoExampleAction(net, er, example)
NETWORK *net;
EXAMPLE_REC *er;
struct EXAMPLE *example;
{
  int i, k, size;
  real *v = NULL;
  int net_n;
  int output_width;
  int input_width;
#ifdef EX_LOAD_ALL_NETS
  assert(net==NULL);
  output_width = neta[0]->p->output_width;
  input_width = neta[0]->p->input_width;
  for (net_n=0, net=neta[0];net_n<n_nets && net!=NULL;net_n++, net=neta[net_n])
    if (output_width!=net->p->output_width||input_width!=net->p->input_width) {
      IError("input or output width of net \"%s\" does not match net \"%s\"",
	     net->name, neta[0]->name);
      return 0;
    }
  net = NULL;
#else
  output_width = net->p->output_width;
  input_width = net->p->input_width;
#endif
  if (er->action==END_EX)
    return 1;
  if (example!=NULL) {
    switch(er->action) {
    case PREDICT_EX:
    case OUTPUT_EX:
      example->output_er = er;
      break;
    case INPUT_EX:
      example->input_er = er;
      break;
    }
  }
  /***
   * find the vector
   */
  size = er->n_values;
  if (er->value != NULL) {
    v = er->value;
  } else if (er->position==NULL) {
    IError("no vector or label for example %d", er->my_n);
    return 0;
  }
#ifdef EXAMPLE_LABELS
  if (er->str!=NULL) {
#ifdef EX_LOAD_ALL_NETS
    for (net_n=0, net=neta[0];net_n<n_nets && net!=NULL;net_n++, net=neta[net_n])
#endif EX_LOAD_ALL_NETS
      net->p->example_label = er->str;
  }
#endif EXAMPLE_LABELS
  if (size<1) {
    IError("size of vector for example %d is less than 1!", er->my_n);
    return 0;
  }

  if (er->position==NULL) {
    if (example!=NULL)
      example->position_loading = 0;
    switch (er->action) {
    case PREDICT_EX:
      /* and fall through to OUTPUT_EX */
      if (example!=NULL)
	example->predict = 1;
    case OUTPUT_EX:
      if (er->action==OUTPUT_EX && example!=NULL)
	example->output=1;
      if (output_width != size) {
	warn("DoExampleAction", "mismatch between output width (%d) and size of example (%d)",
	     output_width, size, er->my_n);
	return 0;
      }
#ifdef EX_LOAD_ALL_NETS
      for (net_n=0, net=neta[0];net_n<n_nets && net!=NULL;net_n++, net=neta[net_n])
#endif
#ifdef VECTOREXLOAD
	LoadTargetVector(net, v, size, 0.0, er->t);
#else
        for (i=0;i<size;i++) LoadTarget(net, i, v[i], er->t);
#endif
      break;
    case INPUT_EX:
      if (example!=NULL) example->input = 1;
      if (input_width != size) {
	IError("mismatch between input width (%d) and size of example (%d), on example %d",
		      input_width, size, er->my_n);
	return 0;
      }
#ifdef EX_LOAD_ALL_NETS
      for (net_n=0, net=neta[0];net_n<n_nets && net!=NULL;net_n++, net=neta[net_n])
#endif
#ifdef VECTOREXLOAD
      LoadInputVector(net, v, size, 0.0, er->t);
#else
      for (i=0;i<size;i++) LoadInput(net, i, v[i], er->t);
#endif
      break;
    default:
      panic("DoExampleAction", "bad example action", er->action);
    }
  } else {
    if (example!=NULL)
      example->position_loading = 1;
    switch (er->action) {
    case INPUT_EX:
      if (example!=NULL)
	example->input = 1;
#ifdef EX_LOAD_ALL_NETS
      for (net_n=0, net=neta[0];net_n<n_nets && net!=NULL;net_n++, net=neta[net_n])
#endif
#ifdef VECTOREXLOAD
      LoadInputVector(net, NULL, input_width, 0.0, er->t);
#else
      for (i=0;i<input_width;i++)
	LoadInput(net, i, 0.0, er->t);
#endif
      for (i=0; i<er->n_values; i++) {
	k = er->position[i];
	if (k < 0 || k >= input_width) {
	  IError("mismatch between input width (%d) and position of example (%d), on example %d",
		 input_width, k, er->my_n);
	  return 0;
	}
#ifdef EX_LOAD_ALL_NETS
	for (net_n=0, net=neta[0];net_n<n_nets && net!=NULL;net_n++, net=neta[net_n])
#endif
	  LoadInput(net, k, er->value==NULL ? 1.0 : er->value[i], er->t);
      }
      break;
    case PREDICT_EX:
      /* and fall through to OUTPUT_EX */
      if (example!=NULL) example->predict = 1;
    case OUTPUT_EX:
      if (example!=NULL) example->output = 1;
#ifdef EX_LOAD_ALL_NETS
      for (net_n=0, net=neta[0];net_n<n_nets && net!=NULL;net_n++, net=neta[net_n])
#endif
#ifdef VECTOREXLOAD
      LoadTargetVector(net, NULL, output_width, 0.0, er->t);
#else
      for (i=0;i<output_width;i++)
	LoadTarget(net, i, 0.0, er->t);
#endif
      for (i=0; i<er->n_values; i++) {
	k = er->position[i];
#ifdef EX_LOAD_ALL_NETS
      for (net_n=0, net=neta[0];net_n<n_nets && net!=NULL;net_n++, net=neta[net_n])
	if (k < 0 || k >= output_width) {
	  IErrorAbort("DoExampleAction mismatch between output width (%d) and position of example (%d), example # %d",
		      output_width, k, er->my_n);
	}
#endif
	LoadTarget(net, k, er->value==NULL ? 1.0 : er->value[i], er->t);
      }
      break;
    default:
      panic("DoExampleAction", "bad example action", er->action);
    }
  }
  return 1;
}

command_show_el(tokc, tokv)
int tokc;
char *tokv[];
{
  EXAMPLE_REC *er;
  if (tokc==SHORT_HELP || tokc==LONG_HELP) {
    help_string = "print the current list of examples";
    return 0;
  }
  if (current_el==NULL) {
    fprintf(dout, "The current example list is completely empty\n");
  } else {
    fprintf(dout, "# Example list #%d contains %d examples:%s\n",
	    current_el->my_n, current_el->n_examples,
	    (current_el->complete ? "" : " (missing a final end)"));
    for (er=current_el->head;er!=NULL;er=er->next)
      (void) ShowExample(er);
    fprintf(dout, "# End examples\n");
  }
  return 0;
}

int ShowExample(er)
EXAMPLE_REC *er;
{
  assert(er!=NULL);
  switch(er->action) {
  case INPUT_EX:
  case OUTPUT_EX:
  case PREDICT_EX:
    ShowParameterizedAction(action_name[(int) er->action], er);
    ShowErValues(er);
    break;
  case END_EX:
    if (!example_show_end)
      break;
    ShowParameterizedAction(action_name[(int) er->action], er);
    ShowErValues(er);
    break;
  default:
    fprintf(dout, "Unimplemented example action %d\n", er->action);
  }
}

ShowParameterizedAction(name, er)
char *name;
EXAMPLE_REC *er;
{
  fprintf(dout, "%-10s", name);
  if (er->t!=T_SENTINEL)
    fprintf(dout, " -t %d", er->t);
  if (er->comment!=NULL)
    fprintf(dout, " -c \"%s\"", er->comment);
  if (er->str!=NULL)
    fprintf(dout, " \"%s\"", er->str);
  if (er->action!=END_EX)
    fprintf(dout, " :");
}

ShowErValues(er)
EXAMPLE_REC *er;
{
  int i;
  string str;
  assert(er!=NULL);
  for (i=0;i<er->n_values;i++) {
    if (isnan(er->value[i]))
      str = "-";
    else
      str = IPrintRealFormatted(er->value[i], example_format);
    if (i>0) fputc(' ', dout);
    if (str!=NULL)
      fputs(str, dout);
  }
  fputc('\n', dout);
}

int command_end(tokc, tokv)
int tokc;
char *tokv[];
{
  EXAMPLE_REC *er;
  if (tokc==SHORT_HELP || tokc==LONG_HELP) {
    help_string = "end a training example";
    return;
  }
  if (tokc>1) {
    IError("%s takes no arguments", tokv[0]);
    return;
  }
  /* TODO should check on current_el */
  assert(current_el!=NULL);
  if (!(er = NewExampleRec(END_EX, 0, NULL))) return;
  TryToAdd(er);
}

int AddComment(er,comment)
EXAMPLE_REC *er;
char *comment;
{
  er->comment = malloc(strlen(comment)+1);
  if (er->comment==NULL) {
    IOutOfMemory("comment");
    return 0;
  }
  strcpy(er->comment,comment);
  return 1;
} 

int command_input(tokc, tokv)
int tokc;
char *tokv[];
{
  int n;
  string str;
  EXAMPLE_REC *er;
  if (GiveHelp(tokc)) {
    ISynopsis("put an input record on the example list");
    IHelp(IHelpArgs,
	  NULL);
    return;
  }

  if (!ExampleDataOk(tokv[0], tokv+1, usage, &n, &str)) return;
  if (!(er = NewExampleRec(INPUT_EX, n, str))) return;
  if (!ReadExampleData(tokv, er)) return;
  TryToAdd(er);
}

int command_input_n(tokc,tokv)
int tokc;
char *tokv[];
{
  int n;
  int c = 0;
  EXAMPLE_REC *er;
  char *comment = NULL;
  static struct OPTION options[] =
    {
      {OPTPRES|OPTSTRARG,"c",	0},
      {OPTLAST,NULL,0}
    };
  IUsage("[%pc <comment>] <n-vals>");
  if (GiveHelp(tokc)) {
    ISynopsis("put an input record on the example list");
    IHelp(IHelpArgs,
	  "Reads <n-vals> values from subsequent lines",
	  NULL);
    return;
  }
  tokv++;	/* skip command name */
  while (IsOption(tokv[0])) {
    if (OptionMatch(tokv[0],"c",options,&c)) {
      comment = *++tokv; tokc--;
    } else {
      IError("option not understood: \"%s\"",tokv[0]);
      return;
    }
    tokv++;
    tokc--;
  }
  if (!IIsInteger(tokv[0])) {
    IError(IPrintUsage("input_n",usage));
    return;
  }
  n = atoi(tokv[0]);
  if (!(er = NewExampleRec(INPUT_EX, n, NULL))) return;
  if (c && !AddComment(er, comment)) return;
  if (!ReadFollowingExampleData(er)) return;
  TryToAdd(er);
}


int command_output_n(tokc,tokv)
int tokc;
char *tokv[];
{
  int n;
  int c = 0;
  EXAMPLE_REC *er;
  char *comment = NULL;
  static struct OPTION options[] =
    {
      {OPTPRES|OPTSTRARG,"c",	0},
      {OPTLAST,NULL,0}
    };
  IUsage("[%pc <comment>] <n-vals>");
  if (GiveHelp(tokc)) {
    ISynopsis("put an output record on the example list");
    IHelp(IHelpArgs,
	  "Reads <n-vals> values from subsequent lines",
	  NULL);
    return;
  }
  tokv++;
  while (IsOption(tokv[0])) {
    if (OptionMatch(tokv[0],"c",options,&c)) {
      comment = *++tokv; tokc--;
    } else {
      IError("option not understood: \"%s\"",tokv[0]);
      return;
    }
    tokv++;
    tokc--;
  }
  if (!IIsInteger(tokv[0])) {
    IError(IPrintUsage("input_n",usage));
    return;
  }
  n = atoi(tokv[0]);
  if (!(er = NewExampleRec(OUTPUT_EX,n,NULL))) return;
  if (c && !AddComment(er,comment)) return;
  if (!ReadFollowingExampleData(er)) return;
  TryToAdd(er);
}


int command_output(tokc, tokv)
int tokc;
char *tokv[];
{
  int n;
  string str;
  EXAMPLE_REC *er;
  if (GiveHelp(tokc)) {
    ISynopsis("put an output record on the example list");
    IHelp(IHelpArgs,
	  NULL);
    return;
  }

  if (!ExampleDataOk(tokv[0], tokv+1, usage, &n, &str)) return;
  if (!(er = NewExampleRec(OUTPUT_EX, n, str))) return;
  if (!ReadExampleData(tokv, er)) return;
  TryToAdd(er);
}

int command_predict(tokc, tokv)
int tokc;
char *tokv[];
{
  int n;
  string str;
  EXAMPLE_REC *er;
  if (GiveHelp(tokc)) {
    ISynopsis("put an predict record on the example list");
    IHelp(IHelpArgs,
	  NULL);
    return;
  }
  
  if (!ExampleDataOk(tokv[0], tokv+1, usage, &n, &str)) return;
  if (!(er = NewExampleRec(PREDICT_EX, n, str))) return;
  if (!ReadExampleData(tokv, er)) return;
  TryToAdd(er);
}

int ExampleDataOk(command, tokv, usage, n, str)
char *command;
char *tokv[];
char *usage;
int *n;
string *str;
{
  int i = 0;
  int er_type = 0;
  *n = 0;
  *str = 0;
  while (tokv[i] && strcmp(tokv[i], ":")) {
    if (!strcmp("-c", tokv[i])) { 				/* a comment */
      if (tokv[++i]!=NULL) i++;
    } else if (tokv[i][0]!='-' && !IIsNumber(tokv[i])) {	/* a label */
      *str = tokv[i++];
    } else if (!strcmp("-t", tokv[i])) {				/* a time */
      if (tokv[++i]!=NULL) i++;
    } else
      if (tokv[i][0] == '-' && 					/* a field */
	ITableLookup(itf_table, tokv[i]+1, IGetType("EXAMPLE-REC", NULL))!=NULL) {
	if (tokv[++i]!=NULL) i++;
    } else {
      IError("unexpected token \"%s\"", tokv[i]);
      IError(IPrintUsage(command, usage));
      return 0;
    }
  }
  /* check for a colon */
  if (tokv[i]!=NULL && !strcmp(":", tokv[i])) {
    i++;
    while (   tokv[i]!=NULL
	   && (IIsNumber(tokv[i]) || !(tokv[i][0]=='-'&&tokv[i][1]=='\0'))) {
      i++;
      (*n)++;
    }
  }
  if (tokv[i]!=NULL) {
    IError(IPrintUsage(command, usage));
    return 0;
  }
  return 1;
}

int ReadExampleData(tokv, er)
char *tokv[];
EXAMPLE_REC *er;
{
  int i = 1;
  int n = 0;
  char *command = tokv[0];
  struct ADDRESS_ENV env;
  struct TBL_ITEM *item;
  /* check for flags */
  while (tokv[i] && tokv[i][0]=='-') {
    if (strprefix(tokv[i]+1, "c")) {
      if (tokv[i+1]) {
	er->comment = malloc(strlen(tokv[i+1])+1);
	if (er->comment==NULL) {
	  IOutOfMemory("ReadExampleData");
	  return 0;
	}
	strcpy(er->comment, tokv[i+1]);
	i++;
      }
    } else if (strprefix(tokv[i]+1, "t")) {
      if (tokv[i+1])
	er->t = atoi(tokv[++i]);
    } else {
      char *field_name = ITempCalloc(char, strlen(tokv[i])+1);
      strcpy(field_name, tokv[i]);
      field_name[0]='.';
      ICreateEnv(&env, "EXAMPLE-REC", er);
      if (!IContParseObject(field_name, &env, NULL)) {
       IError("\"%s\": unknown field in EXAMPLE-REC \"%s\"", command, tokv[i]+1);
	return 0;
      }
      ISetSimpleValue(&env, tokv[++i]);
    } /* replaced this with the field stuff  else {
      IError("\"%s\": unknown flag \"%s\"", command, tokv[i]);
      return 0;
    } */
    i++;
  }
  /* check for a label */
  if (tokv[i]!=NULL && !IIsNumber(tokv[i]) && strcmp(":", tokv[i])) {
    if (   (er->str = IGetName(tokv[i]))==NULL
	|| IGetVectorRealAddr(tokv[i], &n)==NULL) {
      if (!itf_errno) IError("can't find vector \"%s\"");
      return 0;
    }
    i++;
  }
  /* check for a colon & read following values */
  if (tokv[i]!=NULL && !strcmp(":", tokv[i])) {
    i++;
    while (tokv[i]!=NULL && IIsNumber(tokv[i])) {
      if (n>er->n_values) break;
      if (tokv[i][0]=='-' && tokv[i][1]=='\0')
	er->value[n] = quiet_nan(0);
      else
	er->value[n] = atof(tokv[i]);
      n++;
      i++;
    }
  }
  return 1;
}

int ReadFollowingExampleData(er)
EXAMPLE_REC *er;
{
  int i;
  double x;
  IMuckUpLineNumbers();
  for (i=0;i<er->n_values;i++) {
    if (fscanf(din, "%lf", &x)!=1) {
      IError("error reading value %d for \"%s\"", i, action_name[(int)er->action]);
      fprintf(stderr, "Discarding bad input: ");
      for (i=0;i!=EOF && i!='\n';) {
	i=fgetc(din);
	fputc(i, stderr);
      }
      return 0;
    }
    er->value[i] = x;
  }
  /* read upto \n */
  for (i=0;i!=EOF && i!='\n';) i=fgetc(din);
  return 1;
}

static int TryToAdd(er)
EXAMPLE_REC *er;
{
  EXAMPLE_ACTION tail_action;
  EXAMPLE_ACTION action = er->action;
  EXAMPLE_REC *er2;
  if (current_el==NULL)
    if (!(current_el = NewExampleList())) {
      warn("OkToAdd", "out of memory");
      FreeExampleRec(er);
      return 0;
    }
  if (current_el->tail!=NULL) tail_action = current_el->tail->action;
  else tail_action = NULL_ACTION;
  assert(VALID_ENUM(EXAMPLE_ACTION, action));
  assert(VALID_ENUM(EXAMPLE_ACTION, tail_action));
  if (action_follow[(int)tail_action][(int)action]==NO_FOLLOW) {
    IError("can't add a \"%s\" after a \"%s\"",
		  action_name[(int)action], action_name[(int)tail_action]);
    FreeExampleRec(er);
    return 0;
  } else if (FollowAction(tail_action, action)==FOLLOW) {
  } else if (FollowAction(tail_action, action)==AUTO_END && er->t==T_SENTINEL) {
    /* add the END record */
    if (!(er2 = NewExampleRec(END_EX, 0, NULL))) {
      FreeExampleRec(er);
      return 0;
    }
    AddEx(er2, current_el);
  }
  AddEx(er, current_el);
  return 1;
}

static int AddEx(er, el)
EXAMPLE_REC *er;
EXAMPLE_LIST *el;
{
  assert(el!=NULL);
  if (el->tail==NULL) {
    assert(el->head==NULL);
    el->head = el->tail = er;
    el->current = NULL;
    el->n_examples = 1;
  } else {
    assert(el->head!=NULL);
    if (el->tail->action==END_EX) el->n_examples++;
    el->tail->next = er;
    el->tail = er;
  }
  if (er->testing==-1)
    er->testing = el->default_testing;
  if (er->training==-1)
    er->training = el->default_training;
  if (er->validation==-1)
    er->validation = el->default_validation;
  if (er->ignore==-1)
    er->ignore = el->default_ignore;
  
  er->my_n = el->n_examples-1;
  if (er->action==END_EX)
    el->complete=1;
  else
    el->complete=0;

   /*
    * why was the following here? - we want this complete flag=0 in this case
    else if (FollowAction(er->action, (int)END_EX))
    el->complete=1; */
  return 1;
}

EXAMPLE_REC *NewExampleRec(action, n, str)
EXAMPLE_ACTION action;
int n;
string str;
{
  EXAMPLE_REC *er;
  if (   (er = Calloc(EXAMPLE_REC, 1))==NULL
      || (n>0 && (er->value = Calloc(real, n))==NULL)
      || (str!=NULL && (er->str = malloc(strlen(str)+1)) ==NULL)) {
    IOutOfMemory("NewExampleRec", "allocating er, vector, or str");
    return NULL;
  }
  er->action = action;
  er->dest_layer = 0;
  er->next = NULL;
  er->n_values = n;
  er->t = T_SENTINEL;
  er->comment = NULL;
  er->my_n = -1;
  er->ignore = -1;
  er->training = -1;
  er->testing = -1;
  er->validation = -1;
  if (str!=NULL)
    strcpy(er->str, str);
  else
    er->str=NULL;
  if (n<=0)
    er->value=NULL;
  ISetDefaults("EXAMPLE-REC", er);
  return er;
}

FreeExampleRec(er)
EXAMPLE_REC *er;
{
  if (er==NULL) return;
  if (er->value!=NULL) free(er->value);
  if (er->str!=NULL) free(er->str);
  free(er);
}

EXAMPLE_LIST *NewExampleList()
{
  EXAMPLE_LIST *el;
  int el_n;
  /***
   * Find a place in the vector of pointers
   * Extend that vector if necessary
   */
  for (el_n=0;el_n<n_els;el_n++) if (ela[el_n]==NULL) break;
  if (el_n>=n_els) {
    if (n_els>0)
      ela = (EXAMPLE_LIST**)realloc(ela, (n_els+1)*sizeof(EXAMPLE_LIST**));
    else
      ela = Calloc(EXAMPLE_LIST*, 1);
    if (ela==NULL) {
      IOutOfMemory("NewExampleList");
      return NULL;
    }
    el_n = n_els;
    n_els += 1;
  }
  fprintf(dout, "Creating new example list #%d\n", el_n);
  el = Calloc(EXAMPLE_LIST, 1);
  ela[el_n] = el;
  if (el==NULL)
    IOutOfMemory("NewExampleList");
  else
    el->my_n = el_n;
  el->head = el->tail = el->current = el->next = NULL;
  el->complete = 0;
  el->example_n = el->n_examples = 0;
  el->inactive = 0;
  el->index = NULL;
  el->index_size = 0;
  el->default_training = 1;
  el->default_testing = 0;
  el->default_validation = 0;
  el->default_ignore = 0;
  
  return el;
}

int command_new_el(tokc, tokv)
int tokc;
char *tokv[];
{
  if (GiveHelp(tokc)) {
    IUsage("");
    ISynopsis("create a new list of examples");
    IHelp(IHelpArgs,
	  NULL);
    return;
  }

  current_el = NewExampleList();
  if (current_el!=NULL)
    fprintf(dout, "New example list #%d created\n", current_el->my_n);
}

int command_use_el(tokc, tokv)
int tokc;
char *tokv[];
{
  int el_n;
  if (GiveHelp(tokc)) {
    IUsage("<example-list#>");
    ISynopsis("use a different example list");
    IHelp(IHelpArgs,
	  "Use a different list of examples from the array of example lists",
	  NULL);
    return;
  }

  if (tokc!=2 || !IIsInteger(tokv[1])) {
    IExpected("one integer argument");
    return;
  }
  el_n = atoi(tokv[1]);
  if (el_n<0 || el_n>=n_els) {
    IError("%d is not in the range of the current example array [%d-%d]", el_n, 0, n_els-1);
    return;
  }
  current_el = ela[el_n];
}

/* ARGSUSED */
int command_example_props(tokc, tokv)
int tokc;
char *tokv[];
{
  struct ARG_DESC *argd;
  int validation = 0;
  int ignore = 0;
  int training = 0;
  int testing = 0;
  char *range_string = NULL;
  int leave_others = 0;
  argd = StartArgs(tokv[0]);
  Args(argd, "[-Svalidation %S]", &validation);
  Args(argd, "[-Signore %S]", &ignore);
  Args(argd, "[-Straining %S]", &training);
  Args(argd, "[-Stesting %S]", &testing);
  Args(argd, "[<example-range>%R]", &range_string);
  Args(argd, "[-Pleave-others %P %?]", &leave_others,
       "Leave unnamed flags unchanged");
  EndArgs(argd);
  if (GiveHelp(tokc)) {
    ISynopsis("change the example properties");
    IHelp(tokc, tokv[0], NULL, synopsis,
	  "If no arguments are given, the current default properties",
	  "are printed out.",
	  "If some flags but no range is given, the default properties",
	  "are set to what the flags indicate.",
	  "If a range is given, then the properties for examples in that",
	  "range are set to what the flags indicate.",
	  NULL);
    return;
  }
  (void) ParseArgs(argd, tokc, tokv, 0);
  if (current_el==NULL) {
    IErrorAbort("The current example list is NULL");
  } else {
    if (tokc == 1) {
      fprintf(dout, "example-props %svalidation %straining %stesting %signore\n",
	      FlagPrefix(current_el->default_validation),
	      FlagPrefix(current_el->default_training),
	      FlagPrefix(current_el->default_testing),
	      FlagPrefix(current_el->default_ignore));
    } else {
      if (range_string==NULL) {
	if (!leave_others || validation!=-1)
	  current_el->default_validation = (validation==-1 ? 0 : validation);
	if (!leave_others || training!=-1)
	  current_el->default_training = (training==-1 ? 0 : training);
	if (!leave_others || testing!=-1)
	  current_el->default_testing = (testing==-1 ? 0 : testing);
	if (!leave_others || ignore!=-1)
	  current_el->default_ignore = (ignore==-1 ? 0 : ignore);
      } else {
	IErrorAbort("range changing not yet implemented!");
      }
    }
  }
}

/* ARGSUSED */
int command_ex_range(tokc, tokv)
int tokc;
char *tokv[];
{
  struct ARG_DESC *argd;
  char *train = "0-";
  char *validate = "X";
  char *test = "X";
  char *ignore = "X";
  int use = 1;
  int quiet = 0;
  int verbose = 0;
  argd = StartArgs(tokv[0]);
  Args(argd, "[-Suse %S %?]", &use, "turn ex-class-override on (or off)");
  Args(argd, "[-Squiet %S %?]", &quiet, "do not echo the settings");
  Args(argd, "[-Sverbose %S]", &verbose, "show settings in long form");
  Args(argd, "[<training-range>%R]", &train);
  Args(argd, "[<validation-range>%R]", &validate);
  Args(argd, "[<testing-range>%R]", &test);
  Args(argd, "[<ignore-range>%R]", &ignore);
  EndArgs(argd);
  if (GiveHelp(tokc)) {
    ISynopsis("set the example class ranges");
    IHelp(tokc, tokv[0], NULL, synopsis,
	  "If no arguments are given, the current setting are shown.",
	  NULL);
    return;
  }
  (void) ParseArgs(argd, tokc, tokv, 0);
  if (tokc==1) {
    fprintf(dout, "ex-range %suse %s %s %s %s\n",
	    FlagPrefix(true(ex_class_override)),
	    ex_training_range, ex_validation_range,
	    ex_testing_range, ex_ignore_range);
    IDoCommandLine("show ex-class-override ex-training-range ex-validation-range ex-testing-range ex-ignore-range");
  } else {
    ISetValue("ex-class-override", (use ? "true" : "false"));
    ISetValue("ex-training-range", train);
    ISetValue("ex-validation-range", validate);
    ISetValue("ex-testing-range", test);
    ISetValue("ex-ignore-range", ignore);
    if (!quiet)
      fprintf(dout, "ex-range %suse %s %s %s %s\n",
	      FlagPrefix(true(ex_class_override)),
	      ex_training_range, ex_validation_range,
	      ex_testing_range, ex_ignore_range);
  }
}

/******************************************************************************
 * New stuff
 */

int ExampleNumber(el,example_str)
EXAMPLE_LIST *el;
char *example_str;
{
  int e;
  EnsureELIndex(el);
  for (e=0; e<el->n_examples; e++)
    if (!strcmp(example_str,el->index[e]->comment)) return e;
  return -1;
}


int EnsureELIndex(el)
EXAMPLE_LIST *el;
{
  if (el->index_size!=el->n_examples||el->index==NULL) MakeELIndex(el);
}


int MatchExample(net,example_n,example_str)
NETWORK *net;
int example_n;
char *example_str;
{
  EnsureELIndex(current_el);
  if (   example_n == net->p->example_n
      || (   example_str != NULL
	  && !strcmp(example_str,
		     current_el->index[net->p->example_n]->comment)))
    return 1;
  else
    return 0;
}


int LoadExample(net,example_n,example_str)
NETWORK *net;
int example_n;
char *example_str;
{
  EnsureELIndex(current_el);
  ResetExamples(net);
  if (example_str) {
    while (NextExample(net) == EXAMPLE_OK)
      if (MatchExample(net,example_n,example_str)) return 1;
    IError("example %s not found",example_str);
    return 0;
  } else if (DoExample(net,example_n) == EXAMPLE_OK)
    return 1;
  IError("example %d not found",example_n);
  return 0;
}


void ReplaceExample(net)
NETWORK *net;
{
  fprintf(dout,"ReplaceExample %d (not implemented)\n",net->p->example_n);
}

int
  command_load_ex(tokc,tokv)
int tokc;
char *tokv[];
{
  int example_n = -1;
  char *example_str = NULL;
  IUsage("<example>");
  if (GiveHelp(tokc)) {
    ISynopsis("Load example.");
    IHelp(IHelpArgs,
"ARGUMENTS",
"example  : example to run (<int> or <comment>)",
	  NULL);
    return;
  }
  if (!NetExists(net)) { IError("current net is null"); return; }
  net->layer[net->p->input_layer_n].p->is_input_layer = TRUE;
  net->layer[net->p->output_layer_n].p->is_output_layer = TRUE;
  tokv++;	/* ignore command name */
  if (IIsInteger(tokv[0])) example_n = atoi(*++tokv);
  else 		           example_str = *++tokv; 
  if (!LoadExample(net,example_n,example_str)) {
    IError("Cannot load example.");
  }
}
