#ifndef _EXAMPLE_BP_
#define _EXAMPLE_BP_

#ifndef _EXAMPLE_
#define _EXAMPLE_

typedef enum EXAMPLE_CODE {
  EXAMPLE_OK,
  EXAMPLE_END,
  EXAMPLE_ERROR,
} EXAMPLE_CODE;

/*
 * These correspond to commands which we find in the example file
 */
typedef enum EXAMPLE_ACTION {
  EXAMPLE_ACTION_BOT,
  NULL_ACTION,
  INPUT,
  OUTPUT,
  PREDICT,
  COPY,
  CLEAR,
  FILL,
  POP,
  PUSH,
  CLEARSTACK,
  ENQUEUE,
  DEQUEUE,
  CLEARQUEUE,
  ECALL,
  EDO,
  END,		/* of loading a particular example */
  EXAMPLE_ACTION_TOP,
} EXAMPLE_ACTION;

typedef enum EXAMPLE_INTERP {DEFAULT,VECTOR,POSITION};

typedef struct EXAMPLE_REC {	/* BIND */
  EXAMPLE_ACTION	action;
  int			dest_layer;
  int			n_values;
  int			t;
  real			*value;	/* counter: n_values */
  struct EXAMPLE_REC	*next;
  string		str;
  string		comment;
  enum EXAMPLE_INTERP	ex_interp;
  int			my_n;
  int			training;	/* Flags that can be set	*/
  int			validation;	/* for the training to		*/
  int			testing;	/* take notice of.		*/
  int			ignore;		/* Skipped by NextExample()	*/
#ifdef EXAMPLE_REC_EXTRA_FIELDS
  EXAMPLE_REC_EXTRA_FIELDS
#endif
} EXAMPLE_REC;

typedef struct EXAMPLE_LIST {	/* BIND */
  struct EXAMPLE_REC	*head;
  struct EXAMPLE_REC	*tail;
  struct EXAMPLE_REC	*current;
  struct EXAMPLE_REC	*next;
  int			complete;     /* true if there is a final END action */
  int			example_n;
  int			n_examples;
  int			my_n;
  int			inactive;
  struct EXAMPLE_REC	**index;
  int			index_size;
  int			default_training;
  int			default_validation;
  int			default_testing;
  int			default_ignore;
} EXAMPLE_LIST;

/* this structure keeps properties of the most recent example */
typedef struct EXAMPLE { /* BIND - used in new version of example.c */
  int			n;
  int			predict;
  int			output;
  int			input;
  int			training;
  int			validation;
  int			testing;
  string		props_in_text;	/* ascii explanation for printing */
} EXAMPLE;

typedef struct EXAMPLE_PROPS { /* BIND - used in old version of example.c */
  int			predict;
  int			output;
  int			input;
};

extern struct EXAMPLE *example;		/* BIND */
extern struct EXAMPLE_LIST *current_el;	/* BIND altname: el */
extern struct EXAMPLE_REC *current_er;	/* BIND altname: er */
extern struct EXAMPLE_LIST **ela;	/* BIND counter: n_els */
extern int n_els;		/* BIND */
extern int example_show_end;	/* BIND */
extern string example_format;	/* BIND */
extern string example_type;	/* BIND should be a prefix of a function */
extern string example_interp;	/* BIND should be "vector", "position" */

typedef EXAMPLE_CODE (*PFEC)();	/* pointer to function return example code */
typedef EXAMPLE_REC *(*PFER)();	/* pointer to function return example code */

extern int ReadExamples();	/* (char *filename) */
extern EXAMPLE_CODE NextExample ARGS((NETWORK *));
extern EXAMPLE_CODE SkipExample ARGS((NETWORK *));
extern EXAMPLE_CODE DoExample ARGS((NETWORK *, int));
extern EXAMPLE_CODE ResetExamples ARGS((NETWORK *));

extern EXAMPLE_CODE LoadNextExample ARGS((NETWORK *, struct EXAMPLE*, int, int, int));
extern EXAMPLE_CODE SkipNextExample ARGS((NETWORK *, struct EXAMPLE*));
extern EXAMPLE_CODE DoExampleN ARGS((NETWORK *, struct EXAMPLE*, int));

extern int add_to_example_list();

extern string ex_training_range;		/* BIND default: 0- */
extern string ex_validation_range;		/* BIND default: X */
extern string ex_testing_range;			/* BIND default: X */
extern string ex_ignore_range;			/* BIND default: X */
extern string ex_class_override;		/* BIND default: false */

#define T_SENTINEL -1

#endif _EXAMPLE_

extern void ComputeExpectedValues();
extern int EnsureELIndex();
extern int LoadExample();
extern int MatchExample();

#endif _EXAMPLE_BP_
