#include <itf/option.h>
#include <itf/itf.h>
#include <sim/sim.h>
#include <lib/example.h>
#include "example-bp.h"
#include "prop-bp.h"

#define DB(x) 

int
  LesionNet(net,slayer,dlayer,sunit_range,dunit_range,lesion_p,lesion_units)
NETWORK *net;
int *slayer, *dlayer;
struct RANGE *sunit_range;
struct RANGE *dunit_range;
real lesion_p;
int lesion_units;
{
  int t;
  int count=0;
  /*
   * For each unit in destintation layer...
   */
  ForEachForwardLayer {
    if (dlayer[layer_n])
      ForEachUnit(UNIT) {
	if (lesion_units) {
	  /*
	   * (Un)lesion units
	   */
	  if (lesion_p==0.0) {
	    DB(printf("Unlesioning %d:%d\n",layer_n,unit_n););
	    unit->lesioned = 0;
	    ApplyToLinksOfUnit(LINK,UNIT,link->lesioned = 0;);
	  } else if (with_prob(lesion_p)) {
	    DB(printf("Lesioning %d:%d\n",layer_n,unit_n););
	    unit->lesioned = 1;  count++;
	    ApplyToLinksOfUnit(LINK,UNIT,link->lesioned = 1;);
	    UNIT_OUTPUT(unit) = 0.0;
#ifdef TIME_NETWORK
	    for (t=0; t<MAX_T; t++) UNIT_OUTPUT_T(unit,t) = 0.0;
#endif
	  }
	} else
	  ForEachLinkSet(LINK,UNIT) {
	    /*
	     * ...if incoming linkset is from source layer...
	     */
	    if (slayer[OtherLayerN]) {
	      if (lesion_p==0.0) {
		/*
		 * ...either "unlesion" each link...
		 */
		DB(printf("Unlesioning links from %d to %d\n",layer_n,OtherLayerN););
		ApplyToLinks(link->lesioned = 0;);
	      } else
		/*
		 * ...or lesion each link with prob <lesion_p>
		 */
		ApplyToLinks
		  (
		   if (with_prob(lesion_p)) {
		     DB(printf("Lesioning %s\n",LinkLocation););
		     link->lesioned = 1;  count++;
		   }
		   );
	    }
	  } EndForEach; /* LinkSet */
      } EndForEach; /* Unit */
  } EndForEach; /* Layer */
  CompileNet(net);
  return count;
}

void
  UnLesionNet(net)
NETWORK *net;
{
  int *slayer, *dlayer;
  struct RANGE *sunit_range;
  struct RANGE *dunit_range;

  AllocDisplayRanges(net,&dlayer,&slayer,&dunit_range,&sunit_range);
  LesionNet(net,slayer,dlayer,sunit_range,dunit_range,0.0,1);
}


void
  DescribeLesion(net,slayer,dlayer,sunit_range,dunit_range,lesion_p,
		 slayer_specified,dlayer_specified,count)
NETWORK *net;
int *slayer, *dlayer;
struct RANGE *sunit_range;
struct RANGE *dunit_range;
real lesion_p;
int slayer_specified,dlayer_specified;
int count;
{
  int sl,dl;
  /*
   * Print out what's going to be done
   */
  if (dlayer_specified) {
    if (slayer_specified) {
      /*
       * (Un)lesioning links
       */
      for (sl=0; sl<net->n_layers; sl++)
	if (slayer[sl])
	  for (dl=0; dl<net->n_layers; dl++)
	    if (dlayer[dl])
	      if (zerop(lesion_p))
		fprintf(dout,"Unlesioning %d-->%d links\n",sl,dl);
	      else
		fprintf(dout,"Lesioning %.2f (%d) of %d-->%d links\n",lesion_p,count,sl,dl);
    } else 
      /*
       * (Un)lesioning units
       */
      for (dl=0; dl<net->n_layers; dl++)
	if (dlayer[dl])
	  if (zerop(lesion_p))
	    fprintf(dout,"Unlesioning layer %d units\n",dl);
	  else
	    fprintf(dout,"Lesioning %.2f (%d) of layer %d units\n",lesion_p,count,dl);
  } else
    /*
     * (Un)lesioning everything
     */
    if (zerop(lesion_p))
      fprintf(dout,"Unlesioning network\n");
    else
      fprintf(dout,"Lesioning %.2f (%d) of network\n",lesion_p,count);
}


void
  SkeletonizeNet(net,slayer,dlayer,threshold)
NETWORK *net;
int *slayer, *dlayer;
real threshold;
{
  ApplyToLinksOfNet
    (UNIT,LINK,UNIT,
     if (fabs(LINK_WEIGHT(link)) <= threshold) link->lesioned = 1;
     );
  CompileNet(net);
}




/******************************************************************************
 * Commands
 */

int
  command_lesion(tokc,tokv)
int tokc;
char *tokv[];
{
  int seed = 0;
  int *slayer, *dlayer;
  int slayer_specified = 0;
  int dlayer_specified = 0;
  int sunit_specified = 0;
  int dunit_specified = 0;
  struct RANGE *sunit_range;
  struct RANGE *dunit_range;
  real lesion_p = 0.0;
  static struct OPTION options[] =
    {
      {OPTPRES,          "slayer",	0},
      {OPTPRES,          "sunit",	0},
      {OPTPRES,          "dlayer",	0},
      {OPTPRES,          "dunit",	0},
      {OPTPRES|OPTREALARG,"p",		0},
      {OPTPRES|OPTINTARG,"seed",	0},
      {OPTLAST,NULL,0}
    };

  IUsage("[[%pslayer <range>] %pdlayer <range>] [%pp <proportion>]\
 [%pseed <int>]");
  if (GiveHelp(tokc)) {
    ISynopsis("lesion a proportion of units, or links between layers.");
    IHelp(IHelpArgs,
"Lesion a proportion of the units in a layer, or the links between layers",
"OPTIONS",
"%pslayer <range> : source layer(s) of links to lesion.",
"%pdlayer <range> : destination layer(s) of links to lesion.",
"                  If slayer is not specified then units are lesioned.",
"%pp <proportion> : proportion of units/links to lesion (default 0.0).",
"                  If zero, all specified units/links are unlesioned.",
"%pseed <n>       : Seed random number generator with <n>.  ",
"                  This allows runs to be repeated.",
"\"lesion\" (without arguments) undoes all lesions.",
NULL);
    return;
  }
  if (!NetExists(net)) {
    IError("current net is null");
    return;
  }
  AllocDisplayRanges(net,&dlayer,&slayer,&dunit_range,&sunit_range);
  /***
   * process options
   */
  tokv++;
  while (IsOption(tokv[0])) {
    if (OptionMatch(tokv[0],"slayer",options,&slayer_specified) ||
	OptionMatch(tokv[0],"dlayer",options,&dlayer_specified)	||
	OptionMatch(tokv[0],"sunit", options,&sunit_specified)	||
	OptionMatch(tokv[0],"dunit", options,&dunit_specified)) {
      if (!SetDisplayRanges(net,&tokv,dlayer,slayer,dunit_range,sunit_range))
	return;
    } else if (OptionMatch(tokv[0],"p",options,NULL)) {
      if (!IExpectNumber(tokv[1])) return;
      lesion_p = atof(*++tokv); tokc--;
    } else if (OptionMatch(tokv[0],"seed",options,NULL)) {
      if (!IExpectInteger(tokv[1])) return;
      seed = atoi(*++tokv); tokc--;
    } else {
      IError("option not understood: \"%s\"",tokv[0]);
      return;
    }
    tokv++;
    tokc--;
  }
  if (slayer_specified && !dlayer_specified) {
    IError("cannot specify \"slayer\" without \"dlayer\"");
    return;
  }
  if (!zerop(lesion_p)) {
    if (!seed) seed = random(time(0));
    srandom(seed);
    fprintf(dout,"[%d] ",seed);
  }
  DescribeLesion(net,slayer,dlayer,sunit_range,dunit_range,lesion_p,slayer_specified,dlayer_specified,
		 LesionNet(net,slayer,dlayer,sunit_range,dunit_range,lesion_p,!slayer_specified));
}


int
  command_skeletonize(tokc,tokv)
int tokc;
char *tokv[];
{
  int *slayer, *dlayer;
  struct RANGE *sunit_range;
  struct RANGE *dunit_range;
  real threshold = 0.0;
  static struct OPTION options[] =
    {
      {OPTPRES,           "sunit",	0},
      {OPTPRES,           "slayer",	0},
      {OPTPRES,           "dunit",	0},
      {OPTPRES,           "dlayer",	0},
      {OPTPRES|OPTREALARG,"min",		0},
      {OPTLAST,NULL,0}
    };

  IUsage("[%pslayer <range>] [%pdlayer <range>]\
 [%pmin <real>]");
  if (GiveHelp(tokc)) {
    ISynopsis("lesion links with small weights.");
    IHelp(IHelpArgs,
"Lesion links with weight magnitude less than or equal to argument",
"   (default 0.0)",
"OPTIONS",
"%pslayer <range> : source layer(s) of links to skeletonize.",
"%pdlayer <range> : destination layer(s) of links to skeletonize.",
"%pmin <real>     : weights with magnitude <= <real> are lesioned.",
NULL);
    return;
  }
  if (!NetExists(net)) {
    IError("current net is null");
    return;
  }
  AllocDisplayRanges(net,&dlayer,&slayer,&dunit_range,&sunit_range);
   /***
   * process options
   */
  tokv++;
  while (IsOption(tokv[0])) {
    if (OptionMatch(tokv[0],"slayer",options,NULL) ||
	OptionMatch(tokv[0],"dlayer",options,NULL) ||
	OptionMatch(tokv[0],"sunit", options,NULL) ||
	OptionMatch(tokv[0],"dunit", options,NULL)) {
      if (!SetDisplayRanges(net,&tokv,dlayer,slayer,dunit_range,sunit_range))
	return;
    } else if (OptionMatch(tokv[0],"min",options,NULL)) {
      if (!IExpectNumber(tokv[1])) return;
      threshold = atof(*++tokv); tokc--;
    } else {
      IError("option not understood: \"%s\"",tokv[0]);
      return;
    }
    tokv++;
    tokc--;
  }
  fprintf(dout,"Lesioning links with weight <= %f\n",threshold);
  SkeletonizeNet(net,slayer,dlayer,threshold);
}
