#include <itf/option.h>
#include <itf/itf.h>
#include <sim/sim.h>
#include <lib/example.h>
#include "example-bp.h"
#include "prop-bp.h"

#define DB(x) x

#ifdef DBM_NETWORK
extern real FreeEnergy();	/* prop-dbm.c */
extern real Energy();		/* prop-dbm.c */
extern real Entropy();		/* prop-dbm.c */
#endif DBM_NETWORK

extern int  ExampleNumber();	/* example-bp.c */
extern real cos_angle();	/* states-bp.c */
extern real correlation();	/* states-bp.c */
extern int  Connected();	/* energy-bp.c */
extern real GaussNoise();

extern int therapy;		/* therapy-bp.c */
extern int therapy_proto;	/* therapy-bp.c */
extern int therapy_sim;		/* therapy-bp.c */
extern int therapy_details;	/* therapy-bp.c */
extern int therapy_split;	/* therapy-bp.c */
extern int therapy_split_size;	/* therapy-bp.c */

real dist, gap, prob;
int  proximity, abs, phon;

real vis_crit, sem_crit, phon_crit;

#define PROB_MULT		-0.01
#define PROB_NONE		-0.02
#define PROB_HILO		-0.03
#define PROB_STAB		-0.04

#define CON_SIM_CRIT		84
#define ABS_SIM_CRIT		95
#define N_CATEGORIES		5
int n_cats			= N_CATEGORIES;

#define N_ITEM_COUNTS		17
static int n_item_counts 	= N_ITEM_COUNTS;
static int **stim_count 	= NULL; /* item counts for objects as stimuli */
static int **resp_count 	= NULL; /* item counts for objects as response */

static int confusion		= FALSE;
static int **confusion_count	= NULL; /* confusion matrix */

#define ITEM_CORRECT		0
#define ITEM_OMISSION		1
#define ITEM_ERROR		2
#define ITEM_TARGET		3
#define ITEM_VISUAL		4
#define ITEM_SEMANTIC		5
#define ITEM_MIXED		6
#define ITEM_OTHER		7
#define ITEM_BLEND		8
#define ITEM_PHONOLOGICAL	9
#define ITEM_PHON_MIXED		10
#define ITEM_CROSS		11
#define ITEM_PAIR		12

#define HEBB_S_I		8	/* S+P  */   
#define HEBB_S_S		9	/* S+C  */
#define HEBB_S_O		10	/* S+U  */
#define HEBB_SIS		11	/* S+nP */
#define HEBB_O_I		12	/* O+P  */
#define HEBB_O_S		13	/* O+C  */
#define HEBB_OOO		14	/* O+U  */
#define HEBB_OSO		15	/* O+nC */
#define HEBB_OIO		16	/* O+nP */

static int hebb,hebb_ex_n,hebb_prime,hebb_processing_prime,hebb_prime_response_type;
static int **hebb_prime_count 	= NULL;	/* item counts for objects as prime */
static int **hebb_confusion_count	= NULL; /* prime-response confusion matrix */

static char *item_header[N_ITEM_COUNTS] = {
  "corr", "omit", "err", "targ", "vis", "sem", "mix", "othr", "blnd", "phon", "pmix", "cros", "N-20", "",     "",     "",     ""
  };
static char *hebb_item_header[N_ITEM_COUNTS] = {
  "corr", "omit", "err", "targ", "vis", "sem", "mix", "othr", "S+P",  "S+C",  "S+U",  "S+nP", "O+P",  "O+C",  "O+U",  "O+nC", "O+nP"
  };
static char *error_char[N_ITEM_COUNTS] = {
  ".",    "^",    "*",   "T",    "V",   "S",   "M",   "O",    "?",    "P",    "X",    "C",    "N",    "",     "",     "",     ""
  };

#define STATS_CON 		1
#define STATS_ABS		2
#define STATS_N_LESIONS		12
static int  stats_slayer,stats_dlayer;
static real stats_severity;
static int  stats_count[N_ITEM_COUNTS];
static real loc_to_severity[STATS_N_LESIONS] = {
  0.05, 0.1, 0.15, 0.2, 0.25, 0.3, 0.4, 0.5, 0.7, 0.8, 0.9, 1.0
  };
static int stats_location[8][8] = {
  {0, 1, 2, 3, 4,  5,  6,  7},	/* 0 */
  {0, 0, 1, 6, 0,  0,  0,  0},	/* 1 */
  {0, 0, 0, 7, 2,  0,  0,  0},	/* 2 */
  {0, 0, 8, 9, 3,  0,  0,  0},	/* 3 */
  {0, 0, 5, 4, 0, 10, 11,  0},	/* 4 */
  {0, 0, 0, 0, 0,  0, 12, 13},	/* 5 */
  {0, 0, 0, 0, 0, 14,  0, 15},	/* 6 */
  {0, 0, 0, 0, 0,  0, 16,  0}	/* 7 */
};

static int sem_el_n;		/* for specific sem distance calculations */
static int sem_layer_n;
static real *weights = NULL;


/******************************************************************************
 * Orthographic/phonological codes
 */

static char *orthtag[27] = {
  "A", "B", "C", "D", "E", "F", "G", "H", "I", "J", "K", "L", "M", 
  "N", "O", "P", "Q", "R", "S", "T", "U", "V", "W", "Y", "X", "Z",
  "_"
  };

static int orthoct[27] = {
  0126,0271,0050,0270,0310,0300,0141,0315,0314,0340,0213,0301,0207,
  0203,0074,0260,0062,0263,0041,0304,0244,0006,0007,0207,0017,0103,
  0000
  };

static char *phontag[9][20] = {
  {"S",  "_"},
  {"P",  "B",  "M",  "T",   "D",  "N", "K", "G",  "F",  "V",   "Z",   "SH", "H",   "CH", "DZ", "S",  "L",  "R", "W", "_"},
  {"L",  "R",  "W",  "Y",   "_"},
  {"EE", "I",  "E",  "A",   "AR", "O", "U", "AI", "OA", "EER", "IRE", "AW", "AIR", "IE", "EW", "OW", "UU"},
  {"L",  "M",  "N",  "S",   "_"},
  {"J", "P",  "B",  "T",   "D",  "K", "G", "F",  "V",  "SH",  "Z",   "_"},
  {"T",  "S",  "_"},
  {"EE", "A",  "U",  "EEO", "_"},
  {"L",  "N",  "_"}
};

static int phonoct[9][20] = {
  {01,00},
  {0070,0030,0134,0050,0010,0106,0060,0020,0074,0034,0016,0046,0024,0042,0002,0056,0005,0015,0035,0000},
  {03,01,07,05,00},
  {0150,0050,0070,0030,0120,0024,0060,0072,0065,0053,0023,0164,0033,0022,0144,0021,0044},
  {010,04,06,03,00},
  {001,016,006,012,002,014,004,017,007,011,003,000},
  {02,03,00},
  {06,02,04,07,00},
  {01,03,00},
};
static int phonalts[9]  = {2,20,5,17,5,12,3,5,3};
static int phonunits[9] = {1, 7,3, 7,3, 4,2,3,2};


static char *phloctag[3][13] = {
  {"B", "D",  "DY", "G", "H",  "J", "K",  "L", "M",  "N",  "P", "R", "T"},
  {"A", "AR", "AW", "E", "EW", "I",  "IE", "O", "OA", "OW", "U"},
  {"B", "D",  "G",  "K", "N",  "M",  "P",  "T", "_"}
};
static int phlocunits[3] = {13,11,9};


static char *phonloctag[9][17] = {
  {"S", "_"},
  {"B", "CH", "D", "J", "F", "G", "H", "K", "M", "N", "P", "SH", "T", "V", "Z", "_"},
  {"L", "R", "W", "Y", "_"},
  {"A", "AI", "AIR", "AR", "AW", "E", "EE", "EER", "EW", "I", "IE", "IRE", "O", "OA", "OW", "U", "UU"},
  {"L", "M", "N", "S", "_"},
  {"B", "D", "J", "F", "G", "K", "P", "SH", "T", "V", "Z", "_"},
  {"S", "T", "Z", "_"},
  {"A", "EE", "EEO", "U", "_"},
  {"L", "N", "_"}
};
static int phonlocunits[9] = {2,16,5,17,5,12,4,5,3};

static int orthunits = 8;

static char *output_tag[9];	/* string output */
static real min_prob;		/* min probability of phon output */


/******************************************************************************
 * Utilities
 */


void
  SavePerformWeights(net)
NETWORK *net;
{
  if (weights==NULL) weights = ((real*)calloc(net->n_links,sizeof(real)));
  net->p->delta_bar_delta = FALSE;
  ForEachLinkCompiled(LINK,UNIT) {
    weights[link_n] = LINK_WEIGHT(link);
  } EndForEach;
}

void
  RestorePerformWeights(net)
NETWORK *net;
{
  fprintf(dout,"# Restoring Weights\n");
  ForEachLinkCompiled(LINK,UNIT) {
    LINK_WEIGHT(link) = weights[link_n];
    link->dEdw = 0.0;
  } EndForEach;
}

int
  percent(x)
real x;
{
  return rint(100.0*x);
}

int
  percent_overlap(v1,v2,n)
real *v1,*v2;
int n;
{
  int i,overlap = 0;
  for_each(i,n) if (v1[i] == v2[i]) overlap++;
  if (overlap==n) return -1;
  return percent(overlap/(real)n);
}

real
  output_one_minus_cos_angle(net,values)
NETWORK *net;
real *values;
{
  int i = 0;
  real v, l12 = 0.0, l1 = 0.0, l2 = 0.0;
  ApplyToEachOutputUnit
    (
     v    = ScaleOutput(values[i]);
     l1  += UNIT_OUTPUT(unit)*UNIT_OUTPUT(unit);
     l2  += v*v;
     l12 += UNIT_OUTPUT(unit)*v;
     i++;
     );
  if (l1==0.0 || l2==0.0) return 0.0;
  return (1.0-l12/sqrt(l1*l2));
}

real
  output_distance(net,values)
NETWORK *net;
real *values;
{
  int u = 0;
  real dst = 0.0;
  if (proximity) return output_one_minus_cos_angle(net,values);
  ApplyToEachOutputUnit
    (
    dst += sqr(UNIT_OUTPUT(unit)-ScaleOutput(values[u++]));
     );
  return sqrt(dst);
}

real
  layer_cos_angle(net,lay_n,values)
NETWORK *net;
int lay_n;
real *values;
{
  real val, outsq = 0.0, valsq = 0.0, cross = 0.0;
  DoLayer(lay_n) {
    ForEachUnit(UNIT) {
      val    = ScaleOutput(values[unit_n]);
      outsq += UNIT_OUTPUT(unit)*UNIT_OUTPUT(unit);
      valsq += val*val;
      cross += UNIT_OUTPUT(unit)*val;
    } EndForEach;
  } EndDo;
  if (outsq==0.0 || valsq==0.0) return 1.0;
  return (cross/sqrt(outsq*valsq));
}

real
  layer_one_minus_cos_angle(net,lay_n,values)
NETWORK *net;
int lay_n;
real *values;
{
  return (1.0 - layer_cos_angle(net,lay_n,values));
}

real
  layer_distance(net,lay_n,values)
NETWORK *net;
int lay_n;
real *values;
{
  real dst = 0.0;
  if (proximity) return layer_one_minus_cos_angle(net,lay_n,values);
  DoLayer(lay_n) {
    ForEachUnit(UNIT) {
      dst += sqr(UNIT_OUTPUT(unit)-ScaleOutput(values[unit_n]));
    } EndForEach;
  } EndDo;
  return sqrt(dst);
}

real
  target_distance(net,values)
NETWORK *net;
real *values;
{
  int u = 0;
  real dst = 0.0;
  ApplyToEachOutputUnit
    (
    dst += sqr(UNIT_TARGET(unit)-ScaleOutput(values[u++]));
     );
  return sqrt(dst);
}

int
  overlap(a,b)
char *a,*b;
{
  int count = 0;
  while (*a && *b) {
    if (*a == *b) count++;
    a++; b++;
  }
  return count;
}

int
  overlap_other_than(a,b,c)
char *a,*b,*c;
{
  int count = 0;
  while (*a && *b) {
    if (*a!=*c && *a==*b) count++;
    a++;  b++;
  }
  return count;
}

void
  PrintExampleName(rec)
EXAMPLE_REC *rec;
{
  if (rec->comment) fprintf(dout,"%s",rec->comment);
  else	            fprintf(dout,"%d",rec->my_n);
}

void PrintItemHeaders()
{
  int i;
  fprintf(dout,"\t");
  for_each(i,n_item_counts) 
    fprintf(dout," %5s",(hebb?hebb_item_header[i]:item_header[i]));
  fprintf(dout,"\n");
}

int
  lesion_i(lesion_p)
real lesion_p;
{
  int i;
  for_each(i,STATS_N_LESIONS) if (loc_to_severity[i]==lesion_p) return i+1;
  return 0;
}

void
  PrintStats(abs_field)
int abs_field;
{
  int i;
  fprintf(dout,"@ %d %d",
	  stats_location[stats_slayer][stats_dlayer],
	  lesion_i(stats_severity));
  if (abs_field) fprintf(dout," %d",abs_field);
  for_each(i,n_item_counts)
    if (i!=ITEM_TARGET && (hebb || (i!=ITEM_OMISSION && i!=ITEM_OTHER)))
      fprintf(dout," %d",stats_count[i]);
  if (hebb) fprintf(dout," %d",hebb_prime_response_type);
  fprintf(dout,"\n");
}

void
  ClearStats()
{
  int i;
  for_each(i,n_item_counts) stats_count[i] = 0;
}



/******************************************************************************
 * Error classification
 */

int
  VisuallySimilar(response,target)
EXAMPLE_REC *response,*target;
{
  int n_share;
  real val;
  if (vis_crit) {
    val = cos_angle(response->value,target->value,response->n_values);
    if (val > vis_crit) return percent(val);
    else                return FALSE;
  } else if (response->comment==NULL || target->comment==NULL) {
    return FALSE;
  } else {
    n_share = overlap(response->comment,target->comment);
    if ((abs || hebb || n_cats==4) && n_share<2) n_share = FALSE;
    return n_share;
  }
}

int
  SemanticallySimilar(response,target)
EXAMPLE_REC *response,*target;
{
  int sim,crit,cat_size,result;
  real val;
  if (sem_crit) {
    val = (sem_layer_n ?
	   cos_angle(ela[sem_el_n]->index[response->my_n]->next->value,
		     ela[sem_el_n]->index[target->my_n]->next->value,
		     ela[sem_el_n]->index[response->my_n]->next->n_values) :
	   cos_angle(response->next->value,
		     target->next->value,
		     response->next->n_values));
    if (val > sem_crit) return percent(val);
    else                return FALSE;
  } else if (abs) {
    sim = (sem_layer_n ?
	   percent_overlap(ela[sem_el_n]->index[response->my_n]->next->value,
		      ela[sem_el_n]->index[target->my_n]->next->value,
		      ela[sem_el_n]->index[response->my_n]->next->n_values) :
	   percent_overlap(response->next->value,target->next->value,net->p->output_width));
    crit = (target->my_n < ela[sem_el_n]->n_examples/2 ? CON_SIM_CRIT : ABS_SIM_CRIT);
    result = (sim >= crit);
  } else {
    cat_size = current_el->n_examples / n_cats;
    result = (response->my_n/cat_size == target->my_n/cat_size);
  }
  return result;
}

int
  PhonologicallySimilar(response,target)
EXAMPLE_REC *response,*target;
{
  int n_share;
  real val;
  if (!phon) return FALSE;
  else if (phon_crit) {
    val = cos_angle(response->next->value,target->next->value,response->n_values);
    if (val > phon_crit) return percent(val);
    else                 return FALSE;
  } else if (!response->comment || !target->comment) {
    return FALSE;
  } else {
    n_share = overlap_other_than(response->next->comment,target->next->comment,"-");
    if (abs && n_share<2) n_share = FALSE;
    return n_share;
  }
}

int
  ErrorType(response,target,print)
EXAMPLE_REC *response,*target;
int print;
{
  int vis = VisuallySimilar(response,target);
  int sem = SemanticallySimilar(response,target);
  int phn = PhonologicallySimilar(response,target);
  if (phn > vis && (!phon_crit==!vis_crit)) vis = 0;
  if (vis && sem) {
    if (print) fprintf(dout,"%s",error_char[ITEM_MIXED]);
    return ITEM_MIXED;
  } else if (phn && sem) {
    if (print) fprintf(dout,"%s",error_char[ITEM_PHON_MIXED]);
    return ITEM_PHON_MIXED;
  } else if (vis) {
    if (print) fprintf(dout,"%s",error_char[ITEM_VISUAL]);
    return ITEM_VISUAL;
  } else if (phn) {
    if (print) fprintf(dout,"%s",error_char[ITEM_PHONOLOGICAL]);
    return ITEM_PHONOLOGICAL;
  } else if (sem) {
    if (print) fprintf(dout,"%s",error_char[ITEM_SEMANTIC]);
    return ITEM_SEMANTIC;
  } else {
    if (print) fprintf(dout,"%s",error_char[ITEM_OTHER]);
    return ITEM_OTHER;
  }
}

int
  HebbErrorType(sr,hs,hr)
int sr,hs,hr;
{
  if (sr==ITEM_SEMANTIC || sr==ITEM_MIXED) {
    if (hr==ITEM_CORRECT) 			return HEBB_S_I;
    if (hr==ITEM_SEMANTIC || hr==ITEM_MIXED) {
      if (hs==ITEM_CORRECT) 			return HEBB_SIS;
      else 					return HEBB_S_S;
    } else 					return HEBB_S_O;
  }
  if (hr==ITEM_CORRECT) 			return HEBB_O_I;
  if (hr==ITEM_SEMANTIC || hr==ITEM_MIXED) 	return HEBB_O_S;
  if (hs==ITEM_CORRECT) 			return HEBB_OIO;
  if (hs==ITEM_SEMANTIC || hs==ITEM_MIXED) 	return HEBB_OSO;
  						return HEBB_OOO;
}

int
  OutputMeetsCriteria(net,best_targets)
NETWORK *net;
real *best_targets;
{
  int p,p_on,i,it;
  int max_t = net->p->max_t;
  int loc = 0, orth = 0;
  int init_pos,end_pos,n_units;
  int start = 0, end = 0;
  int correct = 1, blend = 0;
  int n_iterations = 2;
  long pos_output,winner[2];
  real output,target,outprob;
  real hi = (dist==HUGE ? 0.5  : dist);
  real lo = (gap== -1   ? 1-hi : gap);
  min_prob = 1.0;
  /*
   * Set parameters
   */
  switch (np->output_criteria) {
  case 1:	/* phloc */
    loc++;
    init_pos = 0; end_pos = 2; 
    break;
  case 2: 	/* phonloc */
    loc++;
    switch (np->output_width) {
    case 55: init_pos = 1; end_pos = 5; break;	/* orig */
    case 61: init_pos = 0; end_pos = 6; break;	/* abs */
    case 69: init_pos = 0; end_pos = 8; break;	/* obj */
    }
    break;
  case 3:	/* phon */
    switch (np->output_width) {
    case 24:  init_pos = 1; end_pos = 5; break;	/* orig */
    case 27:  init_pos = 0; end_pos = 6; break;	/* abs */
    case 32:  init_pos = 0; end_pos = 8; break;	/* obj */
    }
    break;
  case 4:	/* orth */
    orth++; 
    init_pos = 0; end_pos = 2; 
    break;
  }
  DoLayer(np->output_layer_n) {
    for (p=0;        p<9;        p++) output_tag[p] = NULL;
    for (p=init_pos; p<=end_pos; p++) output_tag[p] = " ";
    /*
     * For each slot
     */
    for (p=init_pos; p<=end_pos; p++) {
      switch (np->output_criteria) {
      case 1: n_units = phlocunits[p]; 	 break;
      case 2: n_units = phonlocunits[p]; break;
      case 3: n_units = phonunits[p];	 break;
      case 4: n_units = orthunits;	 break;
      }
      end += n_units;
      /*
       * Determine slot output
       */
#ifdef DBM_NETWORK
      n_iterations = 1;
#endif DBM_NETWORK
      for (it=max_t; it>max_t-n_iterations; it--) {
	pos_output = 0;  outprob = 1.0;
	for (i=start; i<end; i++) {
	  DoUnit(UNIT,i) {
	    pos_output   = pos_output  << 1;
	    output       = InvScaleOutput(UNIT_OUTPUT_T(unit,it));
	    target	 = InvScaleOutput(UNIT_TARGET_T(unit,it));
	    if (prob>=0) {
	      if (output > 0.5) {
		if (pos_output++ && loc) {min_prob=PROB_MULT; return FALSE;}
		p_on = i-start;
		outprob *= output;
	      } else {
		outprob *= 1.0-output; 
	      }
	    } else if (output >= hi) {
	      if (pos_output++ && loc) {min_prob=PROB_MULT; return FALSE;}
	      p_on = i-start;
	    } else if (output > lo) {min_prob=PROB_HILO; return FALSE;}
	    if (target!=(pos_output&1)) correct = -1;
	    if (best_targets && (best_targets[i]!=(pos_output&1))) blend =  1;
	  } EndDo; /* Unit */
	} /* For Units */
	if (prob>=0) {
	  if (outprob < min_prob) min_prob = outprob;
	  if (outprob < prob)     return FALSE;		/* fail prob */
	}
	if (loc && !pos_output) {min_prob=PROB_NONE; return FALSE;}
	winner[max_t-it] = pos_output;
      } /* Iteration */
#ifndef DBM_NETWORK
      if (winner[0] != winner[1]) {min_prob=PROB_STAB; return FALSE;}
#endif  DBM_NETWORK
      /*
       * Store tag
       */
      switch (np->output_criteria) {
      case 1: 
	output_tag[p] = phloctag[p][p_on];
	break;
      case 2: 
	output_tag[p] = phonloctag[p][p_on];
	break;
      case 3: 
	for (i=0; i<phonalts[p]; i++) 
	  if (winner[0]==phonoct[p][i]) {output_tag[p]=phontag[p][i]; break;}
	if (output_tag[p]==" ") output_tag[p] = "?";
	break;
      case 4:
	for (i=0; i<27; i++)
	  if (winner[0]==orthoct[i]) {output_tag[p]=orthtag[i]; break;}
	if (output_tag[p]==" ") output_tag[p] = "?";
	break;
      }
      start += n_units;
    } /* position */
  } EndDo; /* Layer */
  if (blend) return -2;
  return correct;					/* succeed */
}


void
  PrintOutputTags()
{
  int p;
  fprintf(dout,"/ ");
  for (p=0; p<9; p++) 
    if (output_tag[p]!=NULL) fprintf(dout,"%s ",output_tag[p]);
  fprintf(dout,"/ ");
}



/*****************************************************************************
 * PerformExamples()
 */

void
  PerformExamples(net,el_n,n_best,example_n,errors,omissions,items,stats)
NETWORK *net;
int el_n,n_best,example_n,errors,omissions,items,stats;
{
  NET_PARAMS *np = net->p;
  int e,ex_n,b;
  int criteria_met,print,response_type,error_type;
  real old_input_noise = np->input_noise;
  int correct,n_correct = 0, n_response = 0, best = 0;
  int vis,sem,bridge;
  int hebb_HS_type,hebb_HR_type,hebb_response_type;
  int old_el_n       = current_el->my_n;
  int example        = (example_n != -1);
  int n_examples     = current_el->n_examples;
  int  *bestn        = ITempCalloc(int,n_best);
  real *bestdist     = ITempCalloc(real,n_best);
  real *distance     = ITempCalloc(real,n_examples);
  int  *sem_bestn;
  int non_sem_count, non_sem_farther;
  real *sem_bestdist, *sem_distance;
  EXAMPLE_CODE ex_status = EXAMPLE_OK;
  EnsureELIndex(ela[el_n]);		/* "new" el */
  EnsureELIndex(ela[old_el_n]);		/* "old" el */
  current_el = ela[el_n];
  /* np->input_noise = 0; */
  if (sem_layer_n) {
    EnsureELIndex(ela[sem_el_n]);
    sem_bestn        = ITempCalloc(int,n_best);
    sem_bestdist     = ITempCalloc(real,n_best);
    sem_distance     = ITempCalloc(real,n_examples);
  }
  if (stats) ClearStats();
  ResetExamples(net);
  for_each (ex_n,n_examples) {
    if (example && ex_n!=example_n) continue;
    if (!LoadExample(net,ex_n,NULL)) return;
    PerformNet(net);
    for_each(e,n_examples) 
      distance[e] = output_distance(net,ela[old_el_n]->index[e]->next->value);

    non_sem_count = non_sem_farther = 0;
    for_each(e,n_examples) 
      if (!SemanticallySimilar(ela[el_n]->index[ex_n],ela[old_el_n]->index[e])) {
	non_sem_count++;
	if (distance[e] > distance[ex_n])
	  non_sem_farther++;
      }
    
    for (b=0,best=0; b<n_best; b++) {
      for (e=1; e<n_examples; e++)
	if (distance[e] < distance[best]) best = e;
      bestn[b]    = best;
      bestdist[b] = distance[best];
      distance[best] = HUGE;
     }
    if (sem_layer_n) {
      for_each(e,n_examples) 
	sem_distance[e] = layer_distance(net,sem_layer_n,ela[sem_el_n]->index[e]->next->value);
      for (b=0,best=0; b<n_best; b++) {
	for (e=1; e<n_examples; e++)
	  if (sem_distance[e] < sem_distance[best]) best = e;
	sem_bestn[b]    = best;
	sem_bestdist[b] = sem_distance[best];
	sem_distance[best] = HUGE;
      }
    } /* sem_layer_n */
    /*
     * bestn[0] now contains number of word that best matches response
     */
    criteria_met =
      (phon ?
       OutputMeetsCriteria(net,ela[old_el_n]->index[bestn[0]]->next->value) :
       (bestdist[0]<dist && (bestdist[1]-bestdist[0])>gap));
    if (criteria_met) {
      n_response++;
      if (confusion) confusion_count[ex_n][bestn[0]]++;
    }
    n_correct += correct = (ex_n==bestn[0] && el_n==old_el_n && criteria_met==1);
    print = (example || (confusion!=-1 && items!=-1 && (criteria_met || omissions) && !(correct && errors)));
    if (!criteria_met) response_type = ITEM_OMISSION;
    else if (correct)  response_type = ITEM_CORRECT;
    else {                     
      response_type = ITEM_ERROR;
      if (criteria_met==-2) {
	error_type = ITEM_BLEND;
	if (print) fprintf(dout,"%s",error_char[ITEM_BLEND]);
      } else {
	error_type = ErrorType(ela[old_el_n]->index[bestn[0]],ela[el_n]->index[ex_n],print);
	if (hebb_processing_prime) 
	  hebb_prime = bestn[0];
	else if (hebb) {
	  if (ex_n==hebb_prime) {hebb_HS_type = ITEM_CORRECT; fprintf(dout,"I");}
	  else hebb_HS_type = ErrorType(ela[el_n]->index[ex_n],        ela[el_n]->index[hebb_prime],print);	/* H-S */
	  if (bestn[0]==hebb_prime) {hebb_HR_type = ITEM_CORRECT; fprintf(dout,"I");}
	  else hebb_HR_type = ErrorType(ela[old_el_n]->index[bestn[0]],ela[el_n]->index[hebb_prime],print);	/* H-R */
	  hebb_response_type = HebbErrorType(error_type,hebb_HS_type,hebb_HR_type);
	  if (stats) stats_count[hebb_response_type]++;
	  if (items) {
	    stim_count[ex_n][hebb_response_type]++;
	    resp_count[bestn[0]][hebb_response_type]++;
	    hebb_prime_count[hebb_prime][hebb_response_type]++;
	  }
	  if (confusion) hebb_confusion_count[hebb_prime][bestn[0]]++;		/* only errors */
	}
	bridge = (sem_layer_n ? sem_bestn[0] : bestn[1]);
	if (print) {
	  if (sem_layer_n && bridge != ex_n && bridge != bestn[0]) {
	    vis = VisuallySimilar(ela[el_n]->index[ex_n],ela[old_el_n]->index[bridge]);
	    sem = SemanticallySimilar(ela[el_n]->index[ex_n],ela[old_el_n]->index[bridge]);
	    fprintf(dout,"%s>",(vis&&sem?"m":(vis?"v":(sem?"s":"o"))));
	    vis = VisuallySimilar(ela[old_el_n]->index[bridge],ela[old_el_n]->index[bestn[0]]);
	    sem = SemanticallySimilar(ela[old_el_n]->index[bridge],ela[old_el_n]->index[bestn[0]]);
	    fprintf(dout,"%s",(vis&&sem?"m":(vis?"v":(sem?"s":"o"))));
	  }
	} /* print */
      } /* non-blend error */
    } /* error */
    if (hebb_processing_prime) hebb_prime_response_type = response_type;
    if (stats) {
      stats_count[response_type]++;
      if (response_type==ITEM_ERROR) {
	stats_count[error_type]++;
	if (abs) {
	  if ((ex_n<n_examples/2) != (bestn[0]<n_examples/2)) 
	    stats_count[ITEM_CROSS]++;
	  if (ex_n%(n_examples/2) == bestn[0]%(n_examples/2)) 
	    stats_count[ITEM_PAIR]++;
	}
      }
    } /* stats */
    if (items) {
      stim_count[ex_n][response_type]++;
      resp_count[bestn[0]][response_type]++;
      if (response_type==ITEM_ERROR) {
	stim_count[bestn[0]][ITEM_TARGET]++;
	stim_count[ex_n][error_type]++;
	resp_count[bestn[0]][error_type]++;
	if (abs && (ex_n%(n_examples/2) == bestn[0]%(n_examples/2))) {
	  stim_count[ex_n][ITEM_PAIR]++;
	  resp_count[bestn[0]][ITEM_PAIR]++;
	}
      }
      if (hebb) {
	hebb_prime_count[hebb_prime][response_type]++;
	if (response_type==ITEM_ERROR) {
	  hebb_prime_count[hebb_prime][error_type]++;
	}
      }
    } /* item */
    if (print) {
      fprintf(dout,"%s",error_char[response_type]);
      PrintExampleName(ela[el_n]->index[ex_n]);
      fprintf(dout,":\t");
      if (phon) PrintOutputTags();
      /*
       * Print nbest response
       */
      for_each(b,n_best) {
	PrintExampleName(ela[old_el_n]->index[bestn[b]]);
	fprintf(dout,"(");
	fprintf(dout,"%.2f",bestdist[b]);
	if (zerop(b)) {
	  if (prob>=0) fprintf(dout," %2d",percent(min_prob));
#ifdef DBM_NETWORK
	  if (Connected(net,np->output_layer_n,np->output_layer_n))
	    fprintf(dout," %.1f",-FreeEnergy(net,np->output_layer_n,np->output_layer_n));
	  fprintf(dout," %.1f",-FreeEnergy(net,0,0));
	  fprintf(dout," %d",net->p->max_t);	/* settling iterations */
	  if (np->input_strength<1.0)
	    fprintf(dout," %.3f",
		    layer_one_minus_cos_angle(net,np->input_layer_n,ela[el_n]->index[ex_n]->value));
#else
	  fprintf(dout," %2d %2d",
		  (proximity ?	/* vis sim */
		   percent(cos_angle(ela[el_n]->index[ex_n]->value,
				     ela[old_el_n]->index[bestn[0]]->value,
				     np->input_width)) :
		   percent_overlap(ela[el_n]->index[ex_n]->value,
				   ela[old_el_n]->index[bestn[0]]->value,
				   np->input_width)),
		  (proximity ?	/* phon sim */
		   percent(cos_angle(ela[el_n]->index[ex_n]->next->value,
				     ela[old_el_n]->index[bestn[0]]->next->value,
				     np->output_width)) :
		   percent_overlap(ela[el_n]->index[ex_n]->next->value,
				   ela[old_el_n]->index[bestn[0]]->next->value,
				   np->output_width)));
#endif DBM_NETWORK
	} /* best */
	fprintf(dout,") ");
      } /* phon best */
      fprintf(dout,"<%d %d> ", non_sem_farther, non_sem_count);
      /*
       * Semantic best
       */
      if (sem_layer_n) {
	fprintf(dout,"| ");
	for_each(b,n_best) {
	  PrintExampleName(ela[sem_el_n]->index[sem_bestn[b]]->next);
	  fprintf(dout,"(");
	  fprintf(dout,"%3.2f",sem_bestdist[b]);
	  if (zerop(b)) {
#ifdef DBM_NETWORK
	    if (Connected(net,sem_layer_n-2,sem_layer_n-1) || Connected(net,sem_layer_n-1,sem_layer_n))
	      fprintf(dout," %.1f",-(Energy(net,sem_layer_n-2,sem_layer_n-1)+Entropy(net,sem_layer_n-1)+Energy(net,sem_layer_n-1,sem_layer_n)));
	    if (Connected(net,sem_layer_n,sem_layer_n))
	      fprintf(dout," %.1f",-FreeEnergy(net,sem_layer_n,sem_layer_n));
	    if (Connected(net,sem_layer_n,sem_layer_n+1) || Connected(net,sem_layer_n+1,sem_layer_n+2))
	      fprintf(dout," %.1f",-(Energy(net,sem_layer_n,sem_layer_n+1)+Entropy(net,sem_layer_n+1)+Energy(net,sem_layer_n+1,sem_layer_n+2)));
# else
	    fprintf(dout," %2d",
		    (proximity ?
		     percent(cos_angle(ela[sem_el_n]->index[ex_n]->next->value,
				       ela[sem_el_n]->index[sem_bestn[0]]->next->value,
				       net->layer[sem_layer_n].n_units)) :
		     percent_overlap(ela[sem_el_n]->index[ex_n]->next->value,
				     ela[sem_el_n]->index[sem_bestn[0]]->next->value,
				     net->layer[sem_layer_n].n_units)));
#endif DBM_NETWORK
	  } /* best */
	  fprintf(dout,") ");
	} /* sem best */
      } /* sem layer */
      fprintf(dout,"\n"); fflush(dout);
    } /* print */
    if (stats && abs && (ex_n==n_examples/2-1)) {
      PrintStats(STATS_CON);
      ClearStats();
    }
  } /* examples */
  if (stats) {
    PrintStats((abs?STATS_ABS:FALSE));
  } else if (!example && items!=-1) {
    fprintf(dout,"(%d correct   %d errors   %d omissions)\n",
	    n_correct,n_response-n_correct,n_examples-n_response);
  }
  current_el = ela[old_el_n];
  np->input_noise = old_input_noise;
}


/*****************************************************************************
 * command_perform()
 */

int
  command_perform(tokc,tokv)
int tokc;
char *tokv[];
{
  int i,j;
  int lesion_rep,lesion_reps = 1;
  int hebb_random = 0, hebb_example = 0, hebb_rep, hebb_reps = 1;
  int el_n,old_el_n,sem = 0;
  int n_best = 3, n_hebb_words = 5;
  char **hebb_word;
  char *confusion_format = " %2d";
  int example = 0, errors = 0, omissions = 0, both = 0, items = 0, stats = FALSE, sum_index;
  int seed = 0, seed_specified = 0, noise = FALSE;
  int n_examples, example_n = -1;
  int *slayer, *dlayer;
  int slayer_specified = 0, dlayer_specified = 0, sunit_specified = 0, dunit_specified = 0;
  struct RANGE *sunit_range;
  struct RANGE *dunit_range;
  int lesion = 0;
  real lesion_p = 0.0;
  static struct OPTION options[] =
    {
      {OPTPRES|OPTINTARG,"reps",	0},
      {OPTPRES|OPTINTARG,"nbest",	0},
      {OPTPRES|OPTREALARG,"dist",   	0},
      {OPTPRES|OPTREALARG,"gap",   	0},
      {OPTPRES|OPTREALARG,"critv",   	0},
      {OPTPRES|OPTREALARG,"crits",   	0},
      {OPTPRES|OPTREALARG,"critp",   	0},
      {OPTPRES|OPTINTARG,"el",		0},
      {OPTPRES|OPTINTARG,"seed",	0},
      {OPTSIGN|OPTINTARG,"therapy",	0},
      {OPTSIGN|OPTINTARG,"split",	0},
      {OPTSIGN,          "details",	0},
      {OPTPRES,		 "proto",   	0},
      {OPTPRES|OPTINTARG,"sim",		0},
      {OPTPRES|OPTINTARG,"ncats",	0},
      {OPTPRES,		 "noise",   	0},
      {OPTPRES,		 "prob",   	0},
      {OPTPRES,		 "cosine",  	0},
      {OPTPRES,		 "hebb",	0},
      {OPTPRES,		 "hebbex",	0},
      {OPTPRES,		 "both",	0},
      {OPTPRES,		 "errors",	0},
      {OPTPRES,		 "omissions",	0},
      {OPTSIGN,		 "items",	0},
      {OPTSIGN,		 "confusion",	0},
      {OPTPRES,		 "ex",		0},
      {OPTPRES,          "slayer",	0},
      {OPTPRES,          "sunit",	0},
      {OPTPRES,          "dlayer",	0},
      {OPTPRES,          "dunit",	0},
      {OPTPRES,		 "abs",   	0},
      {OPTPRES,		 "stats",   	0},
      {OPTPRES|OPTREALARG,"p",		0},
      {OPTPRES|OPTINTARG,"smel",	0},
      {OPTPRES|OPTINTARG,"smlayer",	0},
      {OPTLAST,NULL,0}
    };

  IUsage("[%preps <int>] [%pnbest <n>] [%pex <n-or-comment>]\
 [%pdist <real>] [%pgap <real>] [%pcosine] [%pel [<int>]] [%pboth]\
 [%perrors] [%pomissions] [[%pslayer <range>] %pdlayer <range>]\
 [%pp <proportion>] [%phebb [<reps>]] [%phebbex <n-or-comment>]\
 [%pseed <int>] [%sitems] [%pprob [<th>]] [%pabs] [%pstats]\
 [%sconfusion] [%psmel <int>] [%psmlayer <int>] [%pnoise]\
 [%stherapy] [%sdetails] [%ssplit [<int>]] [%pproto] [%psim <n>]\
 [%pncats <n>] [%pcritv <real>] [%pcrits <real>] [%pcritp <real>]");
  if (GiveHelp(tokc)) {
    ISynopsis("lesion and run network on examples a number of times.");
    IHelp(IHelpArgs,
"For a specified number of repetitions, the network is run on each example",
"(or just one if \"example\" is specified) and its sum-squared-error is",
"printed.  The distance of this output is then compared with the desired",
"outputs of all other examples, and the <nbest> smallest distances are",
"printed if the best is less than <dist> and the difference between the",
"best and second best exceeds <gap>.  If <el> is specified, then the",
"network is run on examples from this example list, but the outputs are",
"still compared with the desired outputs of the current example list.  ",
"A lesion of the network may also be specified.",
"OPTIONS",
"%preps <int>     : Number of repetitions of lesion to perform.",
"%pnbest <int>    : Only show distances from <n> closest other examples.",
"                  Default is to show distances from every other example.",
"%pex <e>         : Summarize performance only for example e, which can",
"                  either be an integer or a comment.  Default is to",
"                  summarize every case.",
"%pdist <d>       : Only summarize cases with best distance less than <d>.",
"%pgap <d>        : Only summarize cases with second best distance farther",
"                  away than <g> from best distance.",
"%pprob [<th>]    : Use probability omission criterion of <th> (defaults",
"                  to \"dist\")",
"%pcosine         : Use cosine (proximity) rather than distance measure",
"%pel <int>       : Number of example list on which to run network.",
"%pboth           : Run on both current and specified example lists.",
"%sitems        : Print out item analysis.  No details on individual errors",
"                  is given is '-' is used.",
"%stherapy      : Administer therapy (training) for each lesion.",
"%sdetails      : Print details of therapy (default %y).",
"%ssplit <n>    : Split trained therapy set in half (default %n).  <n> sets",
"                  therapy by classes of size <n>.",
"%pproto          : Select therapy based on prototypicality.",
"%pncats <n>      : Number of semantic categories.",
"%psim <n>        : Select therapy based on type <n> similarities",
"                  (see therapy-bp.c for details).",
"%pcritv <real>   : Visual similarity requires proximity > <real>.",
"%pcrits <real>   : Semantic similarity requires proximity > <real>.",
"%pcritp <real>   : Phonological similarity requires proximity > <real>.",
"%pnoise          : Add noise with sd = <p> rather than lesion conns.",
"%perrors         : Only print errors (that meet dist and gap criteria).",
"%pomissions      : Force printing of omissions.",
"%pslayer <range> : Source layer(s) of links to lesion.",
"%pdlayer <range> : Destination layer(s) of links to lesion.",
"                  If slayer is not specified then units are lesioned.",
"%pp <real>       : Proportion of units/links to lesion (default 0.0).",
"                  If zero, all specified units/links are unlesioned.",
"%phebb [<n>]     : Perform lesion repetitions after setting hebb weights",
"                  based on each word in current example set.  This only",
"                  makes sense if np.hebb-scale is non-zero.  If <n> is",
"                  specified, perform lesions <n> times, chosing word for",
"                  setting hebb weights randomly.  This only makes sense",
"                  if np.hebb-damping and np.hebb-scale are non-zero.",
"%phebbex <e>     : Set hebb weights only for example e, which can",
"                  either be an integer or a comment.",
"%pseed <n>       : Seed random number generator with <n>.  This allows",
"                  runs to be repeated.",
"%pabs            : Split item display and print 'PAIR' errors" ,
"%pstats          : Generate output for statistical analysis" ,
"%sconfusion    : Generate confusion matrix (no details if %n)" ,
"%psmel <int>     : Example list that specifies semantics (for phon nets)" ,
"%psmlayer <int>  : Number of semantic layer (for phon nets)" ,
"",
"The command 'show np' will list out the current network parameters.",
"The command 'set np.hebb-damping = 0' will set hebb-damping to 0 (e.g.).",
NULL);
    return;
  }
  if (!NetExists(net))    { IError("current net is null");          return; }  
  if (current_el == NULL) { IError("current example set is NULL");  return; }
  /* Globals */
  dist = HUGE;  gap = -1;  prob = -1;
  proximity = FALSE;  abs = FALSE;  
  therapy = FALSE; therapy_proto = FALSE; therapy_sim = FALSE; 
  therapy_details = TRUE;  therapy_split = FALSE; therapy_split_size = NULL;
  vis_crit = sem_crit = phon_crit = FALSE; confusion = FALSE;
  phon = net->p->output_criteria;
  hebb = FALSE;  hebb_ex_n = -1;  hebb_processing_prime = FALSE;
  n_cats = N_CATEGORIES;
  srandom(time(0));
  seed=random();
  n_item_counts = N_ITEM_COUNTS;
  EnsureELIndex(current_el);
  if (net->p->hebb_damping==0.0) n_hebb_words = 1;
  n_examples  = current_el->n_examples;
  el_n = old_el_n = current_el->my_n;
  AllocDisplayRanges(net,&dlayer,&slayer,&dunit_range,&sunit_range);
  /*
   * Process options
   */
  fprintf(dout,"# %s\n",IComposeTokvLine(tokv,0));	/* echo command line */
  tokv++;
  while (IsOption(tokv[0])) {
    if        (OptionMatch(tokv[0],"reps",options,NULL)) {
      if (!IExpectInteger(tokv[1])) return;
      lesion_reps = atoi(*++tokv); tokc--;
    } else if (OptionMatch(tokv[0],"nbest",options,NULL)) {
      if (!IExpectInteger(tokv[1])) return;
      n_best = atoi(*++tokv); tokc--;
    } else if (OptionMatch(tokv[0],"ex",options,&example)) {
      example_n = (IIsInteger(tokv[1]) ? 
		   atoi(*++tokv) : 
		   ExampleNumber(current_el,*++tokv));
      tokc--;
    } else if (OptionMatch(tokv[0],"dist",options,NULL)) {
      if (!IExpectNumber(tokv[1])) return;
      dist = atof(*++tokv); tokc--;
    } else if (OptionMatch(tokv[0],"gap",options,NULL)) {
      if (!IExpectNumber(tokv[1])) return;
      gap = atof(*++tokv); tokc--;
    } else if (OptionMatch(tokv[0],"critv",options,NULL)) {
      if (!IExpectNumber(tokv[1])) return;
      vis_crit = atof(*++tokv); tokc--;
    } else if (OptionMatch(tokv[0],"crits",options,NULL)) {
      if (!IExpectNumber(tokv[1])) return;
      sem_crit = atof(*++tokv); tokc--;
    } else if (OptionMatch(tokv[0],"critp",options,NULL)) {
      if (!IExpectNumber(tokv[1])) return;
      phon_crit = atof(*++tokv); tokc--;
    } else if (OptionMatch(tokv[0],"noise",options,&noise)) {
    } else if (OptionMatch(tokv[0],"el",options,NULL)) {
      if (!IExpectInteger(tokv[1])) return;
      el_n = atoi(*++tokv); tokc--;
    } else if (OptionMatch(tokv[0],"seed",options,&seed_specified)) {
      if (!IExpectInteger(tokv[1])) return;
      seed = atoi(*++tokv); tokc--;
    } else if (OptionMatch(tokv[0],"abs",options,&abs)) {
    } else if (OptionMatch(tokv[0],"stats",options,&stats)) {
    } else if (OptionMatch(tokv[0],"both",options,&both)) {
    } else if (OptionMatch(tokv[0],"errors",options,&errors)) {
    } else if (OptionMatch(tokv[0],"omissions",options,&omissions)) {
    } else if (OptionMatch(tokv[0],"therapy",options,&therapy)) {
      if (IIsInteger(tokv[1])) {np->stop_at_epoch = atoi(*++tokv); tokc--;}
      np->therapy = TRUE;
      if (!therapy) therapy = -1;	/* will not reset training cases */
    } else if (OptionMatch(tokv[0],"details",options,&therapy_details)) {
    } else if (OptionMatch(tokv[0],"split",options,&therapy_split)) {
      if (IIsInteger(tokv[1])) {therapy_split_size = atoi(*++tokv); tokc--;}
    } else if (OptionMatch(tokv[0],"items",options,&items)) {
      if (!items) items = -1;
    } else if (OptionMatch(tokv[0],"confusion",options,&confusion)) {
      if (!confusion) confusion = -1;
    } else if (OptionMatch(tokv[0],"slayer",options,&slayer_specified) ||
	       OptionMatch(tokv[0],"dlayer",options,&dlayer_specified)	||
	       OptionMatch(tokv[0],"sunit", options,&sunit_specified)	||
	       OptionMatch(tokv[0],"dunit", options,&dunit_specified)) {
      if (!SetDisplayRanges(net,&tokv,dlayer,slayer,dunit_range,sunit_range)) return;
      stats_slayer = 0;		/* bias layer - designates lesioning units */
      for (i=0; i<net->n_layers; i++) {
	if (dlayer[i]) stats_dlayer = i;
	if (slayer_specified & slayer[i]) stats_slayer = i;
      }
      /* NOTE: p, prob, proto must be in this order due to overlap */
    } else if (OptionMatch(tokv[0],"p",options,&lesion)) {
      if (!IExpectNumber(tokv[1])) return;
      lesion_p = atof(*++tokv); tokc--;
      stats_severity = lesion_p;
    } else if (OptionMatch(tokv[0],"prob",options,NULL)) {
      if (IIsNumber(tokv[1])) {
	prob = atof(*++tokv); tokc--;
      }
    } else if (OptionMatch(tokv[0],"proto",options,&therapy_proto)) {
    } else if (OptionMatch(tokv[0],"cosine",options,&proximity)) {
    } else if (OptionMatch(tokv[0],"hebb",options,&hebb)) {
      if (IIsInteger(tokv[1])) {
	hebb_reps = atoi(*++tokv); tokc--;
	hebb_random = 1;
      } else {
	hebb_reps = n_examples;
      }
    } else if (OptionMatch(tokv[0],"hebbex",options,&hebb)) {
      hebb_reps = 1;
      hebb_example = 1;
      if (IIsInteger(tokv[1]))
	hebb_ex_n = atoi(*++tokv);
      else if ((hebb_ex_n=ExampleNumber(current_el,*++tokv))==-1) {
	IError("hebb example %d \"%s\" not found",hebb_ex_n,*tokv);
	return;
      }
      tokc--;
    } else if (OptionMatch(tokv[0],"smel",options,&sem)) {
      if (!IExpectInteger(tokv[1])) return;
      sem_el_n = atoi(*++tokv); tokc--;
      EnsureELIndex(ela[sem_el_n]);
    } else if (OptionMatch(tokv[0],"smlayer",options,&sem)) {
      if (!IExpectInteger(tokv[1])) return;
      sem_layer_n = atoi(*++tokv); tokc--;
      n_best = 2;
    } else if (OptionMatch(tokv[0],"sim",options,NULL)) {
      if (!IExpectInteger(tokv[1])) return;
      therapy_sim = atoi(*++tokv); tokc--;
    } else if (OptionMatch(tokv[0],"ncats",options,NULL)) {
      if (!IExpectInteger(tokv[1])) return;
      n_cats = atoi(*++tokv); tokc--;
    } else {
      IError("option not understood: \"%s\"",tokv[0]);
      return;
    }
    tokv++;
    tokc--;
  }
  srandom(seed);
  /*
   * Check parameters
   */
  if (slayer_specified && !dlayer_specified) {
    IError("cannot specify \"slayer\" without \"dlayer\"");
    return;
  }
  if (el_n<0 || el_n>=n_els) {
    IError("%d is not in the range of the current example array [%d-%d]",
	   el_n,0,n_els-1);
    return;
  }
  if (sem) {
    if (!ela[sem_el_n]) {
      IError("no sem example list %d",sem_el_n);
      return;
    }
    if (sem_layer_n >= net->n_layers) {
      IError("net does not have (sem) layer %d",sem_layer_n);
      return;
    }
  } else sem_layer_n = FALSE;	/* will be checked in PerformExamples */
  if (n_best < 2) {
    warn("perform","nbest = %d (must be >= 2), setting to 2",n_best);
    n_best = 2;
  }
  if (!lesion && lesion_reps > 1) warn("perform","Repeating runs without lesioning");
  if (el_n == current_el->my_n || el_n < 0) both = 0;
  if (hebb && np->hebb_scale==0) {
    warn("perform","hebb-scale is zero - not setting hebb weights");
    hebb = 0;
    hebb_reps = 1;
    n_cats = 4;
  }
  if (!hebb) {
    n_item_counts = 13;
    if (!abs)  n_item_counts -= 2;	/* no CROSS and PAIR */
    if (!phon) n_item_counts -= 2;	/* no PHON and PHON_MIXED */
  }
  if (items) {
    sum_index = ela[old_el_n]->n_examples+1;
    stim_count = ITempMatrix(int,0,sum_index,0,N_ITEM_COUNTS);
    resp_count = ITempMatrix(int,0,sum_index,0,N_ITEM_COUNTS);
    for (i=0; i<sum_index; i++)
      for (j=0; j<n_item_counts; j++)
	stim_count[i][j] = resp_count[i][j] = 0;
  }
  if (confusion) {
    confusion_count = ITempMatrix(int,0,ela[old_el_n]->n_examples,0,ela[old_el_n]->n_examples);
    for_each (i,ela[old_el_n]->n_examples) 
      for_each (j,ela[old_el_n]->n_examples) 
	confusion_count[i][j] = 0;
    if (hebb) {
      hebb_confusion_count = ITempMatrix(int,0,ela[old_el_n]->n_examples,0,ela[old_el_n]->n_examples);
      for_each (i,ela[old_el_n]->n_examples) 
	for_each (j,ela[old_el_n]->n_examples) 
	  hebb_confusion_count[i][j] = 0;
    }
  }
  fprintf(dout,"# Network %s\n",net->name);
  if (hebb) {
    fprintf(dout,"# hebb-scale = %f   hebb-damping = %f\n",np->hebb_scale,np->hebb_damping);
    hebb_word = ITempCalloc(char*,n_hebb_words);
    ResetHebbWeights(net);
    hebb_prime_count = ITempMatrix(int,0,sum_index,0,N_ITEM_COUNTS);
    for (i=0; i<sum_index; i++)
      for (j=0; j<n_item_counts; j++) 
	hebb_prime_count[i][j] = 0;
  }
  if (lesion) UnLesionNet(net);
  if (noise) SavePerformWeights(net);
  /*
   * Hebb repetitiions
   */
  for (hebb_rep=1; hebb_rep<=hebb_reps; hebb_rep++) {
    /* start hebb_rep at 0 to run lesion reps with no hebb context */
    if (hebb) {
      fprintf(dout,"************************* (%d/%d) HEBB context: ",hebb_rep,hebb_reps);
      fflush(dout);
      if (hebb_rep==0) {
	for (i=0; i<n_hebb_words; i++) hebb_word[i] = NULL;
	fprintf(dout,"[none]");
      } else {
	if (!hebb_example) {
	  if (hebb_random) hebb_ex_n = RandInt(n_examples);
	  else		   hebb_ex_n = hebb_rep-1;
	}
	for (i=n_hebb_words-1; i>0; i--) hebb_word[i] = hebb_word[i-1];
	hebb_word[0] = current_el->index[hebb_ex_n]->comment;
	for (i=0; i<n_hebb_words; i++) if (hebb_word[i]) fprintf(dout,"%s ",hebb_word[i]);
      }
      fprintf(dout,"\n");
    } /* hebb */
    /*
     * Lesion repetitions
     */
    for (lesion_rep=1; lesion_rep<=lesion_reps; lesion_rep++) {
      if (lesion) {
	srandom(seed);
	fprintf(dout,"========================= (%d/%d) [%d] ",lesion_rep,lesion_reps,seed);
	srandom(seed=random());
	if (noise) {
	  ApplyToEachLinkInRange
	    (slayer,sunit_range,dlayer,dunit_range,
	     LINK_WEIGHT(link) += GaussNoise(lesion_p);
	     );
	  fprintf(dout,"Noise ");
	  DescribeLesion(net,slayer,dlayer,sunit_range,dunit_range,lesion_p,slayer_specified,dlayer_specified,0);
	} else {
	  DescribeLesion(net,slayer,dlayer,sunit_range,dunit_range,lesion_p,slayer_specified,dlayer_specified,
			 LesionNet(net,slayer,dlayer,sunit_range,dunit_range,lesion_p,!slayer_specified));
	}
	fflush(dout);
      }
      /*
       * Run examples
       */
      if (hebb && hebb_rep != 0) {
	fprintf(dout,"h");
	hebb_prime = hebb_ex_n;
	hebb_processing_prime = TRUE;
	PerformExamples(net,current_el->my_n,n_best,hebb_ex_n,FALSE,FALSE,FALSE,FALSE);
	hebb_processing_prime = FALSE;
	UpdateHebbWeights(net);
      }
      if (both) {
	fprintf(dout,"---------- Example list %d (current)\n",current_el->my_n);
	PerformExamples(net,current_el->my_n,n_best,example_n,errors,omissions,FALSE,stats);
	fprintf(dout,"---------- Example list %d\n",el_n);
	fflush(dout);
      }
      PerformExamples(net,el_n,n_best,example_n,errors,omissions,items,stats);
      if (therapy) PerformTherapy(net);
      if (lesion) UnLesionNet(net);
      if (noise) RestorePerformWeights(net);
    } /* end lesion reps */
  } /* end hebb reps */
  if (items) {
    /*
     * Stimulus analysis
     */
    fprintf(dout,"Stimulus Analysis (n = %d):\n",lesion_reps*hebb_reps);
    PrintItemHeaders();
    if (abs) {
      /* CONCRETE WORDS */
      for (i=0; i<ela[old_el_n]->n_examples/2; i++) {
	PrintExampleName(ela[old_el_n]->index[i]);
	fprintf(dout,":\t");
	for (j=0; j<n_item_counts; j++) {
	  fprintf(dout,"%6d",stim_count[i][j]);
	  stim_count[sum_index][j] += stim_count[i][j];
	}
	fprintf(dout,"\n");
      }
      PrintItemHeaders();
      fprintf(dout,"TOTALS:\t");
      for (j=0; j<n_item_counts; j++) {
	fprintf(dout,"%6d",stim_count[sum_index][j]);
	stim_count[0][j] = 0;		/* [0] used for abs subtotals */
      }
      fprintf(dout,"\n");
      /* ABSTRACT WORDS */
      for (i=ela[old_el_n]->n_examples/2; i<sum_index-1; i++) {
	PrintExampleName(ela[old_el_n]->index[i]);
	fprintf(dout,":\t");
	for (j=0; j<n_item_counts; j++) {
	  fprintf(dout,"%6d",stim_count[i][j]);
	  stim_count[0][j] += stim_count[i][j];		/* abs subtotals */
	}
	fprintf(dout,"\n");
      }
      PrintItemHeaders();
      fprintf(dout,"TOTALS:\t");
      for (j=0; j<n_item_counts; j++) {
	fprintf(dout,"%6d",stim_count[0][j]);
	stim_count[sum_index][j] += stim_count[0][j];
      }
      fprintf(dout,"\n");
    } else {
      for (i=0; i<sum_index-1; i++) {
	PrintExampleName(ela[old_el_n]->index[i]);
	fprintf(dout,":\t");
	for (j=0; j<n_item_counts; j++) {
	  fprintf(dout,"%6d",stim_count[i][j]);
	  stim_count[sum_index][j] += stim_count[i][j];
	}
	fprintf(dout,"\n");
      }
    } /* abs */
    /*
     * Response analysis
     */
    fprintf(dout,"Response Analysis (n = %d):\n",lesion_reps*hebb_reps);
    PrintItemHeaders();
    for_each (i,sum_index-1) {
      PrintExampleName(ela[old_el_n]->index[i]);
      fprintf(dout,":\t");
      for_each (j,n_item_counts) 
	fprintf(dout,"%6d",resp_count[i][j]);
      fprintf(dout,"\n");
    }
    /*
     * Prime analysis (hebb)
     */
    if (hebb) {
      fprintf(dout,"Prime Analysis (n = %d):\n",lesion_reps*hebb_reps);
      PrintItemHeaders();
      for_each (i,sum_index-1) {
	PrintExampleName(ela[old_el_n]->index[i]);
	fprintf(dout,":\t");
	for_each (j,n_item_counts) 
	  fprintf(dout,"%6d",hebb_prime_count[i][j]);
	fprintf(dout,"\n");
      }
    } /* hebb */
    /*
     * Grand totals
     */
    PrintItemHeaders();
    fprintf(dout,"TOTALS:\t");
    for (j=0; j<n_item_counts; j++) 
      fprintf(dout,"%6d",stim_count[sum_index][j]);
    fprintf(dout,"\n");
  } /* item */
  if (confusion) {
    fprintf(dout,"Stimulus-Response Confusion Matrix (n = %d):\n",lesion_reps*hebb_reps);
    if ((lesion_reps*hebb_reps)>=100) confusion_format = " %3d";
    for_each (i,current_el->n_examples) {
      PrintExampleName(current_el->index[i]);
      fprintf(dout,"\t");
      for_each (j,current_el->n_examples) fprintf(dout,confusion_format,confusion_count[i][j]);
      fprintf(dout,"\n");
    }
    if (hebb) {
      fprintf(dout,"Prime-Response Confusion Matrix (n = %d):\n",lesion_reps*hebb_reps);
      for_each (i,current_el->n_examples) {
	PrintExampleName(current_el->index[i]);
	fprintf(dout,"\t");
	for_each (j,current_el->n_examples) fprintf(dout,confusion_format,hebb_confusion_count[i][j]);
	fprintf(dout,"\n");
      }
    }
  } /* confusion */
}


/*****************************************************************************
 * command_proximity()
 */

int
  command_proximity(tokc,tokv)
int tokc;
char *tokv[];
{
  int e,e2;
  int n_examples,input_width,output_width;
  real val, in_mean=0.0, out_mean=0.0;
  IUsage("");
  if (GiveHelp(tokc)) {
    ISynopsis("print proximity matrix for current example list.");
    IHelp(IHelpArgs,
"Print out a proximity matrix of the input patterns (lower-left)",
"and output patterns (upper-right) of the current example list.",
NULL);
    return;
  }
  if (current_el == NULL) {
    IError("current example set is NULL");
    return;
  }
  EnsureELIndex(current_el);
  n_examples   = current_el->n_examples;
  input_width  = current_el->index[0]->n_values;
  output_width = current_el->index[0]->next->n_values;
  fprintf(dout,"\n\\ O\t");
  for (e=0; e<n_examples; e++) fprintf(dout,"%2d ",e);
  fprintf(dout,"\nI \\\t");
  for (e=0; e<n_examples; e++) fprintf(dout,"---");
  fprintf(dout,"\n");
  for (e=0; e<n_examples; e++) {
    PrintExampleName(current_el->index[e]);
    if (current_el->index[e]->comment) fprintf(dout,"%d",e);
    fprintf(dout,"\t");
    for (e2=0; e2<n_examples; e2++) {
      if (e>e2) {
	val = cos_angle(current_el->index[e]->value,
			current_el->index[e2]->value,
			input_width);
	in_mean += val;
	fprintf(dout,"%2d ",percent(val));
      } else if (e<e2) {
	val = cos_angle(current_el->index[e]->next->value,
			current_el->index[e2]->next->value,
			output_width);
	out_mean += val;
	fprintf(dout,"%3d",percent(val));
      } else fprintf(dout,"--");
    }
    fprintf(dout,"\n");
  }
  val = 2.0/(n_examples*(n_examples-1.0));
  fprintf(dout,"Input Mean  = %f\nOutput Mean = %f\n",
	  in_mean*val, out_mean*val);
}


/*****************************************************************************
 * command_error_types()
 */

int
  command_error_types(tokc,tokv)
int tokc;
char *tokv[];
{
  int n,stim,resp,prime,vis,sem;
  int err_type,hebb_PR, hebb_PS;
  int n_examples,n_counts;
  int count[N_ITEM_COUNTS],total[N_ITEM_COUNTS];
  static struct OPTION options[] =
    {
      {OPTPRES,		 "abs",   	0},
      {OPTPRES,		 "hebb",   	0},
      {OPTPRES|OPTREALARG,"critv",   	0},
      {OPTPRES|OPTREALARG,"crits",   	0},
      {OPTPRES|OPTREALARG,"critp",   	0},
      {OPTPRES|OPTINTARG,"ncats",	0},
      {OPTPRES|OPTINTARG,"smel",	0},
      {OPTPRES|OPTINTARG,"smlayer",	0},
      {OPTLAST,NULL,0}
    };
  IUsage("[%pabs] [%phebb] [%psmel <int>] [%pncats <int>] [%psmlayer <int>]\
 [%pcritv <real>] [%pcrits <real>] [%pcritp <real>]");
  if (GiveHelp(tokc)) {
    ISynopsis("print possible error types for current example list.");
    IHelp(IHelpArgs,
"Print out a matrix of the error type for each pair of inputs.",
"OPTIONS",
"%pabs            : Split item display and print 'PAIR' errors" ,
"%phebb           : Print hebb error types" ,
"%pncats <n>      : Number of semantic categories.",
"%psmel <int>     : Example list that specifies semantics (for phon nets)" ,
"%psmlayer <int>  : Number of semantic layer (for phon nets)" ,
"%pcritv <real>   : Visual similarity requires proximity > <real>.",
"%pcrits <real>   : Semantic similarity requires proximity > <real>.",
"%pcritp <real>   : Phonological similarity requires proximity > <real>.",
NULL);
    return;
  }
  if (current_el == NULL) {
    IError("current example set is NULL");
    return;
  }
  EnsureELIndex(current_el);
  n_examples   = current_el->n_examples;
  abs = FALSE;  sem_layer_n = FALSE;  phon = FALSE;  hebb = FALSE;
  vis_crit = sem_crit = phon_crit = FALSE;
  n_cats = N_CATEGORIES;
  tokv++;
  while (IsOption(tokv[0])) {
    if (OptionMatch(tokv[0],"abs",options,&abs)) {
    } else if (OptionMatch(tokv[0],"hebb",options,&hebb)) {
    } else if (OptionMatch(tokv[0],"smel",options,&sem)) {
      if (!IExpectInteger(tokv[1])) return;
      sem_el_n = atoi(*++tokv); tokc--;
      if (ela[sem_el_n] == NULL) {IError("example set %d is NULL",sem_el_n); return;      }
      phon = TRUE;
      EnsureELIndex(ela[sem_el_n]);
    } else if (OptionMatch(tokv[0],"ncats",options,NULL)) {
      if (!IExpectInteger(tokv[1])) return;
      n_cats = atoi(*++tokv); tokc--;
    } else if (OptionMatch(tokv[0],"smlayer",options,&sem)) {
      if (!IExpectInteger(tokv[1])) return;
      sem_layer_n = atoi(*++tokv); tokc--;
      if (!NetExists(net)) {IError("current net is null"); return; }  
      phon = TRUE;
    } else if (OptionMatch(tokv[0],"critv",options,NULL)) {
      if (!IExpectNumber(tokv[1])) return;
      vis_crit = atof(*++tokv); tokc--;
    } else if (OptionMatch(tokv[0],"crits",options,NULL)) {
      if (!IExpectNumber(tokv[1])) return;
      sem_crit = atof(*++tokv); tokc--;
    } else if (OptionMatch(tokv[0],"critp",options,NULL)) {
      if (!IExpectNumber(tokv[1])) return;
      phon_crit = atof(*++tokv); tokc--;
    } else {
      IError("option not understood: \"%s\"",tokv[0]);
      return;
    }
    tokv++;
    tokc--;
  }
  /*
   * print header 
   */
  fprintf(dout,"\n\\ O\t");
  for (stim=0; stim<n_examples; stim++) fprintf(dout,"%2d ",stim);
  fprintf(dout,"|");
  n_counts = (hebb?N_ITEM_COUNTS:ITEM_CROSS);
  for (n=ITEM_VISUAL; n<n_counts; n++) 
    fprintf(dout,"%6s",(hebb?hebb_item_header[n]:item_header[n]));
  fprintf(dout,"\nI \\\t");
  for (stim=0; stim<n_examples; stim++) fprintf(dout,"---");
  fprintf(dout,"|");
  for (n=ITEM_VISUAL; n<n_counts; n++) fprintf(dout,"------");
  fprintf(dout,"\n");
  for (n=ITEM_VISUAL; n<n_counts; n++) total[n] = 0;
  for (stim=0; stim<n_examples; stim++) {		/* STIMULUS */
    PrintExampleName(current_el->index[stim]);
    if (current_el->index[stim]->comment) fprintf(dout,"%d",stim);
    fprintf(dout,"\t");
    for (n=ITEM_VISUAL; n<n_counts; n++) count[n] = 0;
    for (resp=0; resp<n_examples; resp++) {		/* RESPONSE */
      if (stim==resp)	/* SR correct */
	fprintf(dout," . ");
      else {		/* SR error */
	fprintf(dout," ");
	count[err_type=ErrorType(current_el->index[stim],current_el->index[resp],TRUE)]++;
	fprintf(dout," ");
	if (hebb) 
	  for (prime=0; prime<n_examples; prime++) {	/* PRIME */
	    hebb_PS = (stim==prime?ITEM_CORRECT:ErrorType(current_el->index[prime],current_el->index[stim],FALSE));
	    hebb_PR = (resp==prime?ITEM_CORRECT:ErrorType(current_el->index[prime],current_el->index[resp],FALSE));
	    count[HebbErrorType(err_type,hebb_PS,hebb_PR)]++;
	  } /* prime */
      } /* error */
    } /* response */
    fprintf(dout,"|");
    for (n=ITEM_VISUAL; n<n_counts; n++) {
      fprintf(dout,"%6d",count[n]);
      total[n] += count[n];
    }
    fprintf(dout,"\n");
  } /* stimulus */
  /*
   * print totals
   */
  for (n=ITEM_VISUAL; n<n_counts; n++) 
    fprintf(dout,"%6s",(hebb?hebb_item_header[n]:item_header[n]));
  fprintf(dout,"\n");
  for (n=ITEM_VISUAL; n<n_counts; n++) fprintf(dout," %5d",total[n]);
  fprintf(dout,"\n");
}


/*****************************************************************************
 * command_error_changes()
 */

int
  command_error_changes(tokc,tokv)
int tokc;
char *tokv[];
{
  int n,e1,e2,n_examples,n_pairs,c;
  real *vis_w, *vis_wo, *sem_w, *sem_wo;
  real old_vis_crit,old_sem_crit,old_phon_crit,mean;
  static struct OPTION options[] =
    {
      {OPTPRES,		 "abs",   	0},
      {OPTPRES|OPTINTARG,"ncats",	0},
      {OPTPRES|OPTREALARG,"critv",   	0},
      {OPTPRES|OPTREALARG,"crits",   	0},
      {OPTPRES|OPTREALARG,"critp",   	0},
      {OPTPRES|OPTINTARG,"smel",	0},
      {OPTPRES|OPTINTARG,"smlayer",	0},
      {OPTLAST,NULL,0}
    };
  IUsage("[%pabs] [%psmel <int>] [%pncats <int>] [%psmlayer <int>]\
 [%pcritv <real>] [%pcrits <real>] [%pcritp <real>]");
  if (GiveHelp(tokc)) {
    ISynopsis("Count changes in error type due to use of criteria.");
    IHelp(IHelpArgs,
"OPTIONS",
"%pabs            : Split item display and print 'PAIR' errors" ,
"%pncats <n>      : Number of semantic categories.",
"%psmel <int>     : Example list that specifies semantics (for phon nets)" ,
"%psmlayer <int>  : Number of semantic layer (for phon nets)" ,
"%pcritv <real>   : Visual similarity requires proximity > <real>.",
"%pcrits <real>   : Semantic similarity requires proximity > <real>.",
"%pcritp <real>   : Phonological similarity requires proximity > <real>.",
NULL);
    return;
  }
  if (current_el == NULL) {
    IError("current example set is NULL");
    return;
  }
  EnsureELIndex(current_el);
  n_examples   = current_el->n_examples;
  n_pairs = n_examples*(n_examples-1)/2;
  vis_w  = ITempCalloc(real,n_pairs);
  vis_wo = ITempCalloc(real,n_pairs);
  sem_w  = ITempCalloc(real,n_pairs);
  sem_wo = ITempCalloc(real,n_pairs);
  n_cats = N_CATEGORIES;
  abs = FALSE;  sem_layer_n = FALSE;  phon = FALSE;
  vis_crit = sem_crit = phon_crit = FALSE;
  tokv++;
  while (IsOption(tokv[0])) {
    if (OptionMatch(tokv[0],"abs",options,&abs)) {
    } else if (OptionMatch(tokv[0],"ncats",options,NULL)) {
      if (!IExpectInteger(tokv[1])) return;
      n_cats = atoi(*++tokv); tokc--;
     } else if (OptionMatch(tokv[0],"smel",options,NULL)) {
      if (!IExpectInteger(tokv[1])) return;
      sem_el_n = atoi(*++tokv); tokc--;
      if (ela[sem_el_n] == NULL) {
	IError("example set %d is NULL",sem_el_n);
	return;
      }
      phon = TRUE;
      EnsureELIndex(ela[sem_el_n]);
    } else if (OptionMatch(tokv[0],"smlayer",options,NULL)) {
      if (!IExpectInteger(tokv[1])) return;
      sem_layer_n = atoi(*++tokv); tokc--;
      if (!NetExists(net)) {
	IError("current net is null");
	return;
      }  
      phon = TRUE;
    } else if (OptionMatch(tokv[0],"critv",options,NULL)) {
      if (!IExpectNumber(tokv[1])) return;
      vis_crit = atof(*++tokv); tokc--;
    } else if (OptionMatch(tokv[0],"crits",options,NULL)) {
      if (!IExpectNumber(tokv[1])) return;
      sem_crit = atof(*++tokv); tokc--;
    } else if (OptionMatch(tokv[0],"critp",options,NULL)) {
      if (!IExpectNumber(tokv[1])) return;
      phon_crit = atof(*++tokv); tokc--;
    } else {
      IError("option not understood: \"%s\"",tokv[0]);
      return;
    }
    tokv++;
    tokc--;
  }
  n = 0;
  for_each(e1,n_examples) 
    for (e2=0; e2<e1; e2++) {
      vis_w[n] = (VisuallySimilar(current_el->index[e1],current_el->index[e2]) != 0);
      sem_w[n] = (SemanticallySimilar(current_el->index[e1],current_el->index[e2]) != 0);
      old_vis_crit  = vis_crit;  old_sem_crit  = sem_crit;  old_phon_crit = phon_crit;
      vis_crit = sem_crit = phon_crit = FALSE;
      vis_wo[n] = (VisuallySimilar(current_el->index[e1],current_el->index[e2]) != 0);
      sem_wo[n] = (SemanticallySimilar(current_el->index[e1],current_el->index[e2]) != 0);
      vis_crit  = old_vis_crit;  sem_crit  = old_sem_crit;  phon_crit = old_phon_crit;
      n++;
    }
  fprintf(dout,"%d pairs\n",n_pairs);
  fprintf(dout,"Visual:  ");
  for (n=0,c=0; n<n_pairs; n++) c += vis_w[n];
  fprintf(dout,"  nwi=%d", c);
  for (n=0,c=0; n<n_pairs; n++) c += vis_wo[n];
  fprintf(dout,"  nwo=%d", c);
  for (n=0,c=0; n<n_pairs; n++) if (vis_w[n] != vis_wo[n]) c++;
  fprintf(dout,"  ham=%d", c);
  fprintf(dout,"  cor=%f\n", correlation(vis_w,vis_wo,n_pairs));
  fprintf(dout,"Semantic:");
  for (n=0,c=0; n<n_pairs; n++) c += sem_w[n];
  fprintf(dout,"  nwi=%d", c);
  for (n=0,c=0; n<n_pairs; n++) c += sem_wo[n];
  fprintf(dout,"  nwo=%d", c);
  for (n=0,c=0; n<n_pairs; n++) if (sem_w[n] != sem_wo[n]) c++;
  fprintf(dout,"  ham=%d", c);
  fprintf(dout,"  cor=%f\n", correlation(sem_w,sem_wo,n_pairs));
}
