#include "utils-bp.h"

#define POSITIVE 		1
#define NEGATIVE 	       -1

#define SHOW_PRINT		1
#define SHOW_DISPLAY 		2

extern real cross_entropy();		/* error-bp.c */
extern real tolerance_cross_entropy();	/* error-bp.c */

#define d_sigmoid(y)		(y)*(1.0-(y))
#define i_sigmoid(y)							\
 (zerop(y) 								\
  ? log(MIN_Y/(1.0-MIN_Y)) 						\
  : (onep(y) 								\
     ? log((1.0-MIN_Y)/MIN_Y) 						\
     : log((y)/(1.0-(y)))))

#define symmetric_sigmoid(x)	(2.0/(1.0+exp(-(x))) - 1.0)
#define d_symmetric_sigmoid(y)	((1.0+(y))*(1.0-(y))*0.25)

#define SQUASH(x)	(np->unit_function ? symmetric_sigmoid(x)   : sigmoid(x))
#define DSQUASH(y)	(np->unit_function ? d_symmetric_sigmoid(y) : d_sigmoid(y))
#define ISQUASH(y)	i_sigmoid(InvScaleOutput(y))

#define ShrinkOutput(y,f) 						\
  (np->unit_function ? ((y)*(f)) : (f*((y)-0.5)+0.5))

#define UnitLowerBound		(np->unit_function ? -1.0 : 0.0)
#define UnitNeutral		(np->unit_function ? 0.0 : 0.5)
#define UnitUpperBound		1.0

#define UNIT_BINARY_OUTPUT(unit) \
  (UNIT_OUTPUT(unit) > UnitNeutral ? UnitUpperBound : UnitLowerBound)

#define UNIT_DERROR_FN(y,d)						\
  (np->error_function							\
   ? tolerance_cross_entropy(y,d)					\
   : cross_entropy(y,d))

#define UNIT_ERROR(unit,t) sqr(UNIT_TARGET_T(unit,t)-UNIT_OUTPUT_T(unit,t))

#define UNIT_DERROR(unit,t) 						\
  UNIT_DERROR_FN(ScaleOutput(UNIT_OUTPUT_T(unit,t)),			\
		 ScaleOutput(UNIT_TARGET_T(unit,t)))

#define IsBadUnit(unit,t)						\
  (UNIT_TARGET_T(unit,t)==UnitLowerBound				\
   ? (UNIT_OUTPUT_T(unit,t) > (UnitLowerBound+np->tolerance))		\
   : (UNIT_TARGET_T(unit,t)==1.0						\
      ? (UNIT_OUTPUT_T(unit,t) < 1.0-np->tolerance)			\
      : (fabs(UNIT_OUTPUT_T(unit,t)-UNIT_TARGET_T(unit,t)) > np->midrange_tolerance)))


/******************************************************************************
 * TESTS
 */

#define IsInputLayer		(layer_n==net->p->input_layer_n)
#define IsOtherInputLayer	(OtherLayerN==net->p->input_layer_n)

#define IsOutputLayer		(layer_n==net->p->output_layer_n)
#define IsOtherOutputLayer	(OtherLayerN==net->p->output_layer_n)

#define IsBiasLayer		(layer_n==net->p->bias_layer_n)
#define IsOtherBiasLayer	(OtherLayerN==net->p->bias_layer_n)

#define IsTrainingIteration(t)	(t > (net->p->max_t-net->p->n_train))
#define IsInputIteration(t)	(!net->p->max_input_t || (t) <= net->p->max_input_t)
#define UpdatingInputStates	(net->p->input_layer_n>=net->p->forward_layer_n)

/******************************************************************************
 * ITERATION
 */

#define IterateForward(t)\
  for (t=1; t<=net->p->max_t; t++) 

#define IterateBackward(t)\
  for (t=net->p->max_t; t>0; t--)


/******************************************************************************
 * TRAVERSAL
 */

#ifdef DBM_NETWORK

#define ApplyToEachInputUnit(P) \
  ForEachLayer { \
    if (layer->p->is_input_layer) \
      ForEachUnit(UNIT) { \
        P \
      } EndForEach; \
  } EndForEach;

#define ApplyToEachOutputUnit(P) \
  ForEachLayer { \
    if (layer->p->is_output_layer) \
      ForEachUnit(UNIT) { \
        P \
      } EndForEach; \
  } EndForEach;

#else

#define ApplyToEachInputUnit(P) \
  DoLayer(net->p->input_layer_n) { \
    ForEachUnit(UNIT) { \
      P \
    } EndForEach; \
  } EndDo;

#define ApplyToEachOutputUnit(P) \
  DoLayer(net->p->output_layer_n) { \
    ForEachUnit(UNIT) { \
      P \
    } EndForEach; \
  } EndDo;

#endif DBM_NETWORK

#define ForEachForwardLayer \
  {\
    LAYER	*layer;\
    int		layer_n;\
    for ((layer_n=net->p->forward_layer_n),\
	 (layer=net->layer+layer_n);\
	 layer_n<=net->n_layers-1;\
	 (layer_n++,layer++)) {


#ifdef DBM_NETWORK
#define ForEachBackwardLayer \
  {\
    LAYER	*layer;\
    int		layer_n;\
    for ((layer_n=net->p->output_layer_n-1),\
	 (layer=net->layer+layer_n);\
	 layer_n>net->p->backward_layer_n;\
	 (layer_n--,layer--)) {
#else  DBM_NETWORK
#define ForEachBackwardLayer \
  {\
    LAYER	*layer;\
    int		layer_n;\
    for ((layer_n=net->n_layers-1),\
	 (layer=net->layer+layer_n);\
	 layer_n>net->p->backward_layer_n;\
	 (layer_n--,layer--)) {
#endif DBM_NETWORK


#define ForEachForwardUnit(UNIT_TYPE) \
  ForEachForwardLayer\
    UNIT_INFO	*info;\
    UNIT_TYPE	*unit;\
    int		unit_n;\
    assertm(layer->unit_tag==TAG(UNIT_TYPE),"bad unit type");\
    for ((unit_n=0),(unit=CONCAT(UNIT_PTR_,UNIT_TYPE)(layer->unit)),\
	            (info=layer->info);\
	 unit_n<layer->n_units;\
	 unit_n++,unit++,info++) 


#define ForEachBackwardUnit(UNIT_TYPE)\
  ForEachBackwardLayer\
      UNIT_INFO	*info;\
      UNIT_TYPE	*unit;\
      int		unit_n;\
      assertm(layer->unit_tag==TAG(UNIT_TYPE),"bad unit type");\
      for ((unit_n=0),(unit=CONCAT(UNIT_PTR_,UNIT_TYPE)(layer->unit)),\
		      (info=layer->info);\
	   unit_n<layer->n_units;\
	   unit_n++,unit++,info++)


/******************************************************************************
 * COMPILED TRAVERSAL
 */

#define ForEachUnitCompiled(UNIT_TYPE)			\
  {							\
    int unit_n;						\
    UNIT_TYPE  *unit;					\
    UNIT_TYPE* *units = (UNIT_TYPE**)net->p->units;	\
    for (unit_n=0; unit_n<net->p->n_units; unit_n++) {	\
      unit = units[unit_n];

#define ForEachLinkCompiled(LINK_TYPE,UNIT_TYPE)	\
  {							\
    int link_n;						\
    LINK_TYPE  *link;					\
    UNIT_TYPE  *s_unit;					\
    UNIT_TYPE  *d_unit;					\
    LINK_TYPE* *links   = (LINK_TYPE**)net->p->links;	\
    UNIT_TYPE* *s_units = (UNIT_TYPE**)net->p->s_units;	\
    UNIT_TYPE* *d_units = (UNIT_TYPE**)net->p->d_units;	\
    for (link_n=0; link_n<net->p->n_links; link_n++) {	\
      link   = links[link_n];				\
      s_unit = s_units[link_n];				\
      d_unit = d_units[link_n];



#define ForEachBackwardUnitCompiled(UNIT_TYPE)		\
  {							\
    int unit_n;						\
    UNIT_TYPE  *unit;					\
    UNIT_TYPE* *units = (UNIT_TYPE**)net->p->bunits;	\
    for (unit_n=0; unit_n<net->p->n_bunits; unit_n++) {	\
      unit = units[unit_n];

#ifdef DBM_NETWORK
#define ForEachBackwardLinkCompiled(LINK_TYPE,UNIT_TYPE)\
  ForEachLinkCompiled(LINK_TYPE,UNIT_TYPE)
#else
#define ForEachBackwardLinkCompiled(LINK_TYPE,UNIT_TYPE)\
  {							\
    int link_n;						\
    LINK_TYPE  *link;					\
    UNIT_TYPE  *s_unit;					\
    UNIT_TYPE  *d_unit;					\
    LINK_TYPE* *links   = (LINK_TYPE**)net->p->blinks;	\
    UNIT_TYPE* *s_units = (UNIT_TYPE**)net->p->s_bunits;\
    UNIT_TYPE* *d_units = (UNIT_TYPE**)net->p->d_bunits;\
    for (link_n=0; link_n<net->p->n_blinks; link_n++) {	\
      link   = links[link_n];				\
      s_unit = s_units[link_n];				\
      d_unit = d_units[link_n];
#endif DBM_NETWORK



#define ApplyToEachLinkInRange(slayer,sunit_range,dlayer,dunit_range,P)	\
  {									\
    int dl, du;								\
    for (dl=0; dl<net->n_layers; dl++)					\
      if (dlayer[dl]) 							\
	for (du=dunit_range[dl].lo; du<=dunit_range[dl].hi; du++) 	\
	  DoLayerUnit(dl,UNIT,du) {					\
	    ForEachLinkSet(LINK,UNIT) {					\
	      if (slayer[OtherLayerN]) 					\
		ApplyToLinks						\
		  (							\
		   if (   OtherUnitN >= sunit_range[OtherLayerN].lo	\
		       && OtherUnitN <= sunit_range[OtherLayerN].hi) {	\
		     P							\
		     }							\
		   );							\
	    } EndForEach;						\
	  } EndDo;							\
  }
