#include <itf/option.h>
#include <itf/itf.h>
#include <sim/sim.h>
#include <lib/example.h>


#define DB(x)

int saveprec = 6;

autoinit_save_weights()
{
  /* just so this gets pulled from the library by the loader */
}


int SaveLinks(net)
NETWORK* net;
{
  fprintf(dout," %d links %s .momentum .gain\n",net->n_links,WEIGHT_NAME);
  ApplyToLinksOfNet
    (UNIT,LINK,UNIT,
     fprintf(dout," %d %d %d %d %.*g %.3g %.3g\n",
	     OtherLayerN,OtherUnitN,layer_n,unit_n,
	     saveprec,LINK_WEIGHT(link),link->momentum,link->gain);
     );
  return 0;
}


int TranslateLinks(net,layermap)
NETWORK *net;
int *layermap;
{
  fprintf(dout," %d links %s .momentum .gain\n",net->n_links,WEIGHT_NAME);
  ApplyToLinksOfNet
    (UNIT,LINK,UNIT,
     fprintf(dout," %d %d %d %d %.*g %.3g %.3g\n",
	     layermap[OtherLayerN],OtherUnitN,layermap[layer_n],unit_n,
	     saveprec,LINK_WEIGHT(link),link->momentum,link->gain);
     );
  return 0;
}


LINK_PTR GetLink(net,sl,su,dl,du)
NETWORK *net;
int sl,su,dl,du;
{
  DoLayerUnit(dl,UNIT,du) {
    ForEachLinkSet(LINK,UNIT) {
      if (OtherLayerN == sl)
	ApplyToEachLink(if (OtherUnitN == su) return link;);
    } EndForEach; /* Link */
  } EndDo; /* Unit */
  /*
   * Have to create link
   *
   *   (This is terribly inefficient - would really like the function
   *    that creates the link to return it, so I wouldn't have to go
   *    find it again.)
   */
  DB(fprintf(dout,"created\n"););
  /* 
   * shoud be
   * Connect(net->my_n,sl,su,net->my_n,dl,du,LINK);
   */
  ConnectUnitBlock(net->my_n,sl,net->my_n,dl,du,su,1,LINK);
  return GetLink(net,sl,su,dl,du);
}


int ReadLink(sl,su,dl,du,w,m,g)
int *sl,*su,*dl,*du;
real *w,*m,*g;
{
  int din_errors = 0;
  double x; /* MUST BE DOUBLE - x is used as the receptor for the scanf */
  if (fscanf(din," %lf",&x)!=1) din_errors++;
  *sl = x;
  if (fscanf(din," %lf",&x)!=1) din_errors++;
  *su = x;
  if (fscanf(din," %lf",&x)!=1) din_errors++;
  *dl = x;
  if (fscanf(din," %lf",&x)!=1) din_errors++;
  *du = x;
  if (fscanf(din," %lf",&x)!=1) din_errors++;
  *w = x;
  if (fscanf(din," %lf",&x)!=1) din_errors++;
  *m = x;
  if (fscanf(din," %lf",&x)!=1) din_errors++;
  *g = x;
  return din_errors;
}


int ReadLinks(net)
NETWORK* net;
{
  int i;
  int n_links;
  int sl,su,dl,du;
  real w,m,g;
  LINK_PTR link;
  int din_errors=0;
  char str[100];
  double x; /* MUST BE DOUBLE - x is used as the receptor for the scanf */
  /*
   * Read link header
   */
  if (fscanf(din,"\n %lf links", &x)!=1) {
    IError("incorrect link header");
    return 1;
  } 
  n_links = x;
  sprintf(str," %s .momentum .gain\n",WEIGHT_NAME);	/* added init <cr> */
  if (!InputMatches(din,str)) {
    IError("expected to find \"%s\"",str);
    return 1;
  }
  if (n_links == net->n_links) 	/* network is built already */
    fprintf(dout,"Reading links...");
  else
    fprintf(dout,"Creating links...");
  fflush(dout);
  /*
   * Read links
   */
  for (i=0; i<n_links; i++) {
    if ((din_errors=ReadLink(&sl,&su,&dl,&du,&w,&m,&g))) {
      IError("error reading link %d",i);
      return din_errors;
    }
    link = GetLink(net,sl,su,dl,du);
    LINK_WEIGHT(link) = w;
    link->momentum    = m;
    link->gain        = g;
  }
  fprintf(dout,"done\n");
  net->p->compiled = 0;
  CompileNet(net);
  return 0;
}


int SaveNet(net)
NETWORK* net;
{
  if (!NetExists(net)) {
    IError("net is null");
    return;
  }
  fprintf(dout,"read-network ");
  IPutQuotedString(net->builder,'"',dout);
  fprintf(dout," %s %d %d",net->name,net->n_layers,net->n_links);
  ForEachLayer {
    fprintf(dout," %d",layer->n_units);
  } EndForEach;
  fprintf(dout,"\n");
  if (SaveLinks(net)) {
    IError("Failed to save links for net \"%s\"",net);
    return 0;
  }
  IShowObject("net",SHOW_SAVE);
  fprintf(dout,"show net.p.epoch-n\n");	/* so it's printed out on read */
  return 1;
}


/******************************************************************************
 * Commands
 */

int command_save_network(tokc,tokv)
int tokc;
char *tokv[];
{
  NETWORK *thisnet;
  EXAMPLE_REC *er;
  int allnets = 0;
  int parameters = -1;
  int examples = 0;
  int i;
  static struct OPTION
    options[] =
      {
	{OPTSIGN,"allnets",0},
	{OPTSIGN,"parameters",0},
	{OPTSIGN,"examples",0},
	{OPTLAST,NULL,0}};
  if (GiveHelp(tokc)) {
    IUsage("[network-name] [%sallnets] [%sexamples] [%sparameters]");
    ISynopsis("print out network description (use instead of `save')");
    IHelp
      (IHelpArgs,
"Print out a network description.",
"To get this output into a file, use redirection, e.g.:",
"  save-network > filename",
"OPTIONS",
"Options can be abbreviated to their first letter.",
"network-name  : save the network with this name, rather than the",
"                current network.",
"%sallnets    : save all of the networks (default is save only current)",
"%sexamples   : save the current examples with the network (default -)",
"%sparameters : save the parameters block (default is nosave unless all",
"                nets are being saved",
"CUSTOMIZATION",
"This command calls the function `SaveLinks' to write out the weights.",
"This command can be customized by writing your own version of",
"SaveLinks()",
	  NULL);
    return;
  }
  if (tokc>1 && !IsOption(tokv[1])) {
    thisnet = (NETWORK*)IGetObjectPtr("NETWORK",tokv[1]);
    if (thisnet==NULL) IErrorAbort("network \"%s\" not found",tokv[1]);
    tokv++;
  } else
    thisnet = net;
  if (!NetExists(thisnet)) {
    IError("network \"%s\" is null",(thisnet!=NULL?thisnet->name:"no net"));
    return;
  }
  tokv++;
  while (*tokv) {
    if        (OptionMatch(*tokv,"allnets",   options,&allnets)) {
    } else if (OptionMatch(*tokv,"parameters",options,&parameters)) {
    } else if (OptionMatch(*tokv,"examples",  options,&examples)) {
    } else {
      IErrorAbort("option \"%s\" not understood",*tokv);
    }
    tokv++;
  }
  if (parameters==-1) parameters = allnets;
  if (allnets) {
    for (i=0;i<n_nets;i++)
      if (neta[i]!=NULL && NetExists(neta[i]))
	if (!SaveNet(neta[i])) {
	  IError("Failed to save net \"%s\"",neta[i]->name);
	  return;
	}
  } else 
    if (!SaveNet(thisnet)) {
      IError("Failed to save net");
      return;
    }
  if (parameters) 
    IShowObject("p",SHOW_SAVE);
  if (examples)
    if (current_el==NULL) {
      fprintf(dout,"# The current example list is completely empty\n");
    } else {
      fprintf(dout,"# Example list #%d contains %d examples:%s\n",
	      current_el->my_n,current_el->n_examples,
	      (current_el->complete ? "" : " (missing a final end)"));
      for (er=current_el->head;er!=NULL;er=er->next)
	(void) ShowExample(er);
      fprintf(dout,"# End examples\n");
    }
}


int command_read_network(tokc,tokv)
int tokc;
char *tokv[];
{
  int net_n,i,n;
  char *name; /* the net name */
  IUsage("<build-command> <name> <#layers> <#links> <layer0 size> ...");
  if (GiveHelp(tokc)) {
    ISynopsis("[internal]");
    IHelp(IHelpArgs,
"This command is used for reading from a file a network description",
"generated by the `save-network' command.",
"It is not intended to be used interactively.",
	  NULL);
    return;
  }
  if (tokc<5) {
    IError(IPrintUsage(tokv[0],usage));
    return;
  }
  if (!IIsLegalVarName(tokv[2])) {
    IExpectedToken("<network-name>",tokv[2],NULL);
    return;
  }
  name = tokv[2];
  /***
   * build the net if necessary
   */
  net_n = FindOrBuildNetSkel(name);
  if (net_n>=0) net = neta[net_n];
  else net = NULL;
  if (!NetExists(net)) {
    fprintf(dout,"Building new net \"%s\" ...\n",name);
    if ((n=IDoCommandLine(tokv[1]))!=1) {	/* used to be != 0 */
      IError("error in building new net (%d)",n);
      return; /* there was an error! */
    }
  } else
    fprintf(dout,"Found net \"%s\", checking similarity...\n",name);
  np = net->p;
  if (strcmp(net->builder,tokv[1]))
    fprintf(dout,
       "Warning: builder strings are different:\n%-13s %s\n%-13s %s\n%s\n",
       "existing net:",net->builder,"saved net:",tokv[1],
       "Continuing anyway...");
  /***
   * Check the # of layers, the # of links, and the size of each layer
   */
  n = atoi(tokv[3]);
  if (net->n_layers!=n) {
    IError("Mismatch: net has %d layers, but %s were requested",
	   net->n_layers,tokv[3]);
    return;
  }
  if (tokc!=5+n) {
    IError("number of arguments does not match number of layers");
    return;
  }
  for (i=0;i<net->n_layers;i++) {
    if (atoi(tokv[i+5])!=net->layer[i].n_units) {
      IError("wrong number of units in layer %d (net has %d, command has %s)",
	     i,net->layer[i].n_units,tokv[i+5]);
      return;
    }
  }
  if (ReadLinks(net)>0) {
    IError("error occurred reading links");
    return;
  }
}


int command_save_links(tokc,tokv)
int tokc;
char *tokv[];
{
  int i;
  int inc;
  int debug = 0;
  int *layermap;
  static struct OPTION options[] =
    {
     {OPTPRES|OPTINTARG,"map",	0},
     {OPTPRES|OPTINTARG,"inc",	0},
     {OPTPRES,          "debug",0},
     {OPTLAST,NULL,0}
   };
  IUsage("[%pinc <n>] [%pmap <old-layer-n> <new-layer-n>]* [%pdebug]");
  if (GiveHelp(tokc)) {
    ISynopsis("print out link descriptions");
    IHelp(IHelpArgs,
"Print out the links in a network, possibly renumbering some layers.",
"Different argument orders can produce different renumberings.",
"To get this output into a file, use redirection, e.g.:",
"  save-links > filename",
"OPTIONS",
"%pinc <n>     : Increment layer numbers by <n> (except for layer 0).",
"%pmap <n> <m> : Renumber layer <n> to be layer <m> when writing out links.",
"                This option can be repeated.",
"%pdebug       : Just print out specified layer mappings.",
	  NULL);
    return;
  }
  if (!NetExists(net)) {
    IError("current net is null");
    return;
  }
  layermap = ITempCalloc(int,net->n_layers);
  for (i=0; i<net->n_layers; i++) layermap[i] = i;
  tokv++;
  while (*tokv) {
    if        (OptionMatch(*tokv,"inc",options,NULL)) {
      if (!IExpectInteger(tokv[1])) return;
      inc = atoi(*++tokv);
      for (i=1; i<net->n_layers; i++) layermap[i] += inc;
    } else if (OptionMatch(*tokv,"map",options,NULL)) {
      if (!IExpectInteger(tokv[1]) || !IExpectInteger(tokv[2])) return;
      layermap[atoi(*++tokv)] = atoi(*++tokv);
    } else if (OptionMatch(*tokv,"debug",options,&debug)) {
    } else {
      IErrorAbort("option \"%s\" not understood",*tokv);
    }
    tokv++;
  }
  if (debug)
    for (i=0; i<net->n_layers; i++)
      fprintf(dout,"# Mapping layer %d to %d\n",i,layermap[i]);
  else {
    fprintf(dout,"read-links\n");
    if (TranslateLinks(net,layermap)) {
      IError("Failed to save links for net \"%s\"",net);
      return 0;
    }
  }
  return 1;
}


int command_read_links(tokc,tokv)
int tokc;
char *tokv[];
{
  IUsage("");
  if (GiveHelp(tokc)) {
    ISynopsis("[internal]");
    IHelp(IHelpArgs,
"Reads in link descriptions generated by the `save-links' command.",
"It is not intended to be used interactively.",
	  NULL);
    return;
  }
  if (!NetExists(net)) {
    IError("current net is null");
    return;
  }
  if (ReadLinks(net)>0) {
    IError("error occurred reading links");
    return 0;
  }
  return 1;
}

