#include <itf/option.h>
#include <itf/itf.h>
#include <sim/sim.h>
#include <signal.h>
#include <lib/example.h>
#include <lib/stop-flag.h>
#include "utils-bp.h"
#include "train-bp.h"

struct TRACE train1;
struct TRACE train2;
struct TRACE train2a;
struct TRACE train3;
struct TRACE train3a;
struct TRACE train4;
struct TRACE train4a;
struct TRACE train5;
struct TRACE btrack;

extern real safediv();
extern real GaussNoise();	/* example-bp.c */



int RandIntWithoutReplacement(n,max_n,narr)
int n,max_n,*narr;
{
  int i,index,r;
  if (n == 0)
    for (i=0; i<max_n; i++)
      narr[i] = i;
  else if (n >= max_n) {
    IError("n = %d must be less than max_n = %d",n,max_n);
    return;
  }
  index = RandInt(max_n-n);
  r = narr[index];
  for (i=index; i<max_n-n-1; i++)
    narr[i] = narr[i+1];
  return r;
}


/*
 * train for some epochs
 */
void
  EpochTrain(net)
NETWORK* net;
{
  EXAMPLE_CODE ex_status;
  NET_PARAMS *np = net->p;
  int n_examples_shown,example_n;
  int n_examples = current_el->n_examples;
  int *examples = ITempCalloc(int,n_examples);
  stop_flag = 0;
  np->total_links = net->n_links;
  np->max_example = n_examples;
  if (np->modify_input==1 && !np->computed_expected_values)
    ComputeExpectedValues(net);
  ZeroLinkDeltas(net);
  maybe_do_trace(&train1,0);
  EnsureELIndex(current_el);
  do {
    /*
     * FORWARD and BACKWARD PROPAGATION for each example
     */
    ResetExamples(net);
    maybe_do_trace(&train2,p->trace_freq);
    maybe_do_trace(&train2a,0);
    ResetWeightStats(np);
    n_examples_shown   = 0;
    np->sumsq_errors   = 0;
    np->total_n_errors = 0;
    while (stop_flag<=1) {
      if (n_examples_shown == n_examples) {
	ex_status = EXAMPLE_END;
	break;
      }
      example_n = (np->randomize_examples ?
		   RandIntWithoutReplacement(n_examples_shown,n_examples,examples) :
		   n_examples_shown);
      if ((ex_status=DoExample(net,example_n)) != EXAMPLE_OK) break;
      if (np->replace_ex_function) ReplaceExample(net);
      n_examples_shown++;
      ForwardProp(net);
      ComputeError(net);
#ifndef DBM_NETWORK
      if (np->therapy) PrintExamplePerformance(net);
#endif  DBM_NETWORK
      if (   (np->ex_n_errors || np->always_bp_example)
	  && (!np->therapy || !current_el->index[example_n]->training))
	BackwardProp(net);
      maybe_do_trace(&train3,p->trace_freq);
      maybe_do_trace(&train3a,0);
    }
    /*
     * BATCH UPDATE of links
     */
    if (ex_status==EXAMPLE_END && stop_flag<=1) {
      if (np->allow_reset_links) MaybeResetLinks(net);
      if (np->calc_cosines)
	UpdateWeightsAndCosines(net);
      else
	UpdateWeights(net);
      UpdateWeightStats(np);
      maybe_do_trace(&train4,p->trace_freq);
      maybe_do_trace(&train4a,0);
    }
    np->epoch_n++;
  } while ((!np->stop_at_epoch || np->epoch_n <= np->stop_at_epoch)
	   && (ex_status==EXAMPLE_END)
	   && (np->total_n_errors || !np->stop_when_correct)
	   && !stop_flag);
  maybe_do_trace(&train5,0);

 error:
  if (stop_flag) 
    ResetExamples(net);
}


int 
  command_train(tokc,tokv)
int tokc;
char *tokv[];
{
  int n;
  if (GiveHelp(tokc)) {
    IUsage("<number-of-epochs>");
    ISynopsis("train the network on the examples");
    IHelp
      (IHelpArgs,
       "Train the network for specified number of epochs, or until",
       "there are no errors.",
       "Tracing can be done at the following points:",
       "*   train1",
       "    do { /* some epochs */",
       "*     train2 (if epoch_n % p.trace_freq == 0)",
       "*     train2a",
       "      while (LoadExample()) {",
       "        ... Forward- and Back-prop ...",
       "*       train3 (if epoch_n % p.trace_freq == 0)",
       "*       train3a",
       "      }",
       "*     train4 (if epoch_n % p.trace_freq == 0)",
       "*     train4a",
       "    }",
       "*   train5",
       NULL);
    return;
  }
  
  if (tokc>2 || tokc==2 && !IIsInteger(tokv[1])) {
    IExpected("only one integer argument",NULL);
    return 1;
  }
  if (!NetExists(net)) {
    IError("current net is NULL");
    return 1;
  }
  if (current_el == NULL) {
    IError("current example list is NULL");
    return 1;
  }
  if (tokc>1) {
    n = atoi(tokv[1]);
    net->p->stop_at_epoch = net->p->epoch_n + n-1;
  } else
    net->p->stop_at_epoch = NULL;
  (void) signal(SIGINT,IncStopFlag);
  EpochTrain(net);
}

