#ifndef DBM_NETWORK
#define TIME_NETWORK
#define MAX_T 21
#endif DBM_NETWORK

#define WEIGHT_STATS
#define DELTA_BAR_DELTA
#define LINKS_HAVE_MOMENTUM
#define ELMAN_STYLE
#define EXAMPLE_LABELS
#define VECTOREXLOAD

#define WEIGHT_NAME		".w"
#define UNIT_OUTPUT_NAME	".y"
#define LINK_WEIGHT(link)	(link)->w
#define UNIT_OUTPUT(unit)	(unit)->y
#define UNIT_TARGET(unit)	(unit)->d

#ifdef TIME_NETWORK
#define UNIT_INPUT_T(unit,t)	(unit)->y_t[t]	/* input loaded here */
#define UNIT_OUTPUT_T(unit,t)	(unit)->y_t[t]
#define UNIT_TARGET_T(unit,t)	(unit)->d
#else
#define UNIT_INPUT_T(unit,t)	(unit)->y	/* input loaded here */
#define UNIT_OUTPUT_T(unit,t)	(unit)->y
#define UNIT_TARGET_T(unit,t)	(unit)->d
#endif TIME_NETWORK

#define sym2asym(y)    		(0.5*((y)+1.0))
#define asym2sym(y)		(2.0*(y)-1.0)

#define ScaleOutput(y)		(np->unit_function ? asym2sym(y) : (y))
#define InvScaleOutput(y)	(np->unit_function ? sym2asym(y) : (y))

#define HEBB_NAME			".h"
#define LINK_HEBB_WEIGHT(link)		(link)->h
#define UNIT_HEBB_OUTPUT(unit) \
  (np->unit_function ? UNIT_OUTPUT(unit) : asym2sym(UNIT_OUTPUT(unit)))
#define UNIT_HEBB_OUTPUT_T(unit,t) \
  (np->unit_function ? UNIT_OUTPUT_T(unit,t) : asym2sym(UNIT_OUTPUT_T(unit,t)))


typedef struct LINK {			/* LINK, BIND */
  real	w;		/* weight save: */
  real	h;		/* hebb weight */
  real	dEdw;
  real	previous_dEdw;
  real	momentum;
  real	delta_bar;
  real	gain;
  int	restrict;	/* default: 0 (-1 nonpos;  1 nonneg) */
  int	lesioned;	/* default: 0 */
} LINK;

typedef struct UNIT {			/* UNIT, BIND */
  real	x;		/* nosave: */
  real	y;		/* save: */
#ifdef TIME_NETWORK
  real	y_t[MAX_T];
#endif TIME_NETWORK
  real	d;
  real  dEdx;
  real  dEdy;
  int	lesioned;	/* default: 0 */
} UNIT;


typedef struct LAYER_PARAMS { /* BIND */
  int	is_input_layer;		/* default: 0 */
  int	is_output_layer;	/* default: 0 */
#ifdef DBM_NETWORK
  int	clamped;		/* default: 0 */
  real	initial_y;		/* y_t[0] default: 0.5 */
#else
  real	initial_y;		/* y_t[0] default: 0.2 */
#endif DBM_NETWORK
} LAYER_PARAMS;


typedef struct PARAMS { /* BIND */
  int		trace_freq;
} PARAMS;


typedef struct NET_PARAMS {		/* BIND */
  /* learning parameters */ 
  real		epsilon;		/* learning rate default: 1.0 */
  real		alpha;			/* momentum factor default: 0.9 */
  real		alpha_kick_in;		/* epoch after which alpha gets used default: 10 */
  real		rand_link_field; 	/* randomize weights +/- default: 0.5 */
  real		weight_decay;		/* default: 0.0 */
  real		input_strength;		/* default: 1.0 */
  int		max_t;			/* number of itations to run network default: 7 */
  int		max_input_t;		/* default: 0 (clamped) */
  int		n_train;		/* num training iterations default: 3 */
#ifdef DBM_NETWORK
  int		max_iterations;		/* default: 300 */
  real		fp_tolerance;		/* default: 0.01 */
  real		temp;		
  real		initial_temp;		/* default: 50.0 */
  real		final_temp;	 	/* default: 1.0 */
  real		temp_decay;		/* default: 0.9 */
  real		unit_damping;		/* default: 0.6 */
  int		settle_positive;	/* default: 1 */
  int		save_fp_freq;		/* default: 0 */
  real		restore_strength;	/* default: 1.0 */
  real		unit_noise;		/* default: 0.0 see SumInput() prop-dbm.c */
  int		mask_onset;		/* default: 0 */
  real		mask_prob;		/* default: 0.5 */
#endif DBM_NETWORK
  /* input modification */
  int		modify_input;		/* default: 0 (see ModifyInput() in example-bp.c) */
  real		input_noise;		/* default: 0.0 */
  real 		input_feature_prob;	/* default: 1.0 (see SampleInputFeatures() in example-bp.c)*/
  int		computed_expected_values; /* nsi: default: 0 */
  real		*input_expected_value;	/* nsi: see ComputeExpectedValues() in example-bp.c */
  real		*output_expected_value;	/* nsi: see ComputeExpectedValues() in example-bp.c */
  /* hebb weights */
  real		hebb_scale;		/* hebb-weights scale factor default: 0 */
  real		hebb_damping;		/* hebb-weights damping factor default: 0.0 */
  /* delta bar delta */
  int		delta_bar_delta;	/* whether to use dbd default: 0   */
  int		dbd_kick_in;		/* init epoch to use dbd default: 0  */
  real		beta;			/* integration factor default: 0.5 */
  real		gain_incr;		/* amount to increment gain default: 0.1 */
  real		gain_decr;		/* amount to mult gain on decr default: 0.8 */
  real		max_gain;		/* maximum gain value default: 50.0 */
  /* reset links */
  int		allow_reset_links;	/* reset momentum if gradient increases by reset_factor default: 0 */
  real		reset_momentum;		/* integrated sumsq_gradient */
  real		reset_ratio;		/* save for displaying */
  real		reset_factor;		/* reset integration factor default: 0.9 */
  real		reset_threshold;	/* gradient must increase by this factor to reset default: 2.0 */
  real		reset_scale;		/* factor to multiply gain by default: 0.1 */
  /* Network Compilation */
  int		compiled;		/* nsi: whether network is compiled default: 0 */
  int		n_units;		/* number of forward units */
  int		n_links;		/* number of backward links */
  UNIT		**units;		/* nsi: vector of pointers to (non-input) units */
  LINK		**links;		/* nsi: vector of pointers to links */
  UNIT		**s_units;		/* nsi: vector of pointers to (source) units */
  UNIT		**d_units;		/* nsi: vector of pointers to (destination) units */
  int		n_bunits;		/* number of backward units */
  int		n_blinks;		/* number of backward links */
  UNIT		**bunits;		/* nsi: vector of pointers to (non-input) units */
  LINK		**blinks;		/* nsi: vector of pointers to links */
  UNIT		**s_bunits;		/* nsi: vector of pointers to (source) units */
  UNIT		**d_bunits;		/* nsi: vector of pointers to (destination) units */
  /* Network Characteristics */
  int		input_layer_n;		
  int		output_layer_n;		
  int		bias_layer_n;		
  int		forward_layer_n;	/* default: 2 start bp here */
  int		backward_layer_n;	/* default: 1 stop bp here */
  int		input_width;		
  int		output_width;		
  int		total_links;		
  int		have_targets;		/* nsi: */
  int		elman_style;		/* nsi: */
  int		output_shift;		/* nsi: */
  int		input_shift;		/* nsi: */
  int		predict;		/* nsi: */
  int		input_function;		
  int		error_function;		/* error function default: 0 CrEnt, 1 w/ TolCrEn */
  int		unit_function;		/* default: 0 */
  int		output_criteria;	/* default: 0 (see OutputMeetsCriteria() in perform-ibp.c) */
  /* Treatment of example values on input and output */
  int		randomize_examples;	/* sample examples randomly default: 0 */
  int		ensure_ex_closest;	/* default: 0 (see AddInputNoise() in example-bp.c) */
  int		replace_ex_function;	/* default: 0 */
  int		auto_associative;	/* nsi: */
  int		max_example;		/* nsi: nosave: */
  int		example_n;		/* nsi: nosave: */
  string	example_label;		/* nsi: nosave: */
  real		midrange_tolerance;	/* error tolerance for d!=0/1 default: 0.05 */
  real		tolerance;		/* error tolerance for d=0/1  default: 0.1 */
  /* Algorithm control */
  int		therapy;		/* default: 0 */
  int		online;			/* nsi: default: 0 1=on-line, 0=batch */
  int		backtrack;		/* nsi: default: 0 */
  int		epoch_n;
  int		stop_at_epoch;		/* nsi: nosave: */
  int		stop_when_correct;	/* default: 1 */
  int		vary_epsilon;		/* nsi: default: 0 */
  int		rand_seed;
  int		always_bp_example;	/* default: 1 */
  int		always_bp_unit;		/* default: 1 */
  int		calc_cosines;		/* default: 1 */
  int		update_one_weight;	/* nsi: nosave: */
  /* Execution statistics */
  real		sumsq_errors;		/* nsi: */
  real		sumsq_errors_prev;	/* nsi: invisible: */
  real		sumsq_errors_change;	/* nsi: */
  real		sum_grad_by_update;	/* nsi: */
  real		sum_old_grad_by_new;	/* nsi: */
  real		sumsq_gradient;		/* nsi: */
  real		prev_sumsq_gradient;	/* nsi: */
  real		sumsq_update;		/* nsi: */
  real		sum_gradient;		/* nsi: */
  real		sum_update;		/* nsi: */
  real		sum_weight;		/* nsi: */
  real		sumsq_weight;		/* nsi: */
  real		avg_link_size;		/* nsi: */
  real		ex_sumsq_errors;	/* nsi: */
  real		gradient_cos;		/* nsi: */
  real		update_cos;		/* nsi: */
  real		gradient_size;		/* nsi: */
  real		update_size;		/* nsi: */
  int		total_n_errors;		/* nsi: */
  int		ex_n_errors;		/* nsi: */
  real		avg_sumsq_error;	/* nsi: per example */
  real		sum_cos_angle;		/* nsi: */
  real		avg_cos_angle;		/* default: 0 */
  int		n_ex_errors_counted;	/* nsi: used to calc avg */
} NET_PARAMS;
