/*
 * $Id: bind.h,v 1.1 90/12/07 11:50:10 drew Exp $
 * 
 * $Log:	bind.h,v $
# Revision 1.1  90/12/07  11:50:10  drew
# Initial revision
# 
 */

/**********************************************************************
 * Copyright 1990,1991 by University of Toronto, Toronto, Ontario, Canada.
 * 
 *			 All Rights Reserved
 * 
 * Permission to use, copy, modify, distribute, and sell this software
 * and its documentation for any purpose is hereby granted without fee, 
 * provided that the above copyright notice appears in all copies and that 
 * both the copyright notice and this permission notice appear in 
 * supporting documentation, and that the name of University of Toronto 
 * not be used in advertising or publicity pertaining to distribution 
 * of the software without specific, written prior permission.  
 * University of Toronto makes no representations about the suitability 
 * of this software for any purpose. It is provided "as is" without 
 * express or implied warranty. 
 *
 * UNIVERSITY OF TORONTO DISCLAIMS ALL WARRANTIES WITH REGARD TO THIS 
 * SOFTWARE, INCLUDING ALL IMPLIED WARRANTIES OF MERCHANTABILITY AND 
 * FITNESS, IN NO EVENT SHALL UNIVERSITY OF TORONTO BE LIABLE FOR ANY 
 * SPECIAL, INDIRECT OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES WHATSOEVER 
 * RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN ACTION OF 
 * CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF OR IN 
 * CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
 *
 **********************************************************************/

#define VALUE_BUFFER 1024 /* max size of values --- exceed this & DIE */

/* reserved features */
#define ITF_SYSTEM	0x00800000L
#define ITF_MARK	0x01000000L
#define ITF_VISIBLE	0x02000000L
#define ITF_READONLY	0x04000000L
#define ITF_SAVE	0x08000000L
#define ITF_RESETABLE	0x10000000L
#define ITF_ORPHANABLE	0x20000000L
#define ITF_VARLEN	0x40000000L
#define ITF_EXPORT	0x80000000L
#define ITF_RESERVED_FEATURES 0xFFFF0000L

#define ITF_STRING		1
#define ITF_REAL		2
#define ITF_FLOAT		3
#define ITF_DOUBLE		4
#define ITF_LONG		5
#define ITF_INT			6
#define ITF_SHORT		7
#define ITF_CHAR		8
#define ITF_BOOLEAN		9
#define ITF_PSUEDO_FIELD	10
#define ITF_STRUCT		11
#define ITF_VAR_TYPES		12

/* max number of characters in a var name: */
#define MAX_VAR_NAME	60
#define MASSAGE_VAR_NAME(name,buf) \
  {if (name!=buf){int _ii;for(_ii=0;name[_ii];_ii++)if(name[_ii]=='_')buf[_ii]='-';else buf[_ii]=name[_ii];buf[_ii]=0;}}

/* work out what the type real is */
#ifdef DOUBLE
typedef double	real;
#define REAL_DEFINED
#endif
#ifdef SINGLE
typedef float real;
#define REAL_DEFINED
#endif
#ifndef REAL_DEFINED
typedef float real;
#define SINGLE
#define REAL_DEFINED
#endif

/* PFPC - pointer to a function returning pointer to a char */
typedef char	*(*PFPC)();
typedef char	*string;
/* typedef int	boolean;
/* typedef char	generic; /* the generic type of pointers to structures */

/* Union Pointer Or Int (POI) */

typedef union POI {
  struct {
    short	offset;
    short	type;
  }		field;
  long		longint;
  PFPC		pfpc;
  string	string;
  real		*real_ptr;
  float		*float_ptr;
  double	*double_ptr;
  char		*char_ptr;
  string	*string_ptr;
  long		*long_ptr;
  int		*int_ptr;
  short		*short_ptr;
  boolean	*boolean_ptr;
  generic	*ptr;
  generic	**pptr;
  generic	**ppptr;
} POI;

#define MAX_INDIRECTION	3

/*
 * These are used in itype[i]
 */
#define FIXED_ARRAY	0x1
#define COUNTED_ARRAY	0x2
#define POINTER		0x4
#define AUTO_POINTER	0x8
#define ENV_IS_ARRAY	(FIXED_ARRAY|COUNTED_ARRAY)
#define ENV_IS_POINTER	(POINTER|AUTO_POINTER)

/*
 * These are bits that object_prop (in BINDING) can take
 */
#define IS_FIELD	0x01	/* a field descriptor */
#define IS_STRUCT	0x02	/* a struct */
#define IS_SIMPLE	0x04	/* long, float, string, etc. */
#define IS_COUNTER	0x08
#define IS_ARRAY	0x10
#define	IS_UNION	0x20
#define IS_TAG		0x40
#define HAS_DEFAULT	0x80
#define OBJECT_IS_PROP	0xFF

/*
 * The following are just used for convenient input to Ibind_*
 */

#define OBJECT		0x0020
#define P_OBJECT	0x0040	/* the POI is a ptr to a C object */
#define PP_OBJECT	0x0080	/* the POI is a ptr to a C object ptr */
#define PPP_OBJECT	0x0100	/* the POI is a ptr to a C object ptr ptr */
#define P_COUNT		0x0200
#define PP_COUNT	0x0400
#define PPP_COUNT	0x0800
#define P_FIXED		0x1000
#define PP_FIXED	0x2000
#define PPP_FIXED	0x4000

#define FIXED_PS	(P_FIXED|PP_FIXED|PPP_FIXED)
#define COUNT_PS	(P_COUNT|PP_COUNT|PPP_COUNT)
#define ARRAY_PS	(P_COUNT|PP_COUNT|PPP_COUNT|P_FIXED|PP_FIXED|PPP_FIXED)

#define TEST(X)		((X)!=0)

/***
 * Following are used as flags to the MoveTo functions
 */
#define SWAP	1
#define NO_SWAP	0

typedef struct BINDING {
  union {
    real	real;
    POI		poi;
  } value;
  short		itype[MAX_INDIRECTION+1];
  short		array_count[MAX_INDIRECTION+1];
  short		ilevel;
  short		reference_count;
  short		object_prop;
  short		element_size;
  short		var_type;	/* takes values ITF_STRING, etc. */
  long		features;
  TBL_ITEM	*tbl_entry;
  PFPC		func_to_call;
  char		*default_val;
  /* the following are for the printing format, e.g. %9.5g would be: */
  char		format_code;	/* 'g' */
  char		format_width;	/* 9 (as an integer) */
  char		format_prec;	/* 5 (as an integer) */
} BINDING;

typedef struct ADDRESS_ENV {
  BINDING	*binding;
  BINDING	*parent_binding;	/* for fields */
  POI		poi;
  POI		parent_poi;
  char		*name;
  short		ilevel;
  short		have_index[MAX_INDIRECTION+1];
} ADDRESS_ENV;

typedef enum ITF_FEATURE_ACTION {SET_FEATURE,UNSET_FEATURE,ENQUIRE};

typedef enum SHOW_STYLE {SHOW_DISPLAY,SHOW_SAVE,SHOW_BRIEF};

#define RESET_ENV(X) {(X)->binding=NULL;(X)->poi.ptr=NULL;(X)->have_index=0;(X)->is_complex=0;(X)->index=0;}

extern	void	IBindInit();		/* (max_vars) */
extern	int	IAssignString();/* (string*,string) */

/* all of the following will create a name if one is not given,
   and will return it */
extern	char* 	IBindReal();	/* (name,*real,mode,count)*/
extern	char*	IBindLong();	/* (name,*long,mode,count)*/
extern	char*	IBindString();	/* (name,**char,mode,count)*/
extern	char*	IBindFloat();	/* (name,*float,mode,count)*/
extern	char*	IBindDouble();	/* (name,*double,mode,count)*/
extern	char*	IBindInt();		/* (name,*int,mode,count)*/
extern	char*	IBindShort();	/* (name,*short,mode,count)*/
extern	char*	IBindChar();	/* (name,*char,mode,count)*/
extern	char*	IBindStruct();	/* (name,typename,*char,mode,count)*/
extern	char*	IBindStructNoTouch();/* (name,typename,*char,mode,count)*/
extern	char*	IGetDefault();	/* (name) */
extern	void	ISetBindingDefaults(); /* (BINDING*) */

extern	BINDING*IBindPoi();		/* (name,type,poi,mode,count)*/

extern	int	IEnterType();	/* (typename,prop) */
extern	int	IEnterField();	/* (struct_name,field_type,name,offset,prop,ind,counter)*/

extern	int	IGetType();	/* (name) */
extern	char*	IGetTypeName();	/* (type-id) */
extern	PFPC	ISetFunction();	/* (name,PFPC) */
extern	long	IGetFeatures();	/* (name) */
extern	int	ISetFeatures();	/* (name,long,0 or 1) */
extern	int	ISetReadonly();	/* (name,0 or 1) */
extern	int	IIsReadonly();	/* (name) */
extern	int	ISetVisible();	/* (name,0 or 1) */
extern	int	IIsVisible();	/* (name) */
extern	int	ISetReport();	/* (name,0 or 1) */
extern	int	IIsReport();	/* (name) */
extern	int	ISetMark();		/* (name,0 or 1) */
extern	int	IIsMarked();	/* (name) */
extern	long	ISetFeatureMask();	/* (long) */
extern	int	ISetValue();	/* (name,char* value) */
extern	char	*IGetName();	/* (name) - so that we can store
					   a reference to a variable as a
					   pointer - DON'T MODIFY the name! */
extern	char	*IGetValue();	/* (name,char *format) */
extern	char	*IGetValueCompositeName ARGS(());	/* varargs */
extern	char	*IGetValueNull();	/* (name) */
extern	char	*IGetValueVar();	/* (BINDING*) */
extern	char	*IGetSimpleValue();	/* (BINDING*,take_address,format) */

extern	char	*IStructDim();	/* (name,type) */
extern	char	*IVarAddress();	/* (name) */
extern	TBL_ITEM*IParseFieldOrObject();
extern	int	IIsStructType();/* (int) */
extern	int	IIsArray();	/* (BINDING*) */
extern	int	IIsStruct();	/* (BINDING*) */
extern	int	IIsField();	/* (BINDING*) */
extern	int	IIsCounter();	/* (BINDING*) */
extern	char	*IArrayCounter();
extern	int	IEnvIsSimple();
extern	int	IEnvIsArray();
extern	int	IEnvIsStruct();
extern	int	IEnvIsPointer();
extern	generic*	IGetObjectPtr(); /* (object-type,name) */
extern	int	ISetupStructFieldEnv();
extern	int	IRecalcFieldEnv();

extern	int	IIsObjectName();	/* (name) */
extern  int	IIsLegalVarName();	/* (name) */
extern  int	ICheckExistSvar();	/* (name) - creates a string variable it if it doesn't exist - returns -1 if bad name */
extern	void	IShowVarValues();
extern	void	ISetEnvDefaults();
extern	void	ISetDefaults();	/* ("STRUCT_NAME",*struct) */
extern	void	ISetFormat();
extern	char*	IGetFormat();
extern	char*	IMakeFormatStr();
extern	int	IParseFormat();
extern	char*	IPrintStringFormatted();
extern	char*	IPrintRealFormatted();

extern char*	IMakeScalar();		/* (type,name) */
extern int*	IGetScalarIntAddr();	/* (name) */
extern real*	IGetScalarRealAddr();	/* (name) */
extern char*	IMakeVector();		/* (type,name,len) */
extern int*	IGetVectorIntAddr();	/* (name,&len) */
extern real*	IGetVectorRealAddr();	/* (name,&len) */

extern generic*	IDeleteObject();	/* (name) */
extern void	IShowObject();		/* (name,style) */

extern char*	ltoa();
extern char*	itoa();
extern char*	ftoa();
extern char*	dtoa();
extern char*	btoa();
extern long	atol();
/* extern float	atof(); */
extern double	atod();
extern boolean	atob();

extern int itf_obj_context;
extern int itf_type_context;
extern HASH_TABLE *itf_table;

extern char default_format_code[];
extern char default_format_width[];
extern char default_format_prec[];
extern char *legal_format_codes[];
extern char itf_value_buffer[];
extern char itf_name_buffer[];
extern struct ADDRESS_ENV itf_saved_address;
extern char *itf_type_names[];
extern int itf_top_struct_type;

#define COPY_STRUCT(source,dest) bcopy((char*)(source),(char*)(dest),sizeof(*source))
#ifndef sun
#define bcopy(a,b,n) {int i; for (i=0;i<n;i++) ((char*)(b))[i]=((char*)(a))[i];}
#endif
