
/**********************************************************************
 * $Id: display.c,v 1.1 91/04/30 16:01:38 xerion Exp $
 *
 * $Log:	display.c,v $
 * Revision 1.1  91/04/30  16:01:38  xerion
 * Initial revision
 * 
 *	Revision 1.2  91/03/20  13:35:31  drew
 *	Somewhat stable version, still experimental.
 *	
 *	Revision 1.1  91/03/13  10:07:22  drew
 *	Initial revision
 *	
 **********************************************************************/

/**********************************************************************
 * Copyright 1990,1991 by University of Toronto, Toronto, Ontario, Canada.
 * 
 *			 All Rights Reserved
 * 
 * Permission to use, copy, modify, distribute, and sell this software
 * and its documentation for any purpose is hereby granted without fee, 
 * provided that the above copyright notice appears in all copies and that 
 * both the copyright notice and this permission notice appear in 
 * supporting documentation, and that the name of University of Toronto 
 * not be used in advertising or publicity pertaining to distribution 
 * of the software without specific, written prior permission.  
 * University of Toronto makes no representations about the suitability 
 * of this software for any purpose. It is provided "as is" without 
 * express or implied warranty. 
 *
 * UNIVERSITY OF TORONTO DISCLAIMS ALL WARRANTIES WITH REGARD TO THIS 
 * SOFTWARE, INCLUDING ALL IMPLIED WARRANTIES OF MERCHANTABILITY AND 
 * FITNESS, IN NO EVENT SHALL UNIVERSITY OF TORONTO BE LIABLE FOR ANY 
 * SPECIAL, INDIRECT OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES WHATSOEVER 
 * RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN ACTION OF 
 * CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF OR IN 
 * CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
 *
 **********************************************************************/

#include <stdio.h>

#include <X11/Intrinsic.h>
#include <X11/Xlib.h>
#include <X11/StringDefs.h>
#include <X11/Shell.h>

#include "display.h"

extern int	IDoSingleCommand(); /*dcp (FILE *inStream, FILE *outStream) ;*/

/* Private functions */
static int			enableCommandLineInput() ;
static XtInputCallbackProc	commandLineCB() ; 


/* Top Level widget for the application */
Widget		topLevel ;

/* 
 * these are passed in from the itf routines 
 * I'd rather not have them here, but there is no easy
 * way around it right now.
 */
static FILE	*InStream, *OutStream ;


/**********************************************************************
 *	Name:		initGraphics
 *	Description:	initializes the Xt toolkit and creates the
 *			top level widget. We can't use XtAppInitialize
 *			because it exits when it can't open the display
 *			and we want to continue, but without graphics.
 *	Parameters:
 *		char	*appName - the application name
 *	Return Value:
 *		int	initGraphics - 0 on failure, 1 on success
 **********************************************************************/
int	initGraphics(argc, argv)
  int	*argc ;
  char	**argv ;
{
  extern Widget		topLevel ;
  XtAppContext		appContext ;
  Display		*display ;

  XtToolkitInitialize() ;
  appContext = XtCreateApplicationContext() ;
  display    = XtOpenDisplay(appContext, NULL,
			     argv[0], "Xerion", NULL, 0, argc, argv) ;

  if (display == NULL) {
    XtDestroyApplicationContext(appContext) ;
    fprintf (stderr, "Unable to open display, graphics disabled.\n") ;
    return 0 ;
  }

  topLevel = XtAppCreateShell(argv[0], "Xerion", 
			      applicationShellWidgetClass, display,
			      NULL, (Cardinal)0) ;
  return 1 ;
}
/**********************************************************************/


/***********************************************************************
 *	Name:		realizeGraphics
 *	Description:	realizes the graphical displays for the simulator.
 *			any children of the topLevel widget should be
 *			created in autoinit routines
 *	Parameters:	NONE
 *	Return Value:	
 *		int	buildGraphics - 0 on failure, 1 on success
 ***********************************************************************/
int	realizeGraphics() {
  
  if (topLevel == NULL)
    return 0 ;
  
  XtRealizeWidget(topLevel);
  return 1 ;
}
/**********************************************************************/


/**********************************************************************
 *	Name:		graphicsLoop
 *	Description:	the main event loop for using with the graphics.
 *			It also initializes the alternate input stream.
 *	Parameters:
 *		FILE	*inStream  - the stream for input (NULL if you
 *					don't want one)
 *		FILE	*outStream - the stream for output
 *	Return Value:
 *			0 on error, otherwise NEVER RETURNS.
 **********************************************************************/
int	graphicsLoop(inStream, outStream)
  FILE	*inStream, *outStream;
{
  XtAppContext	appContext ;
  
  appContext =  XtWidgetToApplicationContext(topLevel) ;
  
  if (inStream != NULL) {
    if (outStream == NULL)
      outStream = stdout ;
    if (enableCommandLineInput(inStream, outStream) == 0)
      return 0 ;
  }
  
  XtAppMainLoop(appContext) ;
  
  return 0 ;
}
/**********************************************************************/


/**********************************************************************
 *	Name:		enableCommandLineInput
 *	Description:	adds command line input to the graphics events
 *			and adds the callback to handle the event.
 *	Parameters:
 *		FILE	*inStream - the stream for input
 *		FILE	*outStream - the stream for output
 *	Return Value:
 *		int	enableCommandLineInput - +1 on success, 0 on 
 *			failure.
 **********************************************************************/
static int	enableCommandLineInput(inStream, outStream)
  FILE	*inStream, *outStream;
{
  int			fileDescriptor ;
  XtInputId		id ;
  XtAppContext		appContext ;
  
  if (topLevel == NULL)
    return 0 ;
  
  InStream  = inStream ;
  OutStream = outStream ;
  
  fileDescriptor = fileno(InStream) ;
  if (fileDescriptor < 0)
    return 0 ;
  
  appContext =  XtWidgetToApplicationContext(topLevel) ;
  
  id = XtAppAddInput(appContext, fileDescriptor, XtInputReadMask, 
		     commandLineCB, (caddr_t)NULL) ;
  if (id == 0) 
    return 0 ;
  
  return 1 ;
}
/**********************************************************************/


/**********************************************************************
 *	Name:		commandLineCB
 *	Description:	the callback for when a command is entered
 *			to the commandLine. This calls the routines
 *			from the itf module which actually reads the
 *			command. It must be added to the event list
 *			by XtAppAddInput.
 *	Parameters:
 *		caddr_t		clientData - unused
 *		int		sourceFile - the file descriptor of
 *				the input file
 *		XtInputId	id - the id returned by XtAppAddInput
 *	Return Value:
 *		NONE
 **********************************************************************/
static XtInputCallbackProc	commandLineCB(clientData, sourceFile, id)
  caddr_t	clientData ;
  int		sourceFile ;
  XtInputId	id ;
{
  if (IDoSingleCommand(InStream, OutStream) == 0)
    exit (0) ;
  return (XtInputCallbackProc)1 ;
}
/**********************************************************************/
