/*
 * $Id: hash-table.h,v 1.1 90/12/07 11:50:33 drew Exp $
 * 
 * $Log:	hash-table.h,v $
# Revision 1.1  90/12/07  11:50:33  drew
# Initial revision
# 
 */

/**********************************************************************
 * Copyright 1990,1991 by University of Toronto, Toronto, Ontario, Canada.
 * 
 *			 All Rights Reserved
 * 
 * Permission to use, copy, modify, distribute, and sell this software
 * and its documentation for any purpose is hereby granted without fee, 
 * provided that the above copyright notice appears in all copies and that 
 * both the copyright notice and this permission notice appear in 
 * supporting documentation, and that the name of University of Toronto 
 * not be used in advertising or publicity pertaining to distribution 
 * of the software without specific, written prior permission.  
 * University of Toronto makes no representations about the suitability 
 * of this software for any purpose. It is provided "as is" without 
 * express or implied warranty. 
 *
 * UNIVERSITY OF TORONTO DISCLAIMS ALL WARRANTIES WITH REGARD TO THIS 
 * SOFTWARE, INCLUDING ALL IMPLIED WARRANTIES OF MERCHANTABILITY AND 
 * FITNESS, IN NO EVENT SHALL UNIVERSITY OF TORONTO BE LIABLE FOR ANY 
 * SPECIAL, INDIRECT OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES WHATSOEVER 
 * RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN ACTION OF 
 * CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF OR IN 
 * CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
 *
 **********************************************************************/

#ifndef _HASH_TABLE_
#define _HASH_TABLE_

#ifndef _PFI_
#define _PFI_
typedef int (*PFI)(); /* pointer to function returning int */
#endif

struct BINDING;	/* `forward' declaration of BINDING */

typedef struct TBL_ITEM {
  int			context;
  int			id;	/* begins numbering at 1 */
  union {
    char		*ptr;	/* to an arbitrary structure or function */
    long		token;  /* or a token identifier */
    int			count;
    PFI			pfi;
    struct BINDING	*binding;
  } data;
  char			*name;
  struct TBL_ITEM	*next;	/* in this hash slot */
} TBL_ITEM;

typedef TBL_ITEM *(*ht_item_alloc)();

typedef struct HASH_TABLE {
  TBL_ITEM	**hash_vec;
  TBL_ITEM	**id_vec;
  TBL_ITEM	**alpha_vec; /* unimplemented */
  int		hash_bins;
  int		max_items;
  int		free_id;
  int		free_context;
  ht_item_alloc	free_item;
  ht_item_alloc	new_item;
} HASH_TABLE;

#define HT_NOT_FOUND	1
#define HT_AMBIGUOUS	2
#define HT_DUPLICATE	3
#define HT_BAD_ID	4
#define HT_NO_MEMORY	5
#define HT_BAD_NAME	6
#define HT_OVERFLOW	7
#define HT_CANT_DO	8

extern int hash_errno;
extern char *hash_error[];

HASH_TABLE *INewTable();
		/* (ht, max_items, hash_bins, index_types) */
/* index_types can be one of the following: */
#define ID_INDEX	1
#define ALPHA_INDEX	2
void	  ITableAllocators();
		/* (new_item, free_item) - item allocators & deallocators */
TBL_ITEM *ITableLookup();
		/* (ht, name, context) */
int	  ITableGetId();
		/* (ht, name, context) = 0 for not found */
int	  INewTableContext();
		/* (ht) - allocates a new context */
TBL_ITEM *ITableLookupById();
		/* (ht, id) */
#define ITableInsert(TABLE, NAME, CONTEXT, INFO) ITableInsert_(TABLE, NAME, CONTEXT, (long)INFO)
TBL_ITEM *ITableInsert_();
		/* (ht, name, context, info) - assigns an id */
TBL_ITEM *ITableNext();
		/* (ht, &id) - id is a reference, start with it at 0 */
TBL_ITEM *ITableNextByContext();
		/* (ht, &id, context) - id is a reference, start with it at 0 */
#define ITableFirstByContext(ht, id, context) (*id=0, ITableNextByContext(ht, id, context))
int	  ITableDelete();
		/* (ht, name, context) */
int	  ITableDeleteByContext();
		/* (ht, context) */

#endif _HASH_TABLE_
