/*
 * $Id: mem.h,v 1.1 90/12/07 11:50:48 drew Exp $
 * 
 * $Log:	mem.h,v $
# Revision 1.1  90/12/07  11:50:48  drew
# Initial revision
# 
 */

/**********************************************************************
 * Copyright 1990,1991 by University of Toronto, Toronto, Ontario, Canada.
 * 
 *			 All Rights Reserved
 * 
 * Permission to use, copy, modify, distribute, and sell this software
 * and its documentation for any purpose is hereby granted without fee, 
 * provided that the above copyright notice appears in all copies and that 
 * both the copyright notice and this permission notice appear in 
 * supporting documentation, and that the name of University of Toronto 
 * not be used in advertising or publicity pertaining to distribution 
 * of the software without specific, written prior permission.  
 * University of Toronto makes no representations about the suitability 
 * of this software for any purpose. It is provided "as is" without 
 * express or implied warranty. 
 *
 * UNIVERSITY OF TORONTO DISCLAIMS ALL WARRANTIES WITH REGARD TO THIS 
 * SOFTWARE, INCLUDING ALL IMPLIED WARRANTIES OF MERCHANTABILITY AND 
 * FITNESS, IN NO EVENT SHALL UNIVERSITY OF TORONTO BE LIABLE FOR ANY 
 * SPECIAL, INDIRECT OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES WHATSOEVER 
 * RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN ACTION OF 
 * CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF OR IN 
 * CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
 *
 **********************************************************************/

extern char *calloc();
extern char *malloc();
extern char *realloc();
extern char *_ITempCalloc();
extern char *_ITempRealloc();
extern char **_ITempMatrix();
extern char *_ITempVector();
extern char **_IMatrix();
extern char *_IVector();
extern void _IStealTempAlloc ARGS((char *));

#ifndef ALLOC_MACROS
#define ALLOC_MACROS

#define Malloc(TYPE, SIZE) \
  ((TYPE*)malloc((unsigned) (SIZE)))

#define Realloc(TYPE, OLD, NUMBER) \
  ((TYPE*)realloc((char*)(OLD), sizeof(TYPE)*(unsigned) (NUMBER)))

#define ReallocOrAbort(TYPE, OLD, NUMBER) \
  ((TYPE*)reallocOrAbort((char*)(OLD), sizeof(TYPE)*(unsigned) (NUMBER)))

#define Calloc(TYPE, N) \
  ((TYPE*)calloc((unsigned) (N), (unsigned)sizeof(TYPE)))

#define CallocOrAbort(TYPE, N) \
  ((TYPE*)callocOrAbort((unsigned) (N), (unsigned)sizeof(TYPE)))

#endif ALLOC_MACROS

#define ITempCalloc(TYPE, N) \
  ((TYPE*)_ITempCalloc((unsigned) (N), (unsigned)sizeof(TYPE)))

#define IStealTempAlloc(OLD) \
  _IStealTempAlloc((char*)(OLD))

#define ITempRealloc(TYPE, OLD, N) \
  ((TYPE*)_ITempRealloc((char*)(OLD), (unsigned) (N), (unsigned)sizeof(TYPE)))

#define ITempMatrix(TYPE, NRL, NRH, NCL, NCH) \
  ((TYPE**)_ITempMatrix((unsigned)sizeof(TYPE), (unsigned)(NRL), (unsigned)(NRH), (unsigned)(NCL), (unsigned)(NCH)))

#define ITempVector(TYPE, NL, NH) \
  ((TYPE*)_ITempVector((unsigned)sizeof(TYPE), (unsigned)(NL), (unsigned)(NH)))

#define IMatrix(TYPE, NRL, NRH, NCL, NCH) \
  ((TYPE**)_IMatrix((unsigned)sizeof(TYPE), (unsigned)(NRL), (unsigned)(NRH), (unsigned)(NCL), (unsigned)(NCH)))

#define IFreeMatrix(M, TYPE, NRL, NRH, NCL, NCH) \
  ((TYPE**)_IFreeMatrix((char**)(M), (unsigned)sizeof(TYPE), (unsigned)(NRL), (unsigned)(NRH), (unsigned)(NCL), (unsigned)(NCH)))

#define IVector(TYPE, NL, NH) \
  ((TYPE*)_IVector((unsigned)sizeof(TYPE), (unsigned)(NL), (unsigned)(NH)))

#define IFreeVector(V, TYPE, NL, NH) \
  ((TYPE*)_IFreeVector((char*)(V), (unsigned)sizeof(TYPE), (unsigned)(NL), (unsigned)(NH)))

/***
 * pointers to record allocs during command
 * that should be freed at end of command
 */
typedef struct POINTER_LIST {
  char *p;
  struct POINTER_LIST *next;
};

typedef struct TEMP_ALLOC {
  struct POINTER_LIST *head;
  struct POINTER_LIST *tail;
  struct TEMP_ALLOC *prev;	/* there can be a stack of these things */
};

extern struct TEMP_ALLOC *temp_alloc;

extern void IFreeTempAllocs();
extern void IBeginTempAlloc();
