/*
 * $Id: background.c,v 1.1 90/12/07 11:55:28 drew Exp $
 * 
 * $Log:	background.c,v $
# Revision 1.1  90/12/07  11:55:28  drew
# Initial revision
# 
 */

/**********************************************************************
 * Copyright 1990,1991 by University of Toronto, Toronto, Ontario, Canada.
 * 
 *			 All Rights Reserved
 * 
 * Permission to use, copy, modify, distribute, and sell this software
 * and its documentation for any purpose is hereby granted without fee, 
 * provided that the above copyright notice appears in all copies and that 
 * both the copyright notice and this permission notice appear in 
 * supporting documentation, and that the name of University of Toronto 
 * not be used in advertising or publicity pertaining to distribution 
 * of the software without specific, written prior permission.  
 * University of Toronto makes no representations about the suitability 
 * of this software for any purpose. It is provided "as is" without 
 * express or implied warranty. 
 *
 * UNIVERSITY OF TORONTO DISCLAIMS ALL WARRANTIES WITH REGARD TO THIS 
 * SOFTWARE, INCLUDING ALL IMPLIED WARRANTIES OF MERCHANTABILITY AND 
 * FITNESS, IN NO EVENT SHALL UNIVERSITY OF TORONTO BE LIABLE FOR ANY 
 * SPECIAL, INDIRECT OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES WHATSOEVER 
 * RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN ACTION OF 
 * CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF OR IN 
 * CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
 *
 **********************************************************************/

#include <itf/itf.h>
#include <sim/sim.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <signal.h>

/* ARGSUSED */
int command_background(tokc,tokv)
int tokc;
char *tokv[];
{
  extern int errno;
  extern char *sys_errlist[];
  struct stat statbuf;
  int stat_ok;
  int out_exists;
  char *out_name;
  char *in_name;
  FILE *out_file = NULL;
  FILE *in_file = NULL;
  extern int mypid;
  extern char *myname;
  char c;
  char namebuf[100];
  int out_d,in_d;
  SignalHandler handler;
  IUsage("[infile] [outfile]");
  if (GiveHelp(tokc)) {
    ISynopsis("reconnects stdin and stdout to files");
    IHelp
      (IHelpArgs,
       "Disconnects stdin and stdout from the terminal and connects them",
       "to files.  This allows the process to be stopped and backgrounded",
       "and the user can logout.",
       "The command can be used in the following ways:",
       "  background infile > outfile",
       "  background infile outfile",
       "",
       "USE WITH FEP",
       "If you are running the simulator under fep, you need to exit from",
       "fep to run the simulator in the background.  Do this by typing",
       "\"fep-exit\", which will cause fep to exit.",
       "The signals SIGTERM and SIGHUP are set to be ignored, unless the",
       "action specified for them is not the default action, in which case",
       "they are unchanged.  This is done because fep sends these signals",
       "in an effort to kill the program when \"fep-exit\" is typed.",
       NULL);
    return;
  }
  /* these two don't work yet - don't know why
       "  background > outfile         (reads input from terminal until EOF)",
       "  background outfile           (reads input from terminal until EOF)",
       */

  /***
   * check if we have output redirection - if we do, use that for outfile
   */
  if (current_outstream!=command_stdout)
    out_file = current_outstream->file;
  /***
   * check if we have input redirection - if we do, use that for infile
   */
  if (current_instream!=command_stdin)
    in_file = current_instream->file;

  if ((tokc + (out_file!=NULL) + (in_file!=NULL)) > 3)
    IErrorAbort("too many arguments!");


  /***
   * If we don't already have an outfile, look for one as the last argument
   */
  if (out_file==NULL) {
    if (tokc<2) IErrorAbort("expecting an outfile name");
    if (stat(tokv[tokc-1],&statbuf)==0)
      IErrorAbort("\"%s\" already exists - can't use as outfile",tokv[tokc-1]);
    out_file = fopen(tokv[tokc-1],"w");
    if (out_file==NULL)
      IErrorAbort("cannot open \"%s\" for writing: %s",
		  tokv[tokc-1],sys_errlist[errno]);
    tokc--;
  }

  
  /***
   * If we don't already have an input, look for one as the last argument,
   * or read from stdin and store in in a temporary file on /tmp
   */
  if (in_file==NULL) {
    if (tokc>1) {
      in_name = tokv[1];
      tokc--;
    } else {
      int i;
      FILE *tmp;
      char *p;
      int size = 0;
      IErrorAbort("expecting an infile name");
      sprintf(namebuf,"/tmp/%s.%d.bgin",myname,mypid);
      for (p=namebuf;*p!=0;p++) ;
      while (stat(namebuf,&statbuf)==0) {
	sprintf(p,"%d",i);
	if (i>100) {
	  fclose(out_file);
	  IErrorAbort("cannot find file to use on /tmp - last try was \"%s\"",
		      namebuf);
	}
      }
      tmp = fopen(namebuf,"w");
      if (tmp==NULL) {
	fclose(out_file);
	IErrorAbort("cannot open \"%s\" for writing: %s",
		    namebuf,sys_errlist[errno]);
      }
      fprintf(stderr,"Type the commands to be executed, end with CTRL-D\n");
      while ((c=getc(din)) && c!=EOF) {
	putc(c,tmp);
	size++;
      }
      fclose(tmp);
      fprintf(stderr,"Commands have been stored in \"%s\"\n",namebuf);
      c = 0;
      /*
       * now we have to wait for the file to be available for reading
       * this takes time on an nfs system
       */
      sleep(1);
      do {
	(void) stat(namebuf,&statbuf);
	c++;
	if (statbuf.st_size<size) {
	  fprintf(stderr,"Waiting for \"%s\" to appear\n",namebuf);
	  sleep(1);
	}
      } while (c<100 && statbuf.st_size<size);
      in_name = namebuf;
    }

    in_file = fopen(in_name,"r");
    if (in_file==NULL) {
      fclose(out_file);
      IErrorAbort("cannot open \"%s\" for reading: %s",
		  in_name,sys_errlist[errno]);
    }
  }

  if (tokc>1) {
    fclose(out_file);
    fclose(in_file);
    IErrorAbort("too many arguments");
  }

  in_d = fileno(in_file);
  out_d = fileno(out_file);

  /*
  fprintf(stderr,"Input file is:\n");
  while ((c=getc(in_file)) && c!=EOF) {
    putc(c,stderr);
  }
  rewind(in_file);
  */

  if (dup2(in_d,0)==-1) {
    warn("background","cannot join infile and stdin: %s",sys_errlist[errno]);
  }
  
  if (dup2(out_d,1)==-1) {
    warn("background","cannot join outfile and stdout: %s",sys_errlist[errno]);
  }

  handler = signal(SIGHUP,SIG_IGN);
  if (handler!=SIG_DFL) signal(SIGHUP,handler);
  else fprintf(stderr,"Ignoring signal SIGHUP\n");

  handler = signal(SIGTERM,SIG_IGN);
  if (handler!=SIG_DFL) signal(SIGTERM,handler);
  else fprintf(stderr,"Ignoring signal SIGTERM\n");

  fprintf(stderr,"Now type \"fep-exit\" if you are using fep, or else type CTRL-Z and \"bg\"\n");

  if (dup2(1,2)==-1)
    warn("background","cannot join stdout and stderr: %s",sys_errlist[errno]);

  ISetValue("echo","true");
}
