/*
 * $Id: example.h,v 1.1 90/12/07 11:56:08 drew Exp $
 * 
 * $Log:	example.h,v $
# Revision 1.1  90/12/07  11:56:08  drew
# Initial revision
# 
 */

/**********************************************************************
 * Copyright 1990,1991 by University of Toronto, Toronto, Ontario, Canada.
 * 
 *			 All Rights Reserved
 * 
 * Permission to use, copy, modify, distribute, and sell this software
 * and its documentation for any purpose is hereby granted without fee, 
 * provided that the above copyright notice appears in all copies and that 
 * both the copyright notice and this permission notice appear in 
 * supporting documentation, and that the name of University of Toronto 
 * not be used in advertising or publicity pertaining to distribution 
 * of the software without specific, written prior permission.  
 * University of Toronto makes no representations about the suitability 
 * of this software for any purpose. It is provided "as is" without 
 * express or implied warranty. 
 *
 * UNIVERSITY OF TORONTO DISCLAIMS ALL WARRANTIES WITH REGARD TO THIS 
 * SOFTWARE, INCLUDING ALL IMPLIED WARRANTIES OF MERCHANTABILITY AND 
 * FITNESS, IN NO EVENT SHALL UNIVERSITY OF TORONTO BE LIABLE FOR ANY 
 * SPECIAL, INDIRECT OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES WHATSOEVER 
 * RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN ACTION OF 
 * CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF OR IN 
 * CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
 *
 **********************************************************************/

#ifndef _EXAMPLE_
#define _EXAMPLE_

typedef enum EXAMPLE_CODE {
  EXAMPLE_OK,
  EXAMPLE_END,
  EXAMPLE_ERROR,
} EXAMPLE_CODE;

/*
 * These correspond to commands which we find in the example file
 */
typedef enum EXAMPLE_ACTION {
  EXAMPLE_ACTION_BOT,
  NULL_ACTION,
  INPUT_EX,
  OUTPUT_EX,
  LAYER_EX,
  PREDICT_EX,
  COPY_EX,
  CLEAR_EX,
  FILL_EX,
  POP_EX,
  PUSH_EX,
  CLEARSTACK_EX,
  ENQUEUE_EX,
  DEQUEUE_EX,
  CLEARQUEUE_EX,
  ECALL_EX,
  EDO_EX,
  END_EX,	/* of loading a particular example */
  EXAMPLE_ACTION_TOP,
} EXAMPLE_ACTION;

typedef struct EXAMPLE_REC {	/* BIND */
  EXAMPLE_ACTION	action;
  int			dest_layer;
  int			n_values;
  int			t;
  int			block;
  real			weight;
  real			*value;		/* counter: n_values */
  int			*position;	/* counter: n_values */
  real			default_value;
  char			*position_vector;
  char			*value_vector;
  struct EXAMPLE_REC	*next;
  string		str;
  string		comment;
  int			my_n;
  int			default_position; /* default: 0 if =1 out-of-range
					     positions go at pos 0 */
  int			training;	/* Flags that can be set	*/
  int			validation;	/* for the training to		*/
  int			testing;	/* take notice of.		*/
  int			ignore;		/* Skipped by NextExample()	*/
#ifdef EXAMPLE_REC_EXTRA_FIELDS
  EXAMPLE_REC_EXTRA_FIELDS
#endif
} EXAMPLE_REC;

typedef struct EXAMPLE_LIST {	/* BIND */
  struct EXAMPLE_REC	*head;
  struct EXAMPLE_REC	*tail;
  struct EXAMPLE_REC	*current;
  struct EXAMPLE_REC	*next;
  int			complete;     /* true if there is a final END action */
  int			example_n;
  int			n_examples;
  int			my_n;
  int			inactive;
  struct EXAMPLE_REC	**index;
  int			index_size;
  int			default_training;
  int			default_validation;
  int			default_testing;
  int			default_ignore;
} EXAMPLE_LIST;

/* this structure keeps properties of the most recent example */
typedef struct EXAMPLE { /* BIND - used in new version of example.c */
  int			n;
  int			predict;
  int			output;
  int			input;
  int			layer;
  int			training;
  int			validation;
  int			testing;
  string		props_in_text;	/* ascii explanation for printing */
  int			position_loading;
  int			position;
  EXAMPLE_REC		*output_er;
  EXAMPLE_REC		*predict_er;
  EXAMPLE_REC		*input_er;
} EXAMPLE;

typedef struct EXAMPLE_PROPS { /* BIND - used in old version of example.c */
  int			predict;
  int			output;
  int			input;
};

extern struct EXAMPLE *example;		/* BIND */
extern struct EXAMPLE_LIST *current_el;	/* BIND altname: el */
extern struct EXAMPLE_REC *current_er;	/* BIND altname: er */
extern struct EXAMPLE_LIST **ela;	/* BIND counter: n_els */
extern int n_els;		/* BIND */
extern int example_show_end;	/* BIND */
extern string example_format;	/* BIND */
extern string example_type;	/* BIND should be a prefix of a function */

typedef EXAMPLE_CODE (*PFEC)();	/* pointer to function return example code */
typedef EXAMPLE_REC *(*PFER)();	/* pointer to function return example code */

extern int ReadExamples();	/* (char *filename) */
extern EXAMPLE_CODE NextExample ARGS((NETWORK *));
extern EXAMPLE_CODE SkipExample ARGS((NETWORK *));
extern EXAMPLE_CODE DoExample ARGS((NETWORK *, int));
extern EXAMPLE_CODE ResetExamples ARGS((NETWORK *));

extern EXAMPLE_CODE LoadNextExample ARGS((NETWORK *, struct EXAMPLE*, int, int, int));
extern EXAMPLE_CODE SkipNextExample ARGS((NETWORK *, struct EXAMPLE*));
extern EXAMPLE_CODE DoExampleN ARGS((NETWORK *, struct EXAMPLE*, int));

extern int add_to_example_list();

extern string ex_training_range;		/* BIND default: 0- */
extern string ex_validation_range;		/* BIND default: X */
extern string ex_testing_range;			/* BIND default: X */
extern string ex_ignore_range;			/* BIND default: X */
extern string ex_class_override;		/* BIND default: false */

#define T_SENTINEL (-1)
#define BLOCK_SENTINEL (-1)

#endif
