/*
 * $Id: show-ul.c,v 1.1 90/12/07 11:57:07 drew Exp $
 * 
 * $Log:	show-ul.c,v $
# Revision 1.1  90/12/07  11:57:07  drew
# Initial revision
# 
 */

/**********************************************************************
 * Copyright 1990,1991 by University of Toronto, Toronto, Ontario, Canada.
 * 
 *			 All Rights Reserved
 * 
 * Permission to use, copy, modify, distribute, and sell this software
 * and its documentation for any purpose is hereby granted without fee, 
 * provided that the above copyright notice appears in all copies and that 
 * both the copyright notice and this permission notice appear in 
 * supporting documentation, and that the name of University of Toronto 
 * not be used in advertising or publicity pertaining to distribution 
 * of the software without specific, written prior permission.  
 * University of Toronto makes no representations about the suitability 
 * of this software for any purpose. It is provided "as is" without 
 * express or implied warranty. 
 *
 * UNIVERSITY OF TORONTO DISCLAIMS ALL WARRANTIES WITH REGARD TO THIS 
 * SOFTWARE, INCLUDING ALL IMPLIED WARRANTIES OF MERCHANTABILITY AND 
 * FITNESS, IN NO EVENT SHALL UNIVERSITY OF TORONTO BE LIABLE FOR ANY 
 * SPECIAL, INDIRECT OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES WHATSOEVER 
 * RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN ACTION OF 
 * CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF OR IN 
 * CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
 *
 **********************************************************************/

#include <itf/itf.h>
#include <sim/sim.h>

/* ARGSUSED */
int command_show_unit(tokc,tokv)
int tokc;
char *tokv[];
{
  ADDRESS_ENV env,envcpy;
  char name[100];
  struct RANGE unit_range,layer_range;
  char *unit_range_str=NULL;
  char *field_name=NULL;
  int address = 0;
  int take_address;
  if (GiveHelp(tokc)) {
    IUsage("[+address] [layer-range] [: unit-range] [field-name]");
    ISynopsis("show the units in the network");
    IHelp(IHelpArgs,
	  "Field name should begin with a '.'",
	  NULL);
    return;
  }
  ++tokv;
  if (!CheckNetExists(net)) return;
  SetDefaultRange(&unit_range,NULL);
  SetDefaultRange(&layer_range,NULL);
  if (*tokv && strprefix(*tokv,"+address")) {
    tokv++;
    address=1;
  }
  if (*tokv && **tokv!='.') {
    if (!ParseRange(*tokv,&layer_range,NULL)) return;
    tokv++;
  }
  if (*tokv && !strcmp(":",*tokv) && tokv[1]) {
    unit_range_str = tokv[1];
    tokv+=2;
  }
  if (*tokv && **tokv=='.') {
    field_name = *tokv;
    tokv++;
  }
  if (*tokv)
    IErrorAbort("argument not understood; \"%s\"",tokv[0]);
  ForEachLayer {
    if (!InRange(layer_n,&layer_range,NULL)) continue;
    if (!ParseRange(unit_range_str,&unit_range,&layer->format))
      return;
    ICreateEnv(&env,nameof_unit[layer->unit_tag],NULL);
    ForEachUnitGeneric {
      if (!InRange(unit_n,&unit_range,&layer->format))
	continue;
      sprintf(name,(address?"unit[%d:%d] (%s) 0x%x":"unit[%d:%d] (%s)"),
	      layer_n,unit_n,nameof_unit[layer->unit_tag],unit);
      envcpy = env;
      envcpy.poi.ptr = unit;
      if (field_name) {
	sprintf(name+strlen(name),"%s",field_name);
	if (!IContParseObject(field_name,&envcpy,&take_address))
	  IErrorAbort("trouble parsing field of unit");
      }
      IShowObjectEnv(name,&envcpy,SHOW_DISPLAY,0);
    } EndForEach;
  } EndForEach;
}

/* ARGSUSED */
int command_show_link(tokc,tokv)
int tokc;
char *tokv[];
{
  ADDRESS_ENV env,envcpy;
  char name[60];
  int *dlayer,*slayer;
  int tag = 0;
  int debug = 0;
  int do_out_links=0;
  int show_deactivated=0;
  int address = 0;
  int number = 0;
  int num = 0;
  char *field_name = NULL;
  char buf[20];
  int take_address = 0;
  struct RANGE *dunit_range,*sunit_range,num_range;
  if (GiveHelp(tokc)) {
    IUsage("[+type TYPE] [+address] [+debug] [+/-outlinks] [+/-dlayer <range>] [+/-slayer <range>] [+dunit <layer-range>:<unit-range>] [+sunit <layer-range>:<range>] [field-name]");
    ISynopsis("show the links in the network");
    IHelp(IHelpArgs,
"OPTIONS",
"+type TYPE    : only display links of this type",
"+address      : show the address of the link",
"+number       : show the number of the link in order of traversal  (note",
"                that only displayed links are counted)",
"+/-outlinks   : show links going outward from a unit (default off)",
"+/-deactivated: show deactivated links (default off)",
"<field-name>  : show only this field in links, should begin with '.'",
"SPECIFYING WHICH LINKS TO DISPLAY",
"The dlayer and dunit options control which units have their links",
"displayed.  The slayer and sunit options control which links are",
"displayed for those units.  There can be any number of dlayer",
"and slayer options, but there can be only one sunit and one dunit",
"option for each layer.",
"+(-)dlayer <range>  : do (not) show links for units in layers in <range>",
"+dunit <layerrange>:<unitrange>",
"                    : only show links for these units in these layers",
"+(-)slayer <range>  : do (not) show links from units in layers in <range>",
"+sunit <layerrange>:<unitrange>",
"                    : only show links from these units for these layers",
"+numrange <range>   : only show links with a link number in this range",
"                      (note that link numbers are affected by the dlayer",
"                       and other similar options).",
	  NULL);
    return;
  }
  if (!CheckNetExists(net)) return;
  tokv++;
  SetDefaultRange(&num_range,NULL);
  AllocDisplayRanges(net,&dlayer,&slayer,&dunit_range,&sunit_range);
  while (*tokv) {
    if (strprefix(*tokv,"+type")) {
      if ((tag = LinkTag(tokv[1]))==0)
	IErrorAbort("expected link type after \"%s\"",*tokv);
      tokv++;
    } else if ((**tokv=='-'||**tokv=='+')&&strprefix((*tokv)+1,"outlinks")) {
      do_out_links = (**tokv=='+');
    } else if ((**tokv=='-'||**tokv=='+')&&strprefix((*tokv)+1,"deactivated")){
      show_deactivated = (**tokv=='+');
    } else if (   (**tokv=='-'||**tokv=='+')
	       && (   strprefix((*tokv)+1,"dlayer")
		   || strprefix((*tokv)+1,"slayer")
		   || strprefix((*tokv)+1,"dunit")
		   || strprefix((*tokv)+1,"sunit"))) {
      if (!SetDisplayRanges(net,&tokv,dlayer,slayer,dunit_range,sunit_range))
	return;
    } else if (strprefix(*tokv,"+address")) {
      address = 1;
    } else if (strprefix(*tokv,"+number")) {
      number = 1;
    } else if (strprefix(*tokv,"+debug")) {
      debug = 1;
    } else if (strprefix(*tokv,"+numrange")) {
      number = 1;
      if (!ParseRange(*++tokv,&num_range,NULL)) return;
    } else if (**tokv == '.') {
      field_name = *tokv;
    } else {
      IErrorAbort("bad option: \"%s\"",*tokv);
    }
    if (*tokv) tokv++;
  }
  if (debug) {
    PrintDisplayRanges(net,dlayer,slayer,dunit_range,sunit_range);
    return;
  }
  ForEachLayer {
    if (!dlayer[layer_n]) continue;
    ForEachUnitGeneric {
      if (!InRange(unit_n,&dunit_range[layer_n],&layer->format)) continue;
      ForEachLinkSetGeneric {
	if (!do_out_links && !IsDestSet) continue;
	if (!slayer[OtherLayerN]) continue;
	if (tag!=0 && tag!=ls->link_tag) continue;
	if (ls->deactivate && !show_deactivated) continue;
	ICreateEnv(&env,nameof_link[ls->link_tag],NULL);
	ApplyToEachLinkGeneric
	  (
	   if (InRange(OtherUnitN,&sunit_range[OtherLayerN],
		       &net->layer[OtherLayerN].format)) {
	     if (InRange(num,&num_range,NULL)) {
	       sprintf(name,"link[%s] (%s)",LinkLocation,nameof_link[ls->link_tag]);
	       if (address)
		 sprintf(name+strlen(name)," 0x%x",link);
	       if (number) {
		 sprintf(buf,"<%d>",num);
		 sprintf(name+strlen(name)," %5s",buf);
	       }
	       envcpy = env;
	       envcpy.poi.ptr = link;
	       if (field_name) {
		 strcat(name,field_name);
		 if (!IContParseObject(field_name,&envcpy,&take_address))
		   IErrorAbort("trouble parsing field of link");
	       }
	       IShowObjectEnv(name,&envcpy,SHOW_DISPLAY,0);
	     }
	     num++;
	   }
	   );
      } EndForEach;
    } EndForEach;
  } EndForEach;
}

AllocDisplayRanges(net,dlayer,slayer,dunit_range,sunit_range)
NETWORK *net;
int **dlayer,**slayer;
struct RANGE **dunit_range,**sunit_range;
{
  int i;
  *dlayer = ITempCalloc(int,net->n_layers);
  *slayer = ITempCalloc(int,net->n_layers);
  *dunit_range = ITempCalloc(struct RANGE,net->n_layers);
  *sunit_range = ITempCalloc(struct RANGE,net->n_layers);
  for (i=0;i<net->n_layers;i++) {
    SetDefaultRange(&((*sunit_range)[i]),&net->layer[i].format);
    SetDefaultRange(&((*dunit_range)[i]),&net->layer[i].format);
    (*slayer)[i] = 1;
    (*dlayer)[i] = 1;
  }
}

int SetDisplayRanges(net,atokv,dlayer,slayer,dunit_range,sunit_range)
NETWORK *net;
char ***atokv;
int *dlayer,*slayer;
struct RANGE *dunit_range,*sunit_range;
{
  int i,on;
  char **tokv = *atokv;
  int all_equal=1; /* if all layers are already on or off, switch them */
  struct RANGE range;
  if ((**tokv=='-'||**tokv=='+') && strprefix((*tokv)+1,"dlayer")) {
    on = (**tokv=='+');
    if (!ParseRange(tokv[1],&range,NULL)) return 0;
    for (i=0;i<net->n_layers;i++) if (dlayer[i]!=on) all_equal=0;
    if (all_equal) for (i=0;i<net->n_layers;i++) dlayer[i] = !on;
    for (i=0;i<net->n_layers;i++)
      if (InRange(i,&range,NULL))
	dlayer[i] = on;
    tokv++;
  } else if ((**tokv=='-'||**tokv=='+') && strprefix((*tokv)+1,"slayer")) {
    on = (**tokv=='+');
    if (!ParseRange(tokv[1],&range,NULL)) return 0;
    for (i=0;i<net->n_layers;i++) if (slayer[i]!=on) all_equal=0;
    if (all_equal) for (i=0;i<net->n_layers;i++) slayer[i] = !on;
    for (i=0;i<net->n_layers;i++)
      if (InRange(i,&range,NULL))
	slayer[i] = on;
    tokv++;
  } else if (strprefix(*tokv,"+dunit")) {
    on = 1;
    if (tokv[1]==NULL||tokv[2]==NULL||strcmp(tokv[2],":")||tokv[3]==NULL)
      IErrorAbort("expected <layer-range>:<unit-range> after \"+dunit\"");
    if (!ParseRange(tokv[1],&range,NULL)) return 0;
    for (i=0;i<net->n_layers;i++)
      if (dlayer[i]!=on && IsDefaultRange(&(dunit_range[i]),LayerFormat(i)))
	all_equal=0;
    if (all_equal) for (i=0;i<net->n_layers;i++) dlayer[i] = !on;
    for (i=0;i<net->n_layers;i++)
      if (InRange(i,&range,NULL)) {
	dlayer[i] = 1;
	if (!ParseRange(tokv[3],&(dunit_range[i]),LayerFormat(i)))
	  return 0;
      }
    tokv+=3;
  } else if (strprefix(*tokv,"+sunit")) {
    on = 1;
    if (tokv[1]==NULL||tokv[2]==NULL||strcmp(tokv[2],":")||tokv[3]==NULL)
      IErrorAbort("expected <layer-range>:<unit-range> after \"+sunit\"");
    if (!ParseRange(tokv[1],&range,NULL)) return 0;
    for (i=0;i<net->n_layers;i++)
      if (slayer[i]!=on && IsDefaultRange(&(sunit_range[i]),LayerFormat(i)))
	all_equal=0;
    if (all_equal) for (i=0;i<net->n_layers;i++) slayer[i] = !on;
    for (i=0;i<net->n_layers;i++)
      if (InRange(i,&range,NULL)) {
	slayer[i]=1;
	if (!ParseRange(tokv[3],&(sunit_range[i]),LayerFormat(i)))
	  return 0;
      }
    tokv+=3;
  } else {
    IErrorAbort("bad option for SetDisplayRanges \"%s\"",tokv[0]);
  }
  *atokv = tokv;
  return 1;
}

int PrintDisplayRanges(net,dlayer,slayer,dunit_range,sunit_range)
NETWORK *net;
int *dlayer,*slayer;
struct RANGE *dunit_range,*sunit_range;
{
  int rows,cols;
  fprintf(dout,"Source layers:\n");
  ForEachLayer {
    if (slayer[layer_n]) {
      RangeDimensions(net->layer[layer_n].n_units,&(sunit_range[layer_n]),
		     LayerFormat(layer_n),&rows,&cols);
      fprintf(dout,"  %2d: format:%dx%d",
	      layer_n,
	      net->layer[layer_n].format.rows,
	      net->layer[layer_n].format.cols);
      if (sunit_range[layer_n].depth==1)
	fprintf(dout," range:%d-%d",
	      sunit_range[layer_n].lo,
	      sunit_range[layer_n].hi);
      else
	fprintf(dout," range:%d-%d+%d",
	      sunit_range[layer_n].lo,
	      sunit_range[layer_n].hi,
	      sunit_range[layer_n].depth);
      fprintf(dout," area:%dx%d\n", rows,cols);
    }
  } EndForEach;
  fprintf(dout,"Destination layers:\n");
  ForEachLayer {
    if (dlayer[layer_n]) {
      RangeDimensions(net->layer[layer_n].n_units,&(dunit_range[layer_n]),
		     LayerFormat(layer_n),&rows,&cols);
      fprintf(dout,"  %2d: format:%dx%d range:%d-%d+%d area:%dx%d\n",
	      layer_n,
	      net->layer[layer_n].format.rows,
	      net->layer[layer_n].format.cols,
	      dunit_range[layer_n].lo,
	      dunit_range[layer_n].hi,
	      dunit_range[layer_n].depth,
	      rows,cols);
    }
  } EndForEach;
}
