/*
 * $Id: link.c,v 1.3 91/01/17 15:49:20 drew Exp Locker: xerion $
 * 
 * $Log:	link.c,v $
 * Revision 1.3  91/01/17  15:49:20  drew
 * fixed some bugs as defined by tap.
 * 
 * Revision 1.2  90/12/17  11:37:28  drew
 * added documentation but didn't try to clean up the code.
 * 
 * Revision 1.1  90/12/07  12:00:51  drew
 * Initial revision
 * 
 */

/**********************************************************************
 * Copyright 1990,1991 by University of Toronto, Toronto, Ontario, Canada.
 * 
 *			 All Rights Reserved
 * 
 * Permission to use, copy, modify, distribute, and sell this software
 * and its documentation for any purpose is hereby granted without fee, 
 * provided that the above copyright notice appears in all copies and that 
 * both the copyright notice and this permission notice appear in 
 * supporting documentation, and that the name of University of Toronto 
 * not be used in advertising or publicity pertaining to distribution 
 * of the software without specific, written prior permission.  
 * University of Toronto makes no representations about the suitability 
 * of this software for any purpose. It is provided "as is" without 
 * express or implied warranty. 
 *
 * UNIVERSITY OF TORONTO DISCLAIMS ALL WARRANTIES WITH REGARD TO THIS 
 * SOFTWARE, INCLUDING ALL IMPLIED WARRANTIES OF MERCHANTABILITY AND 
 * FITNESS, IN NO EVENT SHALL UNIVERSITY OF TORONTO BE LIABLE FOR ANY 
 * SPECIAL, INDIRECT OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES WHATSOEVER 
 * RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN ACTION OF 
 * CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF OR IN 
 * CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
 *
 **********************************************************************/

/**********************************************************************
 *	File name:   link.c
 *	Description: functions for adding links into a network
 *
 *	Exported functions:
 *		int	AddLink_()
 *
 *	Exported variables:
 *
 **********************************************************************/

#include <itf/itf.h>
#include "sim-dummy.h"
#include <sim.h>

static LINK_SET	*FindBestLinkSet();
static char	*AddLinkToLinkSet();

/**********************************************************************
 *	Name:         AddLink_
 *	Description:  Adds a link into a network
 *	Parameters:
 *		int	s_net - pointer to the source net
 *		int	s_layer - pointer to the source layer
 *		int	s_unit - pointer to the source unit
 *		int	d_net - pointer to the destination net
 *		int	d_layer - pointer to the destination layer
 *		int	d_unit - pointer to the destination unit
 *		int	link_tag - ???
 *		int	mode - BI_DIRECTIONAL or UNI_DIRECTIONAL
 *		int	group_n - unused
 *	Return Value:
 *		int	AddLink_ - +1 on success, 0 on failure
 **********************************************************************/
int	AddLink_(s_net,s_layer,s_unit,d_net,d_layer,d_unit,
		 link_tag,mode,group_n)
  int	s_net, s_layer, s_unit;
  int	d_net, d_layer, d_unit;
  int	link_tag, mode, group_n;
{
  NETWORK	*net;
  LAYER		*layer;
  LINK_SET	*linkset;
  char		*linkaddr;

  /* Add the forward link */
  linkset = FindBestLinkSet(s_net,s_layer,s_unit,d_net,d_layer,d_unit,
			    INWARD|(mode&(UNI_DIRECTIONAL|BI_DIRECTIONAL)),
			    link_tag, 0);
  if (linkset == NULL) 
    return 0 ;
  linkaddr = AddLinkToLinkSet(s_unit,d_net,d_layer,d_unit,linkset,link_tag,0);

  /* if bi-directional, add the back link */
  if (mode & BI_DIRECTIONAL) {
    linkset = FindBestLinkSet(d_net,d_layer,d_unit,s_net,s_layer,s_unit,
			      OUTWARD|BI_DIRECTIONAL,link_tag, 1);
    if (linkset == NULL) 
      return 0;
    (void) AddLinkToLinkSet(d_unit,s_net,s_layer,s_unit,linkset,link_tag,linkaddr);
  }
  return 1;
}
/**********************************************************************/

/**********************************************************************
 *	Name:		AddLinkToLinkSet
 *	Description:	adds a link into a specified link set
 *	Parameters:
 *		int	s_unit - pointer to the source unit
 *		int	d_net - pointer to the destination net
 *		int	d_layer - pointer to the destination layer
 *		int	d_unit - pointer to the destination unit
 *		LINK_SET *linkset - the set to insert the link into
 *		int	link_tag - ???
 *		char 	*linkaddr - the address of the link to a
 *	Return Value:   
 *		char    *AddLinkToLinkSet - the address of the link
 **********************************************************************/
static char	*AddLinkToLinkSet(s_unit,d_net,d_layer,d_unit,
				  linkset,link_tag,linkaddr)
  int		s_unit;
  int		d_net, d_layer, d_unit;
  LINK_SET	*linkset;
  int		link_tag ;
  char	 	*linkaddr;
{
  char *p;

  assert(   (linkaddr==NULL && (linkset->mode&LINK_VECTOR))
	 || (linkaddr!=NULL && (linkset->mode&LINK_POINTER_VECTOR)));
  assert(linkset->link_tag==link_tag);
  assert(linkset->mode&(CONNECT_BLOCK|CONNECT_PARTIAL));

  if (linkset->mode&LINK_VECTOR) {
    /***
     * have a vector of links - add to it
     */
    if (linkset->l.cp == NULL)
      p = malloc((linkset->n_links+1)*sizeof_link[link_tag]);
    else
      p = realloc(linkset->l.cp,(linkset->n_links+1)*sizeof_link[link_tag]);
    if (p == NULL) return NULL; 
    if (linkset->l.cp!=NULL
	&& p!=linkset->l.cp && (linkset->mode&BI_DIRECTIONAL)) {
      panic("AddLinkToLinkSet","can't cope! too complicated!");
      MoveBidirLinkVector(linkset,d_net,d_layer,d_unit,p);
      /***
       * what we need to do here is go and repair all the bidirectional
       * link stuff - all the pointers in other places
       */
    }
    linkset->l.cp = p;
    linkset->n_links++;
    if (linkset->mode&CONNECT_BLOCK) {
      if (linkset->starting_unit_n==s_unit+1) {
	linkset->starting_unit_n--;
	/***
	 * should probably bcopy up all the info in the links
	 */
	linkaddr = linkset->l.cp;
      } else if (linkset->starting_unit_n+linkset->n_links-1==s_unit) {
	linkaddr = linkset->l.cp + (linkset->n_links-1)*sizeof_link[link_tag];
      } else {
	panic("AddLinkToLinkSet","can't add on link to block");
      }
    }
    if (linkset->mode&CONNECT_PARTIAL) {
      linkaddr = linkset->l.cp;
      if (linkset->unit_id == NULL)
	p = malloc(sizeof(short)*linkset->n_links);
      else
	p = realloc(linkset->unit_id,sizeof(short)*linkset->n_links);
      if (p==NULL)
	panic("AddLinkToLinkSet","out of memory in non-recoverable place");
      linkset->unit_id = (short*)p;
      linkset->unit_id[linkset->n_links-1] = s_unit;
    }
  } else {
    /***
     * have a vector of pointers to links - add to it
     */
    if (linkset->l.cpp == NULL)
      p = malloc((linkset->n_links+1)*sizeof(char*));
    else
      p = realloc(linkset->l.cpp,(linkset->n_links+1)*sizeof(char*));
    if (p==NULL) return NULL; 
    linkset->l.cpp = (char**)p;
    linkset->n_links++;
    if (linkset->mode&CONNECT_BLOCK) {
      if (linkset->starting_unit_n==s_unit+1) {
	int i;
	linkset->starting_unit_n--;
	/* move all the links up one to make room */
	for (i=linkset->n_links-1;i>0;i--)
	  linkset->l.cpp[i] = linkset->l.cpp[i-1];
	linkset->l.cpp[0] = linkaddr;
      } else if (linkset->starting_unit_n+linkset->n_links-1==s_unit) {
	linkset->l.cpp[linkset->n_links-1] = linkaddr;
      } else {
	panic("AddLinkToLinkSet","can't add on link to block");
      }
    }
    if (linkset->mode&CONNECT_PARTIAL) {
      if (linkset->unit_id == NULL)
	p = malloc(sizeof(short)*linkset->n_links);
      else
	p = realloc(linkset->unit_id,sizeof(short)*linkset->n_links);
      if (p==NULL)
	panic("AddLinkToLinkSet","out of memory in non-recoverable place");
      linkset->unit_id = (short*)p;
      linkset->unit_id[linkset->n_links-1] = s_unit;
    }
  }
  return linkaddr;
}
/**********************************************************************/

/**********************************************************************
 *	Name:         FindBestLinkSet
 *	Description:
 *		Finds the best link set going from source from destination.
 *		A block is better than partial (but to be suitable for block,
 *		the source must fit on one end of the block).
 *		If linkpointer is true the mode of the linkset should be
 *		LINK_POINTER_VECTOR, otherwise LINK_VECTOR.
 *		If nothing is found, a new linkset is created
 *	Parameters:   As below
 *	Return Value: 
 *		LINK_SET *FindBestLinkSet - a ptr to the best link set
 ********************************************************************** */
static LINK_SET *FindBestLinkSet(s_net,s_layer,s_unit,d_net,d_layer,d_unit,
				 direction,link_tag,linkpointer,group_n)
{
  NETWORK	*net;
  LAYER		*layer;
  LINK_SET	*linkset;
  UNIT_INFO	*info;
  int		i;
  int		block_n = -1;
  int		partial_n = -1;
  int		mode_match = (linkpointer ? LINK_POINTER_VECTOR : LINK_VECTOR);

  if (d_net<0 || d_net>=n_nets) {
    warn("FindBestLinkSet","net # %d out of range",d_net);
    return 0;
  }
  net = neta[d_net];
  if (d_layer<0 || d_layer>=net->n_layers) {
    warn("FindBestLinkSet","layer # %d out of range",d_layer);
    return 0;
  }
  layer = &(neta[d_net]->layer[d_layer]);
  if (d_unit<0 || d_unit>=layer->n_units) {
    warn("FindBestLinkSet","unit # %d out of range",d_unit);
    return 0;
  }
  info = &(layer->info[d_unit]);
  for (i=0;i<info->n_lss;i++) {
    if (   info->ls[i]->connect_net == s_net
	&& info->ls[i]->link_tag == link_tag
	&& (group_n<1 || info->ls[i]->group_n == group_n)
	&& info->ls[i]->connect_layer == s_layer
	&& (info->ls[i]->mode & mode_match)
	&& !(info->ls[i]->mode &
	     (RECURRENT_LINKS|LINK_MATRIX_00|LINK_MATRIX_COL|LINK_MATRIX_ROW))
	&& !(info->ls[i]->mode & CONNECT_COMPLETE)
	&& ((info->ls[i]->mode&(INWARD|OUTWARD|BI_DIRECTIONAL|UNI_DIRECTIONAL))
	    == direction)
	) {
      if (   (info->ls[i]->mode & CONNECT_BLOCK)
	  && (   (info->ls[i]->starting_unit_n==s_unit+1)
	      || (info->ls[i]->starting_unit_n+info->ls[i]->n_links==s_unit))){
	block_n = i;
	break;
      }
      if (partial_n==-1 && (info->ls[i]->mode & CONNECT_PARTIAL))
	partial_n = i;
      if (partial_n==-1
	  && (info->ls[i]->mode & CONNECT_BLOCK)
	  && (info->ls[i]->n_links==1))
	partial_n = i;
    }
  }
  if (block_n!=-1)
    linkset = info->ls[block_n];
  else if (partial_n!=-1) {
    linkset = info->ls[partial_n];
    if (linkset->mode&CONNECT_BLOCK) {
      /* change it to partial connectivity */
      linkset->mode &= ~CONNECT_BLOCK;
      linkset->mode |= CONNECT_PARTIAL;
      linkset->unit_id = Calloc(short,1);
      linkset->unit_id[0] = linkset->starting_unit_n;
      linkset->starting_unit_n = 0;
    }
  } else {
    linkset = Calloc(LINK_SET,1);
    linkset->connect_net = s_net;
    linkset->connect_layer = s_layer;
    linkset->mode = (CONNECT_BLOCK | direction
		     | (linkpointer ? LINK_POINTER_VECTOR : LINK_VECTOR));
    linkset->n_links = 0;
    linkset->starting_unit_n = s_unit;
    linkset->link_tag = link_tag;
    linkset->l.cp = 0;
    linkset->unit_id = 0;
    AddLinkSet(info,linkset,group_n);
  }
  return linkset;
}
/**********************************************************************/

/**********************************************************************
 *	Name:         MoveBidirLinkVector
 *	Description:  Doesn't work yet
 *	Parameters:   I have no Idea
 *	Return Value: none
 **********************************************************************/
static int	MoveBidirLinkVector(linkset,d_net,d_layer,d_unit,p)
  LINK_SET	*linkset;
  char		*p;
{
  panic("MoveBidirLinkVector","nyi");
}
/**********************************************************************/
