/*
 * $Id: cli_serv.c,v 1.1 90/12/07 12:04:56 drew Exp $
 * 
 * $Log:	cli_serv.c,v $
# Revision 1.1  90/12/07  12:04:56  drew
# Initial revision
# 
 */

/**********************************************************************
 * Copyright 1990,1991 by University of Toronto, Toronto, Ontario, Canada.
 * 
 *			 All Rights Reserved
 * 
 * Permission to use, copy, modify, distribute, and sell this software
 * and its documentation for any purpose is hereby granted without fee, 
 * provided that the above copyright notice appears in all copies and that 
 * both the copyright notice and this permission notice appear in 
 * supporting documentation, and that the name of University of Toronto 
 * not be used in advertising or publicity pertaining to distribution 
 * of the software without specific, written prior permission.  
 * University of Toronto makes no representations about the suitability 
 * of this software for any purpose. It is provided "as is" without 
 * express or implied warranty. 
 *
 * UNIVERSITY OF TORONTO DISCLAIMS ALL WARRANTIES WITH REGARD TO THIS 
 * SOFTWARE, INCLUDING ALL IMPLIED WARRANTIES OF MERCHANTABILITY AND 
 * FITNESS, IN NO EVENT SHALL UNIVERSITY OF TORONTO BE LIABLE FOR ANY 
 * SPECIAL, INDIRECT OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES WHATSOEVER 
 * RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN ACTION OF 
 * CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF OR IN 
 * CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
 *
 **********************************************************************/

From neat.ai.toronto.edu!ken Tue Jan 10 08:37:24 1989
Received: from neat.ai.toronto.edu by ephemeral.ai.toronto.edu with SMTP id 10955; Tue, 10 Jan 89 08:37:22 EST
Received: by neat.ai.toronto.edu id 38118; Tue, 10 Jan 89 08:36:33 EST
From:	Ken Lalonde <ken@ai.toronto.edu>
To:	ssfels@ai.toronto.edu
Subject: inetd stuff
Message-Id: <89Jan10.083633est.38118@neat.ai.toronto.edu>
Date:	Tue, 10 Jan 89 08:36:30 EST
Status: R

Here's a shar file containing sample client/server programs.
The client part is straightforward.  The server code will
run under inetd if compiled -DINETD, otherwise it will run standalone.
I think it's best to use a standalone server for now.
That way you are in complete control of things.  Also if you run
under inetd, a couple of system data files have to be updated and inetd
has to be restarted, which you can't do yourself.

Anyway you'll have to modify your server code so that
in its main loop it also selects on ``s'' (the server socket
file descriptor -- not the best variable name...).  All that should
take is another FD_SET call.  Also be sure that the first argument
to select() is correct after this change.  If select indicates
that ``s'' is readable, there is a pending new connection,
and you should call accept() as done below.  If you want to service
one connection at a time, don't bother forking after accept().

If you run into problems, let me know.

#! /bin/sh
# This is a shell archive.  Remove anything before this line, then unpack
# it by saving it into a file and typing "sh file".  To overwrite existing
# If this archive is complete, you will see the following message at the end:
#		"End of shell archive."
#
# Contents:
#   Makefile client.c server.c
#
# Wrapped by ken@ai.utoronto.ca on Tue Jan 10 08:12:36 1989
#
PATH=/bin:/usr/bin:/usr/ucb ; export PATH
if test -f Makefile -a "${1}" != "-c" ; then 
  echo shar: Will not over-write existing file \"Makefile\"
else
echo shar: Extracting \"Makefile\" \(48 characters\)
sed "s/^X//" >Makefile <<'END_OF_Makefile'
XCFLAGS = -DDEBUG -DPORT=9876
Xall: client server
END_OF_Makefile
if test 48 -ne `wc -c <Makefile`; then
    echo shar: \"Makefile\" unpacked with wrong size!
fi
# end of overwriting check
fi
if test -f client.c -a "${1}" != "-c" ; then 
  echo shar: Will not over-write existing file \"client.c\"
else
echo shar: Extracting \"client.c\" \(1186 characters\)
sed "s/^X//" >client.c <<'END_OF_client.c'
X#include <stdio.h>
X#include <netdb.h>
X#include <sys/types.h>
X#include <sys/socket.h>
X#include <netinet/in.h>
X
X/*
X * Connect to PORT on this machine, and copy whatever the
X * server sends to stdout.
X */
Xmain()
X{
X	int s, c;
X	FILE *fp;
X
X	/* "localhost" is always an alias for the current machine */
X	if ((s = open_remote_stream("localhost")) < 0)
X		exit(1);
X	/* `s'' can be read or written to, using read or write. */
X	fp = fdopen(s, "r");	/* stdio is easier to use here */
X	while ((c = getc(fp)) != EOF)
X		putchar(c);
X	exit(0);
X}
X
X/*
X * Open a stream connection to the PORT service on ``host'',
X * returning the file descriptor or -1.
X */
Xint
Xopen_remote_stream(host)
X	char *host;
X{
X	struct hostent *hp;
X	struct sockaddr_in sin;
X	int s;
X
X	if ((hp = gethostbyname(host)) == NULL) {
X		fprintf(stderr, "%s: unknown host\n", host);
X		return -1;
X	}
X	bzero((char *)&sin, sizeof sin);
X	bcopy(hp->h_addr, (char *)&sin.sin_addr, hp->h_length);
X	sin.sin_family = AF_INET;
X	sin.sin_port = htons(PORT);
X	if ((s = socket(AF_INET, SOCK_STREAM, 0)) < 0) {
X		perror("socket");
X		return -1;
X	}
X	if (connect(s, (char *)&sin, sizeof sin) < 0) {
X		perror("connect");
X		close(s);
X		return -1;
X	}
X	return s;
X}
END_OF_client.c
if test 1186 -ne `wc -c <client.c`; then
    echo shar: \"client.c\" unpacked with wrong size!
fi
# end of overwriting check
fi
if test -f server.c -a "${1}" != "-c" ; then 
  echo shar: Will not over-write existing file \"server.c\"
else
echo shar: Extracting \"server.c\" \(2460 characters\)
sed "s/^X//" >server.c <<'END_OF_server.c'
X/*
X * Server skeleton: calls function doit(fd), where fd is
X * a stream connection to a new client.
X * When running under inetd (compile -DINETD), this is trivial.
X */
X#include <sys/types.h>
X#include <sys/socket.h>
X#include <netinet/in.h>
X#include <sys/ioctl.h>
X#include <netdb.h>
X#include <errno.h>
X#include <signal.h>
X#include <sys/wait.h>
X
Xmain(argc, argv)
X	char **argv;
X{
X	
X#ifdef INETD
X	 /* Inetd provides the client connection on fd 0. */
X	doit(0);
X	exit(0);
X#else
X	struct sockaddr_in sin;
X	int i, s, f;
X	void reapchild();
X	
X#ifndef DEBUG
X	/* close all fd's, disconnect from tty */
X	for (i = getdtablesize(); --i >= 0; )
X		close(i);
X	(void) open("/dev/null", 2);
X	dup2(0, 1);
X	dup2(0, 2);
X	if ((i = open("/dev/tty", 2)) >= 0) {
X		ioctl(i, TIOCNOTTY, (char *)0);
X		close(i);
X	}
X#endif
X	/*
X	 * Grab a socket, bind it to PORT, listen for clients,
X	 * fork a child for each new connection.
X	 */
X	if ((s = socket(AF_INET, SOCK_STREAM, 0)) < 0) {
X		perror("socket");
X		exit(1);
X	}
X	bzero((char *)&sin, sizeof sin);
X	sin.sin_family = AF_INET;
X	sin.sin_port = htons(PORT);
X	if (bind(s, (caddr_t)&sin, sizeof sin) < 0) {
X		perror("bind");
X		exit(1);
X	}
X	if (listen(s, 5) < 0) {
X		perror("listen");
X		exit(1);
X	}
X	signal(SIGCHLD, reapchild);
X	for (;;) {
X		f = accept(s, (struct sockaddr *)0, (int *)0);
X		if (f < 0) {
X			if (errno == EINTR)
X				continue;	/* interrupted; ignore it */
X			perror("accept");
X			exit(1);
X		}
X		switch (fork()) {
X		case 0:
X			close(s);	/* child doesn't need this */
X			signal(SIGCHLD, SIG_DFL);
X			doit(f);
X			exit(0);
X		case -1:
X			perror("fork");
X		default:
X			break;
X		}
X		close(f);
X	}
X#endif INETD
X}
X
X/* this asynchronously cleans up terminated children */
Xstatic void
Xreapchild()
X{
X	union wait status;
X
X	while (wait3(&status, WNOHANG|WUNTRACED, 0) > 0)
X		;
X}
X
X/* deal with client -- here we send back the current date */
Xdoit(fd)
X	int fd;
X{
X	time_t t;
X	char *cp, *ctime();
X
X#ifdef notdef
X	/*
X	 * To get the client address (host name + port number):
X	 */
X	struct sockaddr_in from;
X	int fromlen;
X	struct hostent *hp;
X
X	fromlen = sizeof from;
X	if (getpeername(fd, &from, &fromlen) < 0) {
X		perror("getpeername");
X		exit(1);
X	}
X	/* port is ntohs(from.sin_port) */
X	hp = gethostbyaddr(&from.sin_addr, sizeof(struct in_addr), AF_INET);
X	if (hp == NULL) {
X		char *inet_ntoa();
X		/* (numeric) host name is inet_ntoa(from.sin_addr) */
X	} else
X		/* host name is hp->h_name */
X#endif
X	time(&t);
X	cp = ctime(&t);
X	write(fd, cp, strlen(cp));
X}
END_OF_server.c
if test 2460 -ne `wc -c <server.c`; then
    echo shar: \"server.c\" unpacked with wrong size!
fi
# end of overwriting check
fi
echo shar: End of shell archive.
exit 0

