/*
 * $Id: geometry.c,v 1.1 90/12/07 12:05:07 drew Exp $
 * 
 * $Log:	geometry.c,v $
# Revision 1.1  90/12/07  12:05:07  drew
# Initial revision
# 
 */

/**********************************************************************
 * Copyright 1990,1991 by University of Toronto, Toronto, Ontario, Canada.
 * 
 *			 All Rights Reserved
 * 
 * Permission to use, copy, modify, distribute, and sell this software
 * and its documentation for any purpose is hereby granted without fee, 
 * provided that the above copyright notice appears in all copies and that 
 * both the copyright notice and this permission notice appear in 
 * supporting documentation, and that the name of University of Toronto 
 * not be used in advertising or publicity pertaining to distribution 
 * of the software without specific, written prior permission.  
 * University of Toronto makes no representations about the suitability 
 * of this software for any purpose. It is provided "as is" without 
 * express or implied warranty. 
 *
 * UNIVERSITY OF TORONTO DISCLAIMS ALL WARRANTIES WITH REGARD TO THIS 
 * SOFTWARE, INCLUDING ALL IMPLIED WARRANTIES OF MERCHANTABILITY AND 
 * FITNESS, IN NO EVENT SHALL UNIVERSITY OF TORONTO BE LIABLE FOR ANY 
 * SPECIAL, INDIRECT OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES WHATSOEVER 
 * RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN ACTION OF 
 * CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF OR IN 
 * CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
 *
 **********************************************************************/

/*
 * Geometry: Module containing code dealing with geometries and sizehints.
 *
 * Entry Points:
 *
 *    init_hints    : Routine to init. a XSizeHints structure to no hints.
 *
 *    parse_geometry: Like XParseGeometry but ADDS the result to a already
 *                    mostly filled out XSizeHints structure.  Handles
 *                    negative x & y positions correctly and sets US flags
 *                    as needed.  The program should first call Init_Hints
 *                    on the hints, fill in any hints it wants using P flags
 *                    and then call Parse_Geometry with the user's geometry.
 *                    A value of NULL for the geometry is prefectly ok.
 */

#include <X11/Xlib.h>
#include <X11/Xutil.h>

void init_hints(hints)
        XSizeHints *hints;
{
	hints->flags = 0;
	hints->x = hints->y = 0;
	hints->width = hints->height = 100;
}

void parse_geometry(dpy, screen, geometry, hints)
        Display *dpy;
        int screen;
        char *geometry;
        XSizeHints *hints;
{
	int status;

	if (!geometry)
	  return;

	status = XParseGeometry(geometry, &hints->x, &hints->y, &hints->width,
				&hints->height);
	if (status & (XValue|YValue)) {
	  hints->flags |= USPosition;
	  hints->flags &= ~PPosition;
        }

	if (status & (WidthValue|HeightValue)) {
	  hints->flags |= USSize;
	  hints->flags &= ~PSize;
        }

	if ( (status&XNegative) && (status&XValue) )
	  hints->x += DisplayWidth(dpy, screen) - hints->width;

	if ( (status&YNegative) && (status&YValue) )
	  hints->y += DisplayHeight(dpy, screen) - hints->height;
}
