/*
 * $Id: sock_init.c,v 1.1 90/12/07 12:05:13 drew Exp $
 * 
 * $Log:	sock_init.c,v $
# Revision 1.1  90/12/07  12:05:13  drew
# Initial revision
# 
 */

/**********************************************************************
 * Copyright 1990,1991 by University of Toronto, Toronto, Ontario, Canada.
 * 
 *			 All Rights Reserved
 * 
 * Permission to use, copy, modify, distribute, and sell this software
 * and its documentation for any purpose is hereby granted without fee, 
 * provided that the above copyright notice appears in all copies and that 
 * both the copyright notice and this permission notice appear in 
 * supporting documentation, and that the name of University of Toronto 
 * not be used in advertising or publicity pertaining to distribution 
 * of the software without specific, written prior permission.  
 * University of Toronto makes no representations about the suitability 
 * of this software for any purpose. It is provided "as is" without 
 * express or implied warranty. 
 *
 * UNIVERSITY OF TORONTO DISCLAIMS ALL WARRANTIES WITH REGARD TO THIS 
 * SOFTWARE, INCLUDING ALL IMPLIED WARRANTIES OF MERCHANTABILITY AND 
 * FITNESS, IN NO EVENT SHALL UNIVERSITY OF TORONTO BE LIABLE FOR ANY 
 * SPECIAL, INDIRECT OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES WHATSOEVER 
 * RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN ACTION OF 
 * CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF OR IN 
 * CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
 *
 **********************************************************************/


/* Here's a shar file containing sample client/server programs.
The client part is straightforward.  The server code will
run under inetd if compiled -DINETD, otherwise it will run standalone.
I think it's best to use a standalone server for now.
That way you are in complete control of things.  Also if you run
under inetd, a couple of system data files have to be updated and inetd
has to be restarted, which you can't do yourself.

Anyway you'll have to modify your server code so that
in its main loop it also selects on ``s'' (the server socket
file descriptor -- not the best variable name...).  All that should
take is another FD_SET call.  Also be sure that the first argument
to select() is correct after this change.  If select indicates
that ``s'' is readable, there is a pending new connection,
and you should call accept() as done below.  If you want to service
one connection at a time, don't bother forking after accept().

If you run into problems, let me know.
 */

#include <sys/types.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <sys/ioctl.h>
#include <netdb.h>
#include <errno.h>
#include <signal.h>
#include <sys/wait.h>
#include <stdio.h>

int s;

int sock_set(port)

int port;

{

#ifdef INETD
	 /* Inetd provides the client connection on fd 0. */
	doit(0);
	exit(0);
#else
	struct sockaddr_in sin;
 	int i, f;
        int portinc;
 	void reapchild();
#endif
 	
#ifndef DEBUG
 	/* close all fd's, disconnect from tty */
/* 	for (i = getdtablesize(); --i >= 0; )
 		close(i);
 	(void) open("/dev/null", 2);
 	dup2(0, 1);
 	dup2(0, 2);
 	if ((i = open("/dev/tty", 2)) >= 0) {
 		ioctl(i, TIOCNOTTY, (char *)0);
 		close(i);
 	}
*/
#endif
 	/*
 	 * Grab a socket, bind it to PORT, listen for clients,
 	 * 
 	 */
 	if ((s = socket(AF_INET, SOCK_STREAM, 0)) < 0) {
 		perror("socket");
 		exit(1);
 	}
 	bzero((char *)&sin, sizeof sin);
 	sin.sin_family = AF_INET;
 	sin.sin_port = htons(port);

        portinc = port;

/* try 100 different ports and print out correct one */
        for (i = 0; i<= 100; i++) {
 	    if ((portinc = bind(s, (caddr_t)&sin, sizeof sin)) < 0) {
		bzero((char *)&sin, sizeof sin);
		sin.sin_family = AF_INET;
		sin.sin_port = htons(port+i+1);
 	    } else {
              printf("port opened is %d\n", port+i);
              break;
            }
        }

 	    if (portinc < 0) {
 		perror("bind");
 		exit(1);
            }

 	if (listen(s, 5) < 0) {
 		perror("listen");
 		exit(1);
 	}
 
        return(s);

}

/* 	signal(SIGCHLD, reapchild);
 	for (;;) {
 		f = accept(s, (struct sockaddr *)0, (int *)0);
 		if (f < 0) {
 			if (errno == EINTR)
 				continue;	/ interrupted; ignore it /
 			perror("accept");
 			exit(1);
 		}
 		switch (fork()) {
 		case 0:
 			close(s);	/ child doesn't need this /
 			signal(SIGCHLD, SIG_DFL);
 			doit(f);
 			exit(0);
 		case -1:
 			perror("fork");
 		default:
 			break;
 		}
 		close(f);
 	}
 #endif INETD
 }

*/ 
 /* this asynchronously cleans up terminated children */
 static void
 reapchild()
 {
 	union wait status;
 
 	while (wait3(&status, WNOHANG|WUNTRACED, 0) > 0)
 		;
 }
 
 /* deal with client -- here we send back the current date */
 doit(fd)
 	int fd;
 {
 	time_t t;
 	char *cp, *ctime();
 
#ifdef notdef
 	/*
 	 * To get the client address (host name + port number):
 	 */
 	struct sockaddr_in from;
 	int fromlen;
 	struct hostent *hp;
 
 	fromlen = sizeof from;
 	if (getpeername(fd, &from, &fromlen) < 0) {
 		perror("getpeername");
 		exit(1);
 	}
 	/* port is ntohs(from.sin_port) */
 	hp = gethostbyaddr(&from.sin_addr, sizeof(struct in_addr), AF_INET);
 	if (hp == NULL) {
 		char *inet_ntoa();
 		/* (numeric) host name is inet_ntoa(from.sin_addr) */
 	} else
 		/* host name is hp->h_name */
#endif
 	time(&t);
 	cp = ctime(&t);
 	write(fd, cp, strlen(cp));
 }



int accept_connect(sock_fd)

int sock_fd;

{

   fd_set fd_mask_xands;
   int client_fd;

      FD_ZERO(&fd_mask_xands);
      FD_SET(sock_fd,&fd_mask_xands);

        if (select(sock_fd+1,&fd_mask_xands,0,0,0) >= 0) {

           client_fd = accept(s, (struct sockaddr *)0, (int *)0);
           if ((client_fd < 0) && (errno != EINTR)) {
                  perror("accept");
                  exit(1);
           }

         close(sock_fd);  /* don't need socket anymore */
         }

    return(client_fd);
}
 

char *read_line(client_fd)

int client_fd;

{

int i;
fd_set fd_mask_xands;
char line_buf[1024];
int num;
char c;



   i = 0;


   FD_ZERO(&fd_mask_xands);
   FD_SET(client_fd,&fd_mask_xands);


   /*  get title data from client */

       for (;;) {
        if (select(client_fd+1,&fd_mask_xands,0,0,0) >= 0) {
           /*  sending process has sent initialization information */
           if ((num = read(client_fd,&c,sizeof (char))) > 0) {
              if (c != '\n') {
                 line_buf[i++] = c;
              } else {
                 break;
              }
           }
           else {
           exit(1);
           }
         } else {
            fprintf(stderr, "error on select\n");
            exit(1);
         }
         FD_SET(client_fd,&fd_mask_xands);
       }

           line_buf[i] = '\0';
           return(line_buf);

}


sopen(port)

int port;

{
    int sock_fd;

    sock_fd = sock_set(port);
    return(accept_connect(sock_fd));

}



copen(port)

int port;

{

     int s;

 
 	/* "localhost" is always an alias for the current machine */

 	if ((s = open_remote_stream("localhost",port)) < 0) {
                fprintf(stderr,"Error openning remote stream\n");
 		exit(1);
        }
 	/* `s'' can be read or written to, using read or write. */

       return(s);

 }
 
 /*
  * Open a stream connection to the PORT service on ``host'',
  * returning the file descriptor or -1.
  */
 int
 open_remote_stream(host,port)
       
 	char *host;
        int port;
 {
 	struct hostent *hp;
 	struct sockaddr_in sin;
 	int s;
 
 	if ((hp = gethostbyname(host)) == NULL) {
 		fprintf(stderr, "%s: unknown host\n", host);
 		return -1;
 	}
 	bzero((char *)&sin, sizeof sin);
 	bcopy(hp->h_addr, (char *)&sin.sin_addr, hp->h_length);
 	sin.sin_family = AF_INET;
 	sin.sin_port = htons(port);
 	if ((s = socket(AF_INET, SOCK_STREAM, 0)) < 0) {
 		perror("socket");
 		return -1;
 	}
 	if (connect(s, (char *)&sin, sizeof sin) < 0) {
 		perror("connect");
 		close(s);
 		return -1;
 	}
 	return s;
 }
