/*
 * $Id: system_cli.c,v 1.1 90/12/07 12:05:17 drew Exp $
 * 
 * $Log:	system_cli.c,v $
# Revision 1.1  90/12/07  12:05:17  drew
# Initial revision
# 
 */

/**********************************************************************
 * Copyright 1990,1991 by University of Toronto, Toronto, Ontario, Canada.
 * 
 *			 All Rights Reserved
 * 
 * Permission to use, copy, modify, distribute, and sell this software
 * and its documentation for any purpose is hereby granted without fee, 
 * provided that the above copyright notice appears in all copies and that 
 * both the copyright notice and this permission notice appear in 
 * supporting documentation, and that the name of University of Toronto 
 * not be used in advertising or publicity pertaining to distribution 
 * of the software without specific, written prior permission.  
 * University of Toronto makes no representations about the suitability 
 * of this software for any purpose. It is provided "as is" without 
 * express or implied warranty. 
 *
 * UNIVERSITY OF TORONTO DISCLAIMS ALL WARRANTIES WITH REGARD TO THIS 
 * SOFTWARE, INCLUDING ALL IMPLIED WARRANTIES OF MERCHANTABILITY AND 
 * FITNESS, IN NO EVENT SHALL UNIVERSITY OF TORONTO BE LIABLE FOR ANY 
 * SPECIAL, INDIRECT OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES WHATSOEVER 
 * RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN ACTION OF 
 * CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF OR IN 
 * CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
 *
 **********************************************************************/

/* system.c	1.2	4/9/88 */

/* This file contains only system functions - that is the functions that
 * get the information the performance monitor is monitoring.  No calls
 * to any X routines should be made here.  The reason for doing this is
 * so that as the X toolkit becomes available and the X window system 
 * improves no changes will have to be made to this file, and as this
 * program is made available for a new type of machine, only this file
 * will need to be changed. */

#include "all.h"
#include "system.h"
#include "packet_struct.c"

#include <stdio.h>
#ifdef SYSV
#include <string.h>
#else
#include <strings.h>
#endif

#include <sys/param.h>
#include <sys/socket.h>
#ifndef SYSV
#include <sys/vm.h>
#include <sys/dk.h>
#endif
#include <nlist.h>
#include <sys/buf.h>
#include <errno.h>

#include <net/if.h>
#include <netinet/in.h>
#include <sys/file.h>

extern char window_name[30];


struct
{
  short stats_specified;
  char *stats_used[NUM_POSSIBLE_STATS];
} sys_params;

/*
 * The array stats always has valid info for stats[i], 0 <= i < num_stats.
 * The info for the k-th possible statistic of interest is recorded, if it is
 * recorded at all, in stats[poss_stats[k]].
 */

int			        poss_stats[NUM_POSSIBLE_STATS];
#define WANT_STAT(x)		(poss_stats[(x)] != NO_STAT)

stat_type                       stats;

static	int num_stats;
int client_fd;


int debug = 0;
#define max(a,b) (a>b ? a:b)


void set_sys_defaults()
/* This routine sets up the default system parameters. */
{
int i,j;

  sys_params.stats_specified = 0; /* user has not specified which to use */

  j = 30;   /* set to maximum graph name width */

   for (i=0; i <= 11; i++) {
        if ((sys_params.stats_used[i] = (char *) calloc(j, sizeof (char))) == NULL) {
      fprintf(stderr, "error allocating storage for stats_used[%d]",i);
        }
  }

}

short sys_process_cmdline(argc,argv,args_found)
int argc;
char **argv;
short args_found[];

/* This routine scans the the command line options for options to the
   system sections of this program */
{

  return(TRUE); /* This routine never fails */
}



/*  This is the modified sys_setup which waits for connection from the
    process which wants to use it.  That process will send the
    title of the graph set, plus the labels of each graph along with
    its max and min value.  */

int sys_setup()

{

int sock_fd;

int i;
char min_val[4], max_val[4];
double graph_min, graph_max;
char *graph_label1,*graph_label2;
char line_buf[1024];
char g_header[1024];
int line_ptr,title_begin;

double atof();
    
fd_set fd_mask_xands;

   client_fd = sopen(PORT);

   strcpy(line_buf,read_line(client_fd));

           i = 0;

           title_begin = cpy_next_token(g_header,line_buf,0); 

           if (strcmp(g_header,"title") == 0) {

                 strcpy(window_name,line_buf+title_begin);

           } else {

              fprintf(stderr, "Error in graph title\n");
           }


 /* get graph initialization values */

   strcpy(line_buf,read_line(client_fd));

           i = 0; 

           line_ptr = cpy_next_token(g_header,line_buf,0); 
           if (strcmp(g_header,"graph") == 0) {

         /*  parse the input string */ 
         /* first five tokens are of form: gname min max label1 label2 */

           while (line_buf[line_ptr] != '\0') {

                 line_ptr += cpy_next_token(sys_params.stats_used[i],line_buf,line_ptr);
                 line_ptr += cpy_next_token(min_val,line_buf,line_ptr);
                 line_ptr += cpy_next_token(max_val,line_buf,line_ptr);
                 graph_min = atof(min_val);
                 graph_max = atof(max_val);

                /*  need some spaced for graph_labels since init_stat just */
                /*    copies pointers */

                if ((graph_label1 = (char *)calloc(15,sizeof(char))) == NULL) {
                       fprintf(stderr,"error allocating memory for label1");
                }
                if ((graph_label2 = (char *)calloc(15,sizeof(char))) == NULL) {
                       fprintf(stderr,"error allocating memory for label2");
                }
                 line_ptr += cpy_next_token(graph_label1,line_buf,line_ptr);
                 line_ptr += cpy_next_token(graph_label2,line_buf,line_ptr);

                sys_params.stats_specified = TRUE; 
                poss_stats[i] = i;  

		init_stat(i, graph_min, graph_max, graph_label1, graph_label2);
		i++;
           }
          
         } else {
             fprintf(stderr,"Expecting graph initialization data.\n");
             exit(1);
         }



    num_stats = i;
    return(num_stats);

}

int cpy_next_token(p1,p2,start)

char *p1, *p2;
int start;

{
    int i;
    i = 0;

    while((p2[start+i] != ' ') && (p2[start+i] != '\0')) {
             p1[i] = p2[start + i];
             i++;
    }
             p1[i] = '\0';
    while(p2[start+i] == ' ') {i++;} 

    return(i);
}

/*  This is the routine which is affected to only read stdio for
    values to be graphed, note: collect_stats is the one which is really
    affected.  
*/

void update_stats(data_read)

char *data_read;
{
int i,n;
char g_name[256];
char g_val_str[256];
double g_val;
char g_header[256];
double atof();

   i = 0;

   i = cpy_next_token(g_header,data_read,0);


   while (data_read[i] != '\0') {
	   if (strcmp(g_header, "data") == 0 ) {
	       /* data should be in form g_name value g_name value ... */

	       i += cpy_next_token(g_name,data_read,i);
	       i += cpy_next_token(g_val_str,data_read,i);
	       g_val = atof(g_val_str);

		update_val(g_name,g_val);

	    }  else {
	       fprintf(stderr,"expected data values \n");
               exit(1);
	    }
    }

}


update_val(gr_name,data)

char *gr_name;
double data;

{
int i,j;

     for (i = 0; i < num_stats; i++) {
        for (j = 0; j < num_stats; j++) {
   
           if (strcmp(sys_params.stats_used[j],gr_name) == 0) {
                stats[j].value = data;
                j = num_stats;
           }
        }
     }
  
}

init_stat(index, minval, maxval, label_1, label_2)
int index;
double minval, maxval;
char *label_1, *label_2;
{
	if WANT_STAT(index) {
		index = poss_stats[index];
		stats[index].min_val = minval;
		stats[index].max_val = maxval;
		stats[index].label[0] = label_1;
		stats[index].label[1] = label_2;
	}
}


min (a, b)
	int a,b;
{
	return(a<b ? a:b);
}
