/*
 * $Id: blot.c,v 1.1 90/12/07 12:29:56 drew Exp $
 * 
 * $Log:	blot.c,v $
# Revision 1.1  90/12/07  12:29:56  drew
# Initial revision
# 
 */

/**********************************************************************
 * Copyright 1990,1991 by University of Toronto, Toronto, Ontario, Canada.
 * 
 *			 All Rights Reserved
 * 
 * Permission to use, copy, modify, distribute, and sell this software
 * and its documentation for any purpose is hereby granted without fee, 
 * provided that the above copyright notice appears in all copies and that 
 * both the copyright notice and this permission notice appear in 
 * supporting documentation, and that the name of University of Toronto 
 * not be used in advertising or publicity pertaining to distribution 
 * of the software without specific, written prior permission.  
 * University of Toronto makes no representations about the suitability 
 * of this software for any purpose. It is provided "as is" without 
 * express or implied warranty. 
 *
 * UNIVERSITY OF TORONTO DISCLAIMS ALL WARRANTIES WITH REGARD TO THIS 
 * SOFTWARE, INCLUDING ALL IMPLIED WARRANTIES OF MERCHANTABILITY AND 
 * FITNESS, IN NO EVENT SHALL UNIVERSITY OF TORONTO BE LIABLE FOR ANY 
 * SPECIAL, INDIRECT OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES WHATSOEVER 
 * RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN ACTION OF 
 * CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF OR IN 
 * CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
 *
 **********************************************************************/

#include <stdio.h>
#include <string.h>
static int blot_range_chars = 48;
static int blot_size = 2;	/* 2 characters for a blot */
static int number_of_blots = 24;	/* number of different sized blots */
static int blot_invert = 0;	/* swap +/- colors ? */
static int blot_first_char = 33;/* starting character of blots */
static char area1[100];		/* used to store mode strings */
static char *bold_on;
static char *attr_off;
static int verbose=0;
char *area=area1;

#define DB(X)

static char blot_buffer[10];

int autoinit_blot()
{
  char bp[1024];
  char *env;
  extern char *tgetstr();
  extern char *getenv();
  tgetent(bp,getenv("TERM"));
  bold_on = tgetstr("md",&area);
  attr_off = tgetstr("me",&area);
  if (bold_on==NULL || attr_off==NULL) {
    fprintf(stderr,"autoinit_blot: terminal doesn't have capability to do bold\n");
    exit(1);
  }
  if (strchr(bold_on,'$')) *strchr(bold_on,'$') = 0; /* kill the delay part */
  if (strchr(attr_off,'$')) *strchr(attr_off,'$') = 0; /* kill the delay part*/
  if ((env = getenv("BLOT")) && (atoi(env)!=0)) {
    blot_size = atoi(env);
    blot_invert = 0;
    if (blot_size < 0) {
      blot_invert = 1;
      blot_size = - blot_size;
    }
    if (strchr(env,':')!=NULL)
      number_of_blots = atoi(strchr(env,':')+1);
    else
      number_of_blots = blot_range_chars/blot_size;
  } else {
    /* fprintf(stderr,"Warning: env var BLOT not set - blots could be inappropriate for terminal\n"); */
    blot_invert = 1;
    blot_size = 2;
  }
  DB( fprintf(stderr,"Blot size=%d  blot invert=%d\n",blot_size,blot_invert); )
}

void SetBlotInvert(x)
int x;
{
  if (x) blot_invert = 1;
  else blot_invert = 0;
}

int GetBlotInvert()
{
  return blot_invert;
}

int BlotSize()
{
  return blot_size;
}

char *BlotStart()
{
  return bold_on;
}

char *BlotEnd()
{
  return attr_off;
}

char *RawBlot(x)
float x;
{
  int i,c,sign;
  i = number_of_blots*x;
  if (i>=0)
    sign=0;
  else {
    sign=1;
    i = -i;
  }
  i++;
  if (i>number_of_blots) i=number_of_blots;
  c = blot_first_char + (number_of_blots-i)*blot_size;
  /* be careful - character 127 doesn't work well, so use zero from the lo range */
  if (!(sign ^ blot_invert) && i>1) {
    c += number_of_blots*blot_size;
  }
  if (c==127) c-= blot_size; /* nasty hack -c=127 happens when blot_size==1 */
  if (verbose) printf("\n%sx=%10.3g Blot start is %3d %s",BlotEnd(),x,c,BlotStart());
  for (i=0;i<blot_size;i++)
    blot_buffer[i] = c+i;
  blot_buffer[i]=0;
  /* sprintf(blot_buffer,"%3d ",c); */
  return blot_buffer;
}

#ifdef MAIN
main(argc,argv)
int argc;
char *argv[];
{
  int range=0;
  float x;
  extern double atof();
  if (strcmp(argv[1],"-v")==0) {
    verbose=1;
    argv++;
  }
  if (strncmp(argv[1],"-range",strlen(argv[1]))==0) {
    range=1;
    argv++;
  }
  autoinit_blot();
  printf("%s",BlotStart());
  if (range) {
    for (x= -1;x<=1+1.0/number_of_blots;x+=1.0/number_of_blots)
      printf("%s",RawBlot(x));
  } else {
    while (*++argv)
      if (strcmp("-n",*argv)) printf("%s",RawBlot(atof(*argv)));
      else printf("\n");
  }
  printf("%s\n",BlotEnd());
}
#endif
