/*
 * $Id: cat-socket.c,v 1.1 90/12/07 12:30:06 drew Exp $
 * 
 * $Log:	cat-socket.c,v $
# Revision 1.1  90/12/07  12:30:06  drew
# Initial revision
# 
 */

/**********************************************************************
 * Copyright 1990,1991 by University of Toronto, Toronto, Ontario, Canada.
 * 
 *			 All Rights Reserved
 * 
 * Permission to use, copy, modify, distribute, and sell this software
 * and its documentation for any purpose is hereby granted without fee, 
 * provided that the above copyright notice appears in all copies and that 
 * both the copyright notice and this permission notice appear in 
 * supporting documentation, and that the name of University of Toronto 
 * not be used in advertising or publicity pertaining to distribution 
 * of the software without specific, written prior permission.  
 * University of Toronto makes no representations about the suitability 
 * of this software for any purpose. It is provided "as is" without 
 * express or implied warranty. 
 *
 * UNIVERSITY OF TORONTO DISCLAIMS ALL WARRANTIES WITH REGARD TO THIS 
 * SOFTWARE, INCLUDING ALL IMPLIED WARRANTIES OF MERCHANTABILITY AND 
 * FITNESS, IN NO EVENT SHALL UNIVERSITY OF TORONTO BE LIABLE FOR ANY 
 * SPECIAL, INDIRECT OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES WHATSOEVER 
 * RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN ACTION OF 
 * CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF OR IN 
 * CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
 *
 **********************************************************************/

#include <stdio.h>
#include <errno.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <netdb.h>

  /***
   * Cat-socket connects to a socket that has already been posted.
   * It can then either
   *   -r: read from the socket and write to standard output.
   *   -w: read from standard input and write to the socket (default).
   * Compile with -lbsd flags on sgi machines
   */

/***
 * IOpenNumberedSocket - tries to connect to the given socket on the given
 * host.
 * A file is attached to the socket with the given mode ("r" or "w").
 */

FILE *IOpenNumberedSocket(port, host, mode)
int port;
char *host;
char *mode;
{
  FILE *f;
  int sock;
  struct sockaddr_in server;
  struct hostent *hp, *gethostbyname();
  if (host==NULL) host = "localhost";

  /* create the socket */
  sock = socket(AF_INET, SOCK_STREAM, 0);
  if (sock<0) return NULL;

  /* connect the socket to port & host */
  server.sin_family = AF_INET;
  hp = gethostbyname(host);
  if (hp == 0) return NULL;
  bcopy(hp->h_addr, &server.sin_addr, hp->h_length);
  server.sin_port = htons(port);

  if (connect(sock, &server, sizeof(server)) < 0)
    return NULL;

  f = fdopen(sock, mode);
  return f;
}

FILE *IOpenNamedSocket(name, mode)
char *name;
char *mode;
{
  fprintf(stderr, "IOpenNamedSocket not yet implemented\n");
  return NULL;
}

main(argc, argv)
int argc;
char *argv[];
{
  FILE *f;
  int c;
  char *mode = "w";
  char *my_name = argv[0];
  extern char *sys_errlist[];
  if (argc>1 && !strcmp(argv[1], "-r")) {
    mode = "r";
    argv++;
    argc--;
  } else if (argc>1 && !strcmp(argv[1], "-w")) {
    mode = "w";
    argv++;
    argc--;
  }
  if (argc==2) {
    f = IOpenNamedSocket(argv[1], mode);
  } else if (argc==3) {
    f = IOpenNumberedSocket(atoi(argv[1]), argv[2], mode);
  } else {
    fprintf(stderr,
    "Usage: %s [-r|-w] <port#> <host-name>\nor:    %s [-r|-w] <socket-name> (not yet implemented)\n",
	    my_name, my_name);
    fprintf
      (stderr, "%s\n%s\n%s\n%s\n",
       "Cat-socket connects to a socket that has already been posted.",
       "It can then either",
       "   -r: read from the socket and write to standard output.",
       "   -w: read from standard input and write to the socket (default).");
    exit(1);
  }
  if (f==NULL) {
    fprintf(stderr, "%s: Can't open socket: %s\n",
	    my_name, sys_errlist[errno]);
    exit(1);
  }
  if (*mode=='w') {
    while ((c=getchar())!=EOF) {
      fputc(c, f);
      if (c=='\n') fflush(f);
    }
  } else {
    while ((c=fgetc(f))!=EOF) {
      putchar(c);
    }
  }
}
