/*
 * $Id: post-socket.c,v 1.1 90/12/07 12:30:21 drew Exp $
 * 
 * $Log:	post-socket.c,v $
# Revision 1.1  90/12/07  12:30:21  drew
# Initial revision
# 
 */

/**********************************************************************
 * Copyright 1990,1991 by University of Toronto, Toronto, Ontario, Canada.
 * 
 *			 All Rights Reserved
 * 
 * Permission to use, copy, modify, distribute, and sell this software
 * and its documentation for any purpose is hereby granted without fee, 
 * provided that the above copyright notice appears in all copies and that 
 * both the copyright notice and this permission notice appear in 
 * supporting documentation, and that the name of University of Toronto 
 * not be used in advertising or publicity pertaining to distribution 
 * of the software without specific, written prior permission.  
 * University of Toronto makes no representations about the suitability 
 * of this software for any purpose. It is provided "as is" without 
 * express or implied warranty. 
 *
 * UNIVERSITY OF TORONTO DISCLAIMS ALL WARRANTIES WITH REGARD TO THIS 
 * SOFTWARE, INCLUDING ALL IMPLIED WARRANTIES OF MERCHANTABILITY AND 
 * FITNESS, IN NO EVENT SHALL UNIVERSITY OF TORONTO BE LIABLE FOR ANY 
 * SPECIAL, INDIRECT OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES WHATSOEVER 
 * RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN ACTION OF 
 * CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF OR IN 
 * CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
 *
 **********************************************************************/

#include <stdio.h>
#include <errno.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <netdb.h>

  /***
   * Post-socket posts a socket and then listens for a conection
   * It can then either
   *   -r: read from the socket and write to standard output (default).
   *   -w: read from standard input and write to the socket.
   * Compile with -lbsd flags on sgi machines
   */

#ifndef MAXHOSTNAMELEN
#define MAXHOSTNAMELEN 128
#endif  MAXHOSTNAMELEN

char *my_name;
int verbose = 0;

/***
 * IPostNumberedSocket - posts and listens on the given port number on
 * the current host.
 * A file is attached to the socket with the given mode ("r" or "w").
 */

FILE *IPostNumberedSocket(port, mode)
int port;
char *mode;
{
  FILE *f;
  int msgsock;
  int sock;
  char hostname[MAXHOSTNAMELEN];
  int length;
  struct sockaddr_in server;
  gethostname(hostname, MAXHOSTNAMELEN);
  if ((sock = socket(AF_INET, SOCK_STREAM, 0))<0)
    error("socket");
  server.sin_family = AF_INET;
  server.sin_addr.s_addr = INADDR_ANY;
  server.sin_port = htons((unsigned short) port);
  if (bind(sock, &server, sizeof(server)))
    error("bind");
  /* Find out assigned port number and print it out */
  length = sizeof(server);
  if (getsockname(sock, &server, &length)) error("getsockname");

  if (verbose)
    fprintf(stderr, "Socket has port #%d on %s\n",
	    ntohs(server.sin_port), hostname);

  listen(sock, 5);
  if ((msgsock = accept(sock, 0, 0)) == -1)
    error("accept");
  if (verbose)
    fprintf(stderr, "Accepted connection\n");
  if (!(f = fdopen(msgsock, mode)))
    error("fdopen");

  return f;
}

/*
 * System error: print message and give up.
 */
error(msg)
char *msg;
{
  fprintf(stderr, "%s: ", my_name);
  perror(msg);
  exit(1);
}

FILE *IPostNamedSocket(name, mode)
char *name;
char *mode;
{
  fprintf(stderr, "IPostNamedSocket not yet implemented\n");
  return NULL;
}

main(argc, argv)
int argc;
char *argv[];
{
  FILE *f;
  int c;
  char *mode = "r";
  extern char *sys_errlist[];
  my_name = argv[0];
  if (argc>1 && !strncmp(argv[1], "-v", 2)) {
    verbose = 1;
    argv++;
    argc--;
  }
  if (argc>1 && !strcmp(argv[1], "-r")) {
    mode = "r";
    argv++;
    argc--;
  } else if (argc>1 && !strcmp(argv[1], "-w")) {
    mode = "w";
    argv++;
    argc--;
  }
  if (argc==2 && atoi(argv[1])==0) {
    f = IPostNamedSocket(argv[1], mode);
  } else if (argc==2 && atoi(argv[1])>0) {
    f = IPostNumberedSocket(atoi(argv[1]), mode);
  } else {
    fprintf(stderr,
	    "Usage: %s [-verbose] [-r|-w] <port#>\nor:    %s [-verbose] [-r|-w] <socket-name> (not yet implemented)\n",
	    my_name, my_name);
    fprintf
      (stderr, "%s\n%s\n%s\n%s\n",
       "Post-socket posts a socket and then listens for a conection.",
       "It can then either",
       "   -r: read from the socket and write to standard output (default).",
       "   -w: read from standard input and write to the socket.");
    exit(1);
  }
  if (f==NULL) {
    fprintf(stderr, "%s: Can't post socket: %s\n",
	    my_name, sys_errlist[errno]);
    exit(1);
  }
  if (*mode=='w') {
    while ((c=getchar())!=EOF) {
      fputc(c, f);
      if (c=='\n') fflush(f);
    }
  } else {
    while ((c=fgetc(f))!=EOF) {
      putchar(c);
    }
  }
}
